// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*



////////////////////////////////////////////////////////////////////////

#ifndef  __VSDB_ANALYSIS_TOOL_JOB_H__
#define  __VSDB_ANALYSIS_TOOL_JOB_H__

////////////////////////////////////////////////////////////////////////

#include <iostream>

#include "vx_util/vx_util.h"
#include "vx_data_grids/grid.h"
#include "vx_wrfdata/vx_wrfdata.h"
#include "vx_cal/vx_cal.h"
#include "vx_analysis_util/vsdb_line.h"

////////////////////////////////////////////////////////////////////////

//
// Defaults to be used if not specified by the user
//
static const double default_alpha         = 0.05;
static const int    default_boot_interval = 1;
static const double default_boot_rep_prop = 1.0;
static const int    default_n_boot_rep    = 1000;
static const char   default_boot_rng[]    = "mt19937";
static const char   default_boot_seed[]   = "";

////////////////////////////////////////////////////////////////////////

//
// Enumerate all the possible VSDB Analysis Job Types
//
enum VSDBJobType {

   vsdb_job_filter   = 0, // Filter out the VSDB data and write the
                          // lines to the filename specified.

   vsdb_job_summary  = 1, // Compute min, max, mean, stdev and
                          // percentiles for a column of data.

   vsdb_job_aggr     = 2, // For contingency table line types (fho, ctc,
                          // ctp, cfp, cop) or partial sums line types
                          // (sl1l2, sal1l2, vl1l2, val1l2) aggregate
                          // the raw counts and generate the
                          // corresponding output line type containing
                          // the aggregated counts.  For the intensity
                          // scale line type, aggregate the scores.

   vsdb_job_aggr_cts = 3, // For contingency table line types (fho, ctc,
                          // ctp, cfp, cop), aggregate the raw counts and
                          // generate a contingecy table statistics (cts)
                          // output line.

   vsdb_job_aggr_cnt = 4, // For partial sum line types (SL1L2 and SAL1L2)
                          // aggregate the partial sums and generate a
                          // continuous statistics (cnt) output line.

   vsdb_job_aggr_mpr = 5, // For the matched pair line type (MPR),
                          // aggregate the matched pairs and generate
                          // an output line of the type specified by
                          // the user.

   vsdb_job_go_index = 6, // Compute the GO Index.

   no_vsdb_job_type  = 7  // default value

};
static const int n_vsdbjobtypes = 8;

////////////////////////////////////////////////////////////////////////

class VSDBAnalysisJob {

   private:

      void init_from_scratch();

      void assign(const VSDBAnalysisJob &);



   public:

      VSDBAnalysisJob();
     ~VSDBAnalysisJob();
      VSDBAnalysisJob(const VSDBAnalysisJob &);
      VSDBAnalysisJob & operator=(const VSDBAnalysisJob &);

      void clear();

      void dump(ostream &, int depth = 0) const;

      int  is_keeper(const VSDBLine &) const;

      void parse_job_command(const char *);

      int  set_job_type(const char *);
      void set_col_name(const char *);
      void set_boot_rng(const char *);
      void set_boot_seed(const char *);

      void get_jobstring(char *);

      //
      // Job Type
      //
      VSDBJobType job_type;

      //
      // Variables used to stratify the input VSDB lines
      //
      StringArray model;

      NumArray    fcst_lead; // stored in seconds

      unixtime    valid_beg;
      unixtime    valid_end;

      unixtime    init_beg;
      unixtime    init_end;

      NumArray    init_hour; // stored in seconds

      StringArray obtype;

      StringArray mask;

      StringArray line_type;

      StringArray var;

      StringArray level;

      StringArray interp_mthd;
      NumArray    interp_pnts;

      //
      // Indicate whether the VSDB lines used to compute the statistic
      // are dumped to a file
      //
      StringArray dump_row;

      //
      // Column Name
      //
      char *col_name;

      //
      // Alpha value to be used for computing confidence intervals
      //
      double alpha;

      //
      // Variables to be used for computing bootstrap confidence
      // intervals

      //
      // Type of bootstrap confidence interval method:
      //    0 = BCa, 1 = Percentile (Default = 1)
      //
      int boot_interval;

      //
      // When using the percentile method, this is the proportion
      // of n to be resampled. (Default = 1.0)
      //
      double boot_rep_prop;

      //
      // Number of bootstrap replicates to be done. (Default = 1000)
      //
      int n_boot_rep;

      //
      // Name and seed value for the bootstrap random number generator.
      // (Default = "mnt19937" and "")
      //
      char *boot_rng;
      char *boot_seed;

      //
      // Variables used for the vsdb_job_aggr_mpr job type
      //
      VSDBLineType out_line_type;

      StringArray  threshold;
      SingleThresh thresh;

      StringArray  mask_grid;
      Grid         grid;

      StringArray  mask_poly;
      Polyline     poly;

      int is_in_grid_mask(double, double) const;
      int is_in_poly_mask(double, double) const;
};

////////////////////////////////////////////////////////////////////////

#endif   /*  __VSDB_ANALYSIS_TOOL_JOB_H__  */

////////////////////////////////////////////////////////////////////////
