// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*

using namespace std;

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <cstdio>
#include <cmath>

#include "vx_contable/vx_contable.h"
#include "vx_util/vx_util.h"
#include "vx_met_util/vx_met_util.h"

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::baser(int &status) const {
   double v;

   v = oy_tp(status);

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::baser_ci(int &status, double alpha,
                                    double &cl, double &cu) const {
   double v;

   v = baser(status);

   compute_proportion_ci(v, n(), alpha, cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::fmean(int &status) const {
   double v;

   v = fy_tp(status);

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::fmean_ci(int &status, double alpha,
                                    double &cl, double &cu) const {
   double v;

   v = fmean(status);

   compute_proportion_ci(v, n(), alpha, cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::accuracy(int &status) const {
   int n_int, d_int;
   double num, denom, v;

   status = 0;

   n_int = fy_oy() + fn_on();
   d_int = n();

   if(d_int == 0) {

      status = 1;
      return(bad_data_double);
   }

   num   = (double) n_int;
   denom = (double) d_int;

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::accuracy_ci(int &status, double alpha,
                                       double &cl, double &cu) const {
   double v;

   v = accuracy(status);

   compute_proportion_ci(v, n(), alpha, cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////
//
// Taken from eq. 7.9, page 241 in Wilks
//
////////////////////////////////////////////////////////////////////////

double TTContingencyTable::fbias(int &status) const {
   int n_int, d_int;
   double num, denom, v;

   status = 0;

   n_int = fy();
   d_int = oy();

   if(d_int == 0) {

      status = 1;
      return(bad_data_double);
   }

   num   = (double) n_int;
   denom = (double) d_int;

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////
//
// Taken from eq. 7.7, page 240 in Wilks
//
////////////////////////////////////////////////////////////////////////

double TTContingencyTable::pod_yes(int &status) const {
   int n_int, d_int;
   double num, denom, v;

   status = 0;

   n_int = fy_oy();
   d_int = oy();

   if(d_int == 0) {

      status = 1;
      return(bad_data_double);
   }

   num   = (double) n_int;
   denom = (double) d_int;

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::pod_yes_ci(int &status, double alpha,
                                      double &cl, double &cu) const {
   double v;

   v = pod_yes(status);

   compute_proportion_ci(v, n(), alpha, cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::pod_no(int &status) const {
   int n_int, d_int;
   double num, denom, v;

   status = 0;

   n_int = fn_on();
   d_int = on();

   if(d_int == 0) {

      status = 1;
      return(bad_data_double);
   }

   num   = (double) n_int;
   denom = (double) d_int;

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::pod_no_ci(int &status, double alpha,
                                     double &cl, double &cu) const {
   double v;

   v = pod_no(status);

   compute_proportion_ci(v, n(), alpha, cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::pofd(int &status) const {
   double d, v;

   d = pod_no(status);

   if(status == 1) return(bad_data_double);

   v = 1.0 - d;

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::pofd_ci(int &status, double alpha,
                                   double &cl, double &cu) const {
   double v;

   v = pofd(status);

   compute_proportion_ci(v, n(), alpha, cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////
//
// Taken from eq. 7.8, page 241 in Wilks
//
////////////////////////////////////////////////////////////////////////

double TTContingencyTable::far(int &status) const {
   int n_int, d_int;
   double num, denom, v;

   status = 0;

   n_int = fy_on();
   d_int = fy();

   if(d_int == 0) {

      status = 1;
      return(bad_data_double);
   }

   num   = (double) n_int;
   denom = (double) d_int;

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::far_ci(int &status, double alpha,
                                  double &cl, double &cu) const {
   double v;

   v = far(status);

   compute_proportion_ci(v, n(), alpha, cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////
//
// Taken from eq. 7.6, page 240 in Wilks
//
////////////////////////////////////////////////////////////////////////

double TTContingencyTable::csi(int &status) const {
   int n_int, d_int;
   double num, denom, v;

   status = 0;

   n_int = fy_oy();
   d_int = n() - fn_on();

   if(d_int == 0) {

      status = 1;
      return(bad_data_double);
   }

   num   = (double) n_int;
   denom = (double) d_int;

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::csi_ci(int &status, double alpha,
                                  double &cl, double &cu) const {
   double v;

   v = csi(status);

   compute_proportion_ci(v, n(), alpha, cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////
//
// Equations given by Barb Brown
//
////////////////////////////////////////////////////////////////////////

double TTContingencyTable::gss(int &status) const {
   int a, b, c;
   double q, num, denom, v;

   status = 0;

   a = fy_oy();
   b = fy_on();
   c = fn_oy();

   if(n() == 0) {

      status = 1;
      return(bad_data_double);
   }

   q     = (double) (a + b)*(a + c)/(n());
   num   = (double) (a - q);
   denom = (double) (a + b + c - q);

   if(denom == 0.0) {

      status = 1;
      return(bad_data_double);
   }

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////
//
// Taken from eq. 7.12, page 249 in Wilks
//
////////////////////////////////////////////////////////////////////////

double TTContingencyTable::hk(int &status) const {
   int n_int, d_int;
   int a, b, c, d;
   double num, denom, v;

   status = 0;

   a = fy_oy();
   b = fy_on();
   c = fn_oy();
   d = fn_on();

   n_int = a*d - b*c;
   d_int = (a + c)*(b + d);

   if(d_int == 0) {

      status = 1;
      return(bad_data_double);
   }

   num   = (double) n_int;
   denom = (double) d_int;

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::hk_ci(int &status, double alpha,
                                 double &cl, double &cu) const {
   double v;

   v = hk(status);

   compute_hk_ci(v, alpha, fy_oy(), fy_on(), fn_oy(), fn_on(), cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////
//
// Taken from eq. 7.10, page 249 in Wilks
//
////////////////////////////////////////////////////////////////////////

double TTContingencyTable::hss(int &status) const {
   int n_int, d_int;
   int a, b, c, d;
   double num, denom, v;

   status = 0;

   a = fy_oy();
   b = fy_on();
   c = fn_oy();
   d = fn_on();

   n_int = 2*(a*d - b*c);
   d_int = (a + c)*(c + d) + (a + b)*(b + d);

   if(d_int == 0) {
      
      status = 1;
      return(bad_data_double);
   }

   num   = (double) n_int;
   denom = (double) d_int;

   v = num/denom;

   return(v);
}


////////////////////////////////////////////////////////////////////////
//
// Taken from eq. 7.10, page 249 in Wilks
//
////////////////////////////////////////////////////////////////////////

double TTContingencyTable::odds(int &status) const {
   int s;
   double num, denom, v, py, pn;

   status = 0;

   py = pod_yes(s);
   pn = pofd(s);

   if(py == 1.0 || py == bad_data_double ||
      pn == 1.0 || pn == bad_data_double) {

      status = 1;
      return(bad_data_double);
   }

   num   = py/(1 - py);
   denom = pn/(1 - pn);

   if(denom == 0.0) {

      status = 1;
      return(bad_data_double);
   }

   v = num/denom;

   return(v);
}

////////////////////////////////////////////////////////////////////////

double TTContingencyTable::odds_ci(int &status, double alpha,
                                   double &cl, double &cu) const {
   double v;

   v = odds(status);

   compute_woolf_ci(v, alpha, fy_oy(), fy_on(), fn_oy(), fn_on(), cl, cu);

   return(v);
}

////////////////////////////////////////////////////////////////////////
