// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*

///////////////////////////////////////////////////////////////////////////////
//
//  Filename:   interest.cc
//
//  Description:
//
//
//  Mod#   Date      Name           Description
//  ----   ----      ----           -----------
//  000    04-15-05  Halley Gotway
//
///////////////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "vx_met_util/vx_met_util.h"
#include "vx_wrfmode/interest.h"
#include "vx_math/vx_math.h"

////////////////////////////////////////////////////////////////////////
//
// Code for class SingleFeature
//
////////////////////////////////////////////////////////////////////////

SingleFeature::SingleFeature() {
   clear();
}

////////////////////////////////////////////////////////////////////////

SingleFeature::~SingleFeature() {
   clear();
}

////////////////////////////////////////////////////////////////////////

SingleFeature::SingleFeature(const SingleFeature &s) {
   assign(s);
}

////////////////////////////////////////////////////////////////////////

SingleFeature & SingleFeature::operator=(const SingleFeature &s) {

   if(this == &s) return(*this);

   assign(s);

   return(*this);
}

////////////////////////////////////////////////////////////////////////

void SingleFeature::clear() {

   area = 0.0;
   area_filter = 0.0;
   xbar = 0.0;
   ybar = 0.0;
   angle_degrees = 0.0;
   aspect_ratio = 1.0;
   complexity = 0.0;
   percentile_intensity = 0.0;

   Raw    = (const WrfData *) 0;
   Mask   = (const WrfData *) 0;

   convex_hull.clear();
   boundary.clear();

   return;
}

////////////////////////////////////////////////////////////////////////

void SingleFeature::assign(const SingleFeature & s) {

   clear();

   area = s.area;
   area_filter = s.area_filter;
   xbar = s.xbar;
   ybar = s.ybar;
   angle_degrees = s.angle_degrees;
   aspect_ratio = s.aspect_ratio;
   complexity = s.complexity;
   percentile_intensity = s.percentile_intensity;
   
   Raw = s.Raw;
   Mask = s.Mask;

   convex_hull = s.convex_hull;
   boundary = s.boundary;

   return;
}

////////////////////////////////////////////////////////////////////////

void SingleFeature::set(const WrfData &raw, const WrfData &mask, const double p) {

   clear();

   int x, y;
   int count;
   double * data = (double *) 0;
   double length, width;

   Raw  = &raw;
   Mask = &mask;

   //
   // Object area
   //
   area = Mask->area();

   //
   // area_filter: the area of the non-zero raw field inside
   // the mask area
   //
   area_filter = (double) wrfdata_intersection(*Raw, *Mask);

   //
   // Centroid
   //
   Mask->centroid(xbar, ybar);

   //
   // Axis angle
   //
   angle_degrees = Mask->angle_degrees();

   //
   // Aspect ratio
   //
   Mask->calc_length_width(length, width);
   aspect_ratio = width/length;
   
   //
   // Complexity
   //
   complexity = Mask->complexity();
   
   //
   // Nth Percentile of intensity
   //
   count = 0;
   for(x=0; x<(Mask->get_nx()); x++) {
      for(y=0; y<(Mask->get_ny()); y++) {

         if((Mask->f_is_on(x, y)) && 
            (Raw->is_valid_data(x, y)) &&
            (Raw->get_xy_int(x, y) > 0)) count++;
      }
   }

   data = new double [count];

   if(!data) {
      cerr << "\n\nERROR: SingleFeature::set() -> "
	   << "memory allocation error\n\n" << flush;
      exit(1);
   }

   count = 0;
   for(x=0; x<(Mask->get_nx()); x++) {
      for(y=0; y<(Mask->get_ny()); y++) {

         if((Mask->f_is_on(x, y)) &&
            (Raw->is_valid_data(x, y)) &&
            (Raw->get_xy_int(x, y) > 0)) {
            data[count++] = (double) (Raw->get_xy_double(x, y));
         }
      }
   }
   
   sort(data, count);
   percentile_intensity = percentile(data, count, p/100.0);

   //
   // Convex hull
   //
   convex_hull = Mask->convex_hull();

   //
   // Boundary
   //
   boundary = Mask->single_boundary();

   //
   // Done
   //

   if(data) { delete [] data;  data = (double *) 0; }

   return;
}

////////////////////////////////////////////////////////////////////////
//
// Code for class PairFeature
//
////////////////////////////////////////////////////////////////////////

PairFeature::PairFeature() {
   clear();
}

////////////////////////////////////////////////////////////////////////

PairFeature::~PairFeature() {
   clear();
}

////////////////////////////////////////////////////////////////////////

PairFeature::PairFeature(const PairFeature &p) {
   clear();
   assign(p);
}

////////////////////////////////////////////////////////////////////////

PairFeature & PairFeature::operator=(const PairFeature &p) {

   if(this == &p) return(*this);

   assign(p);

   return(*this);
}

////////////////////////////////////////////////////////////////////////

void PairFeature::clear() {

   centroid_dist              = 0.0;
   boundary_dist              = 0.0;
   convex_hull_dist           = 0.0;
   angle_diff                 = 0.0;
   area_ratio                 = 0.0;
   intersection_area          = 0.0;
   union_area                 = 0.0;
   complexity_ratio           = 0.0;
   percentile_intensity_ratio = 0.0;

   Obs  = (const SingleFeature *) 0;// DON'T delete
   Fcst = (const SingleFeature *) 0;// DON'T delete

   return;
}

////////////////////////////////////////////////////////////////////////

void PairFeature::assign(const PairFeature &p) {

   clear();
   
   centroid_dist              = p.centroid_dist;
   boundary_dist              = p.boundary_dist;
   convex_hull_dist           = p.convex_hull_dist;
   angle_diff                 = p.angle_diff;
   area_ratio                 = p.area_ratio;
   intersection_area          = p.intersection_area;
   union_area                 = p.union_area;
   complexity_ratio           = p.complexity_ratio;
   percentile_intensity_ratio = p.percentile_intensity_ratio;

   Obs = p.Obs;
   Fcst = p.Fcst;

   return;
}

////////////////////////////////////////////////////////////////////////

void PairFeature::set(const SingleFeature &fcst, 
                      const SingleFeature &obs,
                      double max_centroid_dist) {
   clear();

   Fcst = &fcst;
   Obs  = &obs;

   int x, y;
   int fcst_on, obs_on;
   double dx, dy;
   double a1, a2;

   //
   // Centroid distance
   //
   dx = (Obs->xbar) - (Fcst->xbar);
   dy = (Obs->ybar) - (Fcst->ybar);

   centroid_dist = sqrt(dx*dx + dy*dy);
   if(centroid_dist > max_centroid_dist) return;
   
   //
   // Boundary distance
   //
   boundary_dist = polyline_dist(Obs->boundary, Fcst->boundary);
   
   //
   // Convex hull distance
   //
   convex_hull_dist = polyline_dist(Obs->convex_hull, Fcst->convex_hull);

   //
   // Angle diff
   //
   a1 = Obs->angle_degrees;
   a2 = Fcst->angle_degrees;
   angle_diff = angle_between(a1, a2);

   //
   // Area ratio
   //
   area_ratio = min( (Obs->area)/(Fcst->area), 
                     (Fcst->area)/(Obs->area) );
   
   //
   // Intersection and union areas
   //
   intersection_area = union_area = 0.0;
   for(x=0; x<(Fcst->Mask->get_nx()); ++x) {
      for(y=0; y<(Fcst->Mask->get_ny()); ++y) {

         fcst_on = Fcst->Mask->s_is_on(x, y);
         obs_on  =  Obs->Mask->s_is_on(x, y);

         if(fcst_on && obs_on) intersection_area++;
         if(fcst_on || obs_on) union_area++;
      }
   }

   //
   // Complexity Ratio
   //
   
   // Both complexities are non-zero
   if(Obs->complexity > 0.0 && Fcst->complexity > 0.0) {
      complexity_ratio = min( (Obs->complexity)/(Fcst->complexity), 
                              (Fcst->complexity)/(Obs->complexity) );
   }
   // At least one complexity is zero
   else {
      complexity_ratio = 1.0 - max(Obs->complexity, Fcst->complexity);
   }

   //
   // Nth Percentile intensity ratio
   //
   
   // Both percentile intensities are non-zero
   if(Obs->percentile_intensity > 0.0 && Fcst->percentile_intensity > 0.0) {
      percentile_intensity_ratio = min( (Obs->percentile_intensity)/(Fcst->percentile_intensity),
                                        (Fcst->percentile_intensity)/(Obs->percentile_intensity) );
   }
   // At least one percentile intensity is zero                             
   else {
      percentile_intensity_ratio = 0.0;
   }
   
   //
   // Done
   //

   return;
}

////////////////////////////////////////////////////////////////////////
//
// Code for misc functions
//
////////////////////////////////////////////////////////////////////////

ostream & operator<<(ostream & out, const SingleFeature & s) {
   WrfDataHeader h = s.Raw->get_header();
   const char * units = h.units;

   out << "Object Number          = "  << (s.object_number)    << "\n";
   out << "Area                   = "  << nint(s.area)         << "\n";
   out << "Area Filter            = "  << nint(s.area_filter)         << "\n";
   out << "Centroid               = (" << (s.xbar) << ", " << (s.ybar) << ")\n";
   out << "Axis Angle             = "  << (s.angle_degrees)    << "\n";
   out << "Aspect Ratio           = "  << (s.aspect_ratio)     << "\n";
   out << "Complexity             = "  << (s.complexity)     << "\n";
   out << "Percentile Intensity   = "  << (s.percentile_intensity) << " " << units << "\n";
   out.flush();

   return(out);
}

////////////////////////////////////////////////////////////////////////

ostream & operator<<(ostream & out, const PairFeature & p) {

   out << "Fcst Object Number                = " << (p.Fcst->object_number)    << "\n";
   out << "Obs  Object Number                = " << (p.Obs->object_number)     << "\n";
   out << "Pair Number                       = " << p.pair_number              << "\n";
   out << "Centroid Distance                 = " << (p.centroid_dist)          << "\n";
   out << "Boundary Distance                 = " << (p.boundary_dist)          << "\n";
   out << "Convex Hull Distance              = " << (p.convex_hull_dist)       << "\n";   
   out << "Angle Difference                  = " << (p.angle_diff)             << "\n";
   out << "Area Ratio                        = " << (p.area_ratio)             << "\n";
   out << "Intersection Area                 = " << nint(p.intersection_area)  << "\n";
   out << "Union Area                        = " << nint(p.union_area)         << "\n";
   out << "Percentile Intensity Ratio = " << (p.percentile_intensity_ratio) << "\n";
   out.flush();

   return(out);
}

////////////////////////////////////////////////////////////////////////
