// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*

///////////////////////////////////////////////////////////////////////
//
//   Filename:   grid_stat.cc
//
//   Description:
//
//   Mod#   Date      Name            Description
//   ----   ----      ----            -----------
//   000    04/18/07  Halley Gotway   New
//   001    10/05/07  Halley Gotway   Add checks to only compute
//          confidence intervals for sufficiently large sample sizes.
//   002    10/19/07  Halley Gotway   Add ability to smooth the forecast
//          field using the interp_method and interp_width parameters
//   003    01/11/08  Halley Gotway   Modify sprintf statements which
//          use the grib code abbreviation string
//   004    02/06/08  Halley Gotway   Modify to read the updated NetCDF
//          output of PCP-Combine
//   005    02/11/08  Halley Gotway   Remove BIAS from the CNT line
//          since it's the same as the ME.
//   006    02/12/08  Halley Gotway   Fix bug in writing COP line to
//                    write out OY and ON rather than FY and FN.
//   007    02/14/08  Halley Gotway   Apply the smoothing operation to
//                    both forecast and observation fields rather than
//                    just the forecast.
//   008    02/25/08  Halley Gotway   Write the VSDB lines using the
//                    routines in the vx_met_util library.
//   009    02/25/08  Halley Gotway   Add support for the NBRCTS and
//                    NBRCNT line types.
//   010    07/01/08  Halley Gotway   Add the rank_corr_flag to the
//                    config file to disable computing rank correlations.
//
////////////////////////////////////////////////////////////////////////

using namespace std;

#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <ctype.h>
#include <dirent.h>
#include <iostream>
#include <fstream>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "netcdf.hh"
#include "vx_grib_classes/grib_classes.h"

#include "grid_stat_Conf.h"

#include "vx_wrfdata/vx_wrfdata.h"
#include "vx_met_util/vx_met_util.h"
#include "vx_data_grids/grid.h"
#include "vx_util/vx_util.h"
#include "vx_cal/vx_cal.h"
#include "vx_math/vx_math.h"
#include "vx_contable/vx_contable.h"
#include "vx_gsl_prob/vx_gsl_prob.h"
#include "vx_econfig/result.h"

////////////////////////////////////////////////////////////////////////
//
// Create output files using the following naming convention:
//
//    grid_stat_HHMMSSL_YYYYMMDD_HHMMSSV.vsdb/.txt/.nc
//
//    Where L indicates lead time and V indicates valid time.
//
////////////////////////////////////////////////////////////////////////

static const char * program_name = "grid_stat";

////////////////////////////////////////////////////////////////////////

// Pointer to the random number generator to be used
static gsl_rng *rng_ptr = (gsl_rng *) 0;

// Input configuration file
static char config_file[max_str_len];

// Input forecast Grib or NetCDF files
static char fcst_file[max_str_len];
static GribFile fcst_gb_file;
static NcFile  *fcst_nc_file;

// Input observation Grib or NetCDF files
static char obs_file[max_str_len];
static GribFile obs_gb_file;
static NcFile  *obs_nc_file;

// Output NetCDF file
static char    out_nc_file[max_str_len];
static NcFile  *nc_out = (NcFile *) 0;
static NcDim   *lat_dim, *lon_dim;
static NcVar   *lat_var, *lon_var;

// Grid definition data structures
static PlateCarreeData   pc_data;
static LambertData       lc_data;
static StereographicData st_data;

// Variables for command line arguments
static grid_stat_Conf conf;
static char out_dir[max_str_len];
static int verbosity = 1;

// Indices for the output flag types in the configuration file
static const int i_fho    = 0;
static const int i_ctc    = 1;
static const int i_ctp    = 2;
static const int i_cfp    = 3;
static const int i_cop    = 4;
static const int i_cts    = 5;
static const int i_cnt    = 6;
static const int i_sl1l2  = 7;
static const int i_nbrctc = 8;
static const int i_nbrcts = 9;
static const int i_nbrcnt = 10;
static const int i_nc     = 11;

// Grid variables
static Grid grid, fcst_grid, obs_grid;

// Flags to indicate the format of the input files.
//    0 = netCDF (output of pcp_combine)
//    1 = Grib Version 1
static int fcst_fmt_flag = 0;
static int obs_fmt_flag = 0;

// Strings to be output in the VSDB file
static char abbr_str[max_str_len];
static char msg_type[max_str_len];
static char field_name[max_str_len];
static char level_name[max_str_len];

// Number of Grib Codes to be verified
static int n_gc;

// Objects to hold Grib Code information
// Sized as gc_info[n_gc]
static GCInfo *gc_info = (GCInfo *) 0;

// Objects to hold threshold information for each Grib Code
// Sized as gc_ti[n_gc]
static ThreshInfo *gc_ti = (ThreshInfo *) 0;

// Multiple smoothing operations to be performed on the forecast field
// prior to verification
static int  n_interp;
static int *interp_mthd = (int *) 0;
static int *interp_wdth = (int *) 0;

// Multiple values used to define neighborhoods when computing the
// neighborhood verification measures
static int n_nbr_wdth;

// Multiple verification masking regions to be applied to the forecast
// and observation fields
static int      n_masks;
static WrfData *mask_wd    = (WrfData *) 0;
static char   **mask_names = (char **) 0;

// Multiple values of alpha to be used when computing confidence intervals
// for continuous variables
static int n_ci_alpha;

// Threshold to apply to the fractional coverage of the neigbhorhood method
static ThreshInfo frac_ti;

// Output file streams for creating VSDB and text files
static ofstream *vsdb_out;
static char vsdb_file[max_str_len];

static ofstream *fho_out, *ctc_out, *ctp_out,   *cfp_out,    *cop_out;
static ofstream *cts_out, *cnt_out, *sl1l2_out, *nbrctc_out, *nbrcts_out;
static ofstream *nbrcnt_out;

static char fho_file[max_str_len],    ctc_file[max_str_len];
static char ctp_file[max_str_len],    cfp_file[max_str_len];
static char cop_file[max_str_len],    cts_file[max_str_len];
static char cnt_file[max_str_len],    sl1l2_file[max_str_len];
static char nbrctc_file[max_str_len], nbrcts_file[max_str_len];
static char nbrcnt_file[max_str_len];

////////////////////////////////////////////////////////////////////////

static void process_command_line(int, char **);
static void process_fcst_obs_files();
static void process_config();
static void process_scores();
static void process_masks();
static void clean_up();

static void setup_first_pass(const WrfData &);
static void setup_txt_files(unixtime, int);
static void setup_nc_file(unixtime, int);

static void apply_mask(const WrfData &, const WrfData &, int,
                       NumArray &, NumArray &);

static void do_cnt_cts(CNTInfo &, CTSInfo *&,
                       int, int, int,
                       const NumArray &, const NumArray &);

static void do_nbrcnt_nbrcts(NBRCNTInfo &, NBRCTSInfo *&,
                             int, int, int, int,
                             const NumArray &, const NumArray &);

static void write_cnt(CNTInfo &, int, const WrfData &);
static void write_sl1l2(CNTInfo &, int, const WrfData &);
static void write_cts(CTSInfo &, int, const WrfData &);
static void write_nbrcnt(NBRCNTInfo &, int, const WrfData &);
static void write_nbrcts(NBRCTSInfo &, int, const WrfData &);
static void write_nc(const WrfData &, const WrfData &, int, int, int);

static void close_txt_files();

static WrfData smooth_field(WrfData *, int, int);

static void usage(int, char **);

////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[]) {

   // Process the command line arguments
   process_command_line(argc, argv);

   // Process the forecast and observation input files
   process_fcst_obs_files();

   // Process the configuration file
   process_config();

   // Compute the scores and write them out
   process_scores();

   // Close the text files and deallocate memory
   clean_up();

   return(0);
}

////////////////////////////////////////////////////////////////////////

void process_command_line(int argc, char **argv) {
   int i;

   sprintf(out_dir, "%s/out/grid_stat", MET_BASE);

   if(argc < 4) {
      usage(argc, argv);
      exit(1);
   }

   // Store the input forecast and observation file names
   strcpy(fcst_file, argv[1]);
   strcpy(obs_file, argv[2]);
   strcpy(config_file, argv[3]);

   // Parse command line arguments
   for(i=0; i<argc; i++) {

      if(strcmp(argv[i], "-outdir") == 0) {
         strcpy(out_dir, argv[i+1]);
         i++;
      }
      else if(strcmp(argv[i], "-v") == 0) {
         verbosity = atoi(argv[i+1]);
         i++;
      }
   }

   // Read the config file
   conf.read(config_file);

   // List the input files
   if(verbosity > 0) {
      cout << "Forecast File: " << fcst_file << "\n"
           << "Observation File: " << obs_file << "\n"
           << "Configuration File: " << config_file << "\n" << flush;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void process_fcst_obs_files() {

   // Attempt to open the forecast file as a NetCDF file.  If it
   // fails, attempt to read it as Grib.
   fcst_nc_file = new NcFile(fcst_file);

   if(!fcst_nc_file || !(fcst_nc_file->is_valid())) {
      fcst_nc_file->close();
      delete fcst_nc_file;
      fcst_nc_file = (NcFile *) 0;
      fcst_fmt_flag = 1;
   }
   else {
      fcst_fmt_flag = 0;
   }

   // Attempt to open the forecast file as a Grib file.
   if(fcst_fmt_flag == 1) {

      if( !(fcst_gb_file.open(fcst_file)) ) {
         cerr << "\n\nERROR: process_fcst_obs_files() -> "
              << "can't open grib file: "
              << fcst_file << "\n\n" << flush;
         exit(1);
      }
   }

   // Attempt to open the observation file as a NetCDF file.  If it
   // fails, attempt to read it as Grib.
   obs_nc_file = new NcFile(obs_file);

   if(!obs_nc_file || !(obs_nc_file->is_valid())) {
      obs_nc_file->close();
      delete obs_nc_file;
      obs_nc_file = (NcFile *) 0;
      obs_fmt_flag = 1;
   }
   else {
      obs_fmt_flag = 0;
   }

   // Attempt to open the observation file as a Grib file.
   if(obs_fmt_flag == 1) {

      if(!(obs_gb_file.open(obs_file)) ) {
         cerr << "\n\nERROR: process_fcst_obs_files() -> "
              << "can't open grib file: "
              << obs_file << "\n\n" << flush;
         exit(1);
      }
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void process_config() {
   int i, j, n, wdth, n_mthd, n_wdth;

   //
   // Conf: model
   //

   if(strlen(conf.model().sval()) == 0 ||
      check_reg_exp(non_ws_reg_exp, conf.model().sval()) != true) {

      cerr << "\n\nERROR: process_config() -> "
           << "The model name must be set to a non-empty "
           << "string value.\n\n" << flush;
      exit(1);
   }

   //
   // Conf: vx_grib_code
   //

   // Parse out the grib codes to be verified
   n_gc = conf.n_vx_grib_code_elements();

   if(n_gc == 0) {
      cerr << "\n\nERROR: process_config() -> "
           << "At least one value must be provided "
           << "for vx_grib_code.\n\n" << flush;
      exit(1);
   }

   // Allocate space to store the grib code information
   gc_info = new GCInfo [n_gc];

   for(i=0; i<n_gc; i++) {
      parse_gcinfo(conf.vx_grib_code(i).sval(), gc_info[i]);

      // No support for WDIR
      if(gc_info[i].code == wdir_grib_code) {
         cerr << "\n\nERROR: process_config() -> "
              << "the wind direction field may not be verified using grid_stat.\n\n"
              << flush;
         exit(1);
      }
   }

   //
   // Conf: thresholds
   //

   // Check that the number of forecast threshold levels matches n_gc
   if(conf.n_thresholds_elements() != n_gc) {
      cerr << "\n\nERROR: process_config() -> "
           << "The number threshold levels provided must match the number "
           << "of grib codes provided in vx_grib_code.\n\n"
           << flush;
      exit(1);
   }

   // Allocate space to store the threshold information
   gc_ti = new ThreshInfo [n_gc];

   // Parse the threshold information
   for(i=0; i<n_gc; i++)
      gc_ti[i].parse_thresh_str(conf.thresholds(i).sval());

   //
   // Conf: ci_alpha
   //

   // Check that at least one alpha value is provided
   if((n_ci_alpha = conf.n_ci_alpha_elements()) == 0) {
      cerr << "\n\nERROR: process_config() -> "
           << "At least one confidence interval alpha value must be "
           << "specified.\n\n" << flush;
      exit(1);
   }

   // Check that the values for alpha are between 0 and 1
   for(i=0; i<n_ci_alpha; i++) {
      if(conf.ci_alpha(i).dval() <= 0.0 || conf.ci_alpha(i).dval() >= 1.0) {
         cerr << "\n\nERROR: process_config() -> "
              << "All confidence interval alpha values ("
              << conf.ci_alpha(i).dval() << ") must be greater than 0 "
              << "and less than 1.\n\n" << flush;
         exit(1);
      }
   }

   //
   // Conf: boot_interval
   //

   // Check that boot_interval is set to 0 or 1
   if(conf.boot_interval().ival() != boot_bca_flag &&
      conf.boot_interval().ival() != boot_perc_flag) {
      cerr << "\n\nERROR: process_config() -> "
           << "The boot_interval parameter must be set to "
           << boot_bca_flag << " or "
           << boot_perc_flag << "!\n\n" << flush;
      exit(1);
   }

   //
   // Conf: boot_rep_prop
   //

   // Check that boot_rep_prop is set between 0 and 1
   if(conf.boot_rep_prop().dval() <= 0.0 ||
      conf.boot_rep_prop().dval() > 1.0) {
      cerr << "\n\nERROR: process_config() -> "
           << "The boot_rep_prop parameter must be set between "
           << "0 and 1!\n\n" << flush;
      exit(1);
   }

   //
   // Conf: n_boot_rep
   //

   // Check that n_boot_rep is set > 0
   if(conf.n_boot_rep().dval() < 0.0) {
      cerr << "\n\nERROR: process_config() -> "
           << "The number of bootstrap resamples (n_boot_rep) "
           << "must be set to a value >= 0.\n\n" << flush;
      exit(1);
   }

   //
   // Conf: boot_rng and boot_seed
   //

   // Set the random number generator and seed value to be used when
   // computing bootstrap confidence intervals
   rng_set(rng_ptr, conf.boot_rng().sval(), conf.boot_seed().sval());

   //
   // Conf: interp_method
   //

   // Check that at least one interpolation method is provided
   if((n_mthd = conf.n_interp_method_elements()) <= 0) {
      cerr << "\n\nERROR: process_config() -> "
           << "At least one interpolation method must be provided.\n\n"
           << flush;
      exit(1);
   }

   // Check that distance-weighted mean and least squares fit are not selected
   for(i=0; i<n_mthd; i++) {
      if(get_interp_mthd(conf.interp_method(i).sval()) == i_dw_mean ||
         get_interp_mthd(conf.interp_method(i).sval()) == i_ls_fit) {
         cerr << "\n\nERROR: process_config() -> "
              << "The interpolation method may not be set to DW_MEAN "
              << "or LS_FIT for grid_stat.\n\n" << flush;
         exit(1);
      }
   }

   //
   // Conf: interp_width
   //

   // Check that at least one interpolation width is provided
   if((n_wdth = conf.n_interp_width_elements()) <= 0) {
      cerr << "\n\nERROR: process_config() -> "
           << "At least one interpolation width must be provided.\n\n"
           << flush;
      exit(1);
   }

   // Do error checking and compute the total number of
   // interpolations to be performed
   n_interp = 0;
   for(i=0; i<n_wdth; i++) {

      if(conf.interp_width(i).ival() < 1 ||
         conf.interp_width(i).ival()%2 == 0) {
         cerr << "\n\nERROR: process_config() -> "
              << "The interpolation width must be set "
              << "to odd values greater than or equal to 1 ("
              << conf.interp_width(i).ival() << ").\n\n" << flush;
         exit(1);
      }

      wdth = conf.interp_width(i).ival();
      if(wdth == 1) n_interp += 1;
      if(wdth >  1) n_interp += n_mthd;
   }

   // Allocate space for the interpolation methods and widths
   interp_mthd = new int [n_interp];
   interp_wdth = new int [n_interp];

   // Set each interpolation method and width
   n = 0;
   for(i=0; i<n_wdth; i++) {

      interp_wdth[n] = conf.interp_width(i).ival();

      // For an interpolation width of 1, set the method the method to
      // unweighted mean - which is really just nearest neighbor
      if(interp_wdth[n] == 1) {
         interp_mthd[n] = i_uw_mean;
         n++;
         continue;
      }

      for(j=0; j<n_mthd; j++) {
         interp_mthd[n] = get_interp_mthd(conf.interp_method(j).sval());
         n++;
      }
   }

   //
   // Conf: interp_threshold
   //

   // Check that the interpolation threshold is set between
   // 0 and 1.
   if(conf.interp_threshold().dval() < 0.0
   || conf.interp_threshold().dval() > 1.0) {
         cerr << "\n\nERROR: process_config() -> "
              << "The interpolation threshold value must be set "
              << "between 0 and 1.\n\n" << flush;
         exit(1);
   }

   //
   // Conf: nbr_width
   //

   // Check that at least one neighborhood width is provided
   if((n_nbr_wdth = conf.n_nbr_width_elements()) <= 0) {
      cerr << "\n\nERROR: process_config() -> "
           << "At least one neighborhood width must be provided.\n\n"
           << flush;
      exit(1);
   }

   // Do error checking of neighborhood widths
   for(i=0; i<n_nbr_wdth; i++) {

      if(conf.nbr_width(i).ival() < 1 ||
         conf.nbr_width(i).ival()%2 == 0) {
         cerr << "\n\nERROR: process_config() -> "
              << "The neighborhood width must be set "
              << "to odd values greater than or equal to 1 ("
              << conf.nbr_width(i).ival() << ").\n\n" << flush;
         exit(1);
      }
   }

   //
   // Conf: nbr_threshold
   //

   // Check that the neighborhood threshold is set between 0 and 1.
   if(conf.nbr_threshold().dval() < 0.0
   || conf.nbr_threshold().dval() > 1.0) {
         cerr << "\n\nERROR: process_config() -> "
              << "The neighborhood threshold value must be set "
              << "between 0 and 1.\n\n" << flush;
         exit(1);
   }

   //
   // Conf: nbr_frac_threshold
   //

   // Check that at least one neighborhood fractional threshold is provided
   if(conf.n_nbr_frac_threshold_elements() <= 0) {
      cerr << "\n\nERROR: process_config() -> "
           << "At least one neighborhood fraction threshold value "
           << "must be provided.\n\n"
           << flush;
      exit(1);
   }

   // Set the number of thresholds
   frac_ti.set_n_thresh(conf.n_nbr_frac_threshold_elements());

   // Do error checking of neighborhood fractional thresholds
   for(i=0; i<conf.n_nbr_frac_threshold_elements(); i++) {

      // Parse the threshold information
      frac_ti.set_thresh(i, conf.nbr_frac_threshold(i).sval());

      if(frac_ti.thresh[i].get_thresh() < 0.0 ||
         frac_ti.thresh[i].get_thresh() > 1.0) {
         cerr << "\n\nERROR: process_config() -> "
              << "The neighborhood fraction threshold value must be set "
              << "between 0 and 1.\n\n" << flush;
         exit(1);
      }
   }

   //
   // Conf: output_flag
   //

   // Check that at least one output VSDB type is requested
   for(i=i_fho, n=0; i<i_nbrcnt; i++) {
      n += (conf.output_flag(i).ival() > 0.0);
   }

   if(n == 0) {
      cerr << "\n\nERROR: process_config() -> "
           << "At least one output VSDB type must be requested.\n\n"
           << flush;
      exit(1);
   }

   //
   // Conf: rank_corr_flag
   //
   if(conf.rank_corr_flag().ival() != 0 &&
      conf.rank_corr_flag().ival() != 1) {
      cerr << "\n\nERROR: process_config() -> "
           << "The rank_corr_flag (" << conf.rank_corr_flag().ival()
           << ") must be set to 0 or 1.\n\n"
           << flush;
      exit(1);
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void process_scores() {
   int i, j, k, m, n, h, status;
   WrfData fcst_wd, obs_wd, fcst_wd_smooth, obs_wd_smooth;
   GribRecord fcst_r, obs_r;
   char tmp_str[max_str_len];
   NumArray f_na, o_na;
   CNTInfo     cnt_info;
   CTSInfo    *cts_info;
   NBRCNTInfo  nbrcnt_info;
   NBRCTSInfo *nbrcts_info;


   // Loop through each of the fields to be verified
   for(i=0; i<n_gc; i++) {

      // Get the grib code abbreviation string
      get_grib_code_abbr(gc_info[i].code, conf.ncep_defaults().ival(),
                         abbr_str);

      // Process the forecast file in NetCDF format (output of pcp_combine)
      if(fcst_fmt_flag == 0) {
         read_pcp_combine_netcdf(fcst_nc_file, abbr_str, level_name,
                                 fcst_wd, fcst_grid, verbosity,
                                 pc_data, lc_data, st_data);
      }
      // Process the forecast file in Grib format
      else {

         status = get_grib_record(fcst_gb_file, fcst_r,
                                  gc_info[i].code, gc_info[i].lvl_1, gc_info[i].lvl_2,
                                  fcst_wd, fcst_grid, verbosity,
                                  pc_data, lc_data, st_data);

         if(status != 0) {
            cout << "***WARNING***: process_scores() -> "
                 << "grib code " << gc_info[i].code
                 << " with accumulation/level indicator of "
                 << gc_info[i].lvl_str << " not found in grib file: "
                 << fcst_file << "\n" << flush;
            continue;
         }

         // Get the level name
         get_grib_level_str(fcst_r.pds->type, conf.ncep_defaults().ival(),
                            fcst_r.pds->level_info, level_name);
      }

      // Process the observation file in NetCDF format (output of pcp_combine)
      if(obs_fmt_flag == 0) {
         read_pcp_combine_netcdf(obs_nc_file, abbr_str, tmp_str,
                                 obs_wd, obs_grid, verbosity,
                                 pc_data, lc_data, st_data);
      }
      // Process the observation file in Grib format
      else {

         status = get_grib_record(obs_gb_file, obs_r,
                                  gc_info[i].code, gc_info[i].lvl_1, gc_info[i].lvl_2,
                                  obs_wd, obs_grid, verbosity,
                                  pc_data, lc_data, st_data);

         if(status != 0) {
            cout << "***WARNING***: process_scores() -> "
                 << "grib code " << gc_info[i].code
                 << " with accumulation/level indicator of "
                 << gc_info[i].lvl_str << " not found in grib file: "
                 << obs_file << "\n" << flush;
            continue;
         }
      }

      // Check that the grid dimensions match
      if(fcst_grid.nx() != obs_grid.nx() ||
         fcst_grid.ny() != obs_grid.ny()) {
         cerr << "\n\nERROR: process_scores() -> "
              << "Forecast and observation grid dimensions "
              << "do not match (" << fcst_grid.nx() << ", "
              << fcst_grid.ny() << ") != (" << obs_grid.nx()
              << ", " << obs_grid.ny() << ") for grib code "
              << gc_info[i].code << ".\n\n" << flush;
         exit(1);
      }

      // Check that the valid times match
      if(fcst_wd.get_valid_time() != obs_wd.get_valid_time()) {
         cout << "***WARNING***: process_scores() -> "
              << "Forecast and observation valid times do not match "
              << fcst_wd.get_valid_time() << " != "
              << obs_wd.get_valid_time() << " for grib code "
              << gc_info[i].code << ".\n\n" << flush;
      }

      // Check that the accumulation intervals match
      if(fcst_wd.get_accum_time() != obs_wd.get_accum_time()) {
         cout << "***WARNING***: process_scores() -> "
              << "Forecast and observation accumulation times do not match "
              << fcst_wd.get_accum_time() << " != "
              << obs_wd.get_accum_time() << " for grib code "
              << gc_info[i].code << ".\n\n" << flush;
      }

      // This is the first pass through the loop and grid is unset
      if(grid.nx() == 0 && grid.ny() == 0) {

         // Setup the first pass through the data
         setup_first_pass(fcst_wd);
      }
      // For multiple verification fields, check to make sure that the
      // grid dimensions don't change
      else {

         if(fcst_grid.nx() != grid.nx() ||
            fcst_grid.ny() != grid.ny() ||
            obs_grid.nx()  != grid.nx() ||
            obs_grid.ny()  != grid.ny()) {
            cerr << "\n\nERROR: process_scores() -> "
                 << "The grid dimensions must remain constant ("
                 << grid.nx() << ", " << grid.ny() << ") != ("
                 << fcst_grid.nx() << ", " << fcst_grid.ny() << ") or ("
                 << obs_grid.nx() << ", " << obs_grid.ny() << ")\n\n"
                 << flush;
            exit(1);
         }
      }

      // Setup strings for the grib code
      if(gc_info[i].code == apcp_grib_code) strcpy(msg_type, "MC_PCP");
      else                                  strcpy(msg_type, "ANALYS");

      if(nint(fcst_wd.get_accum_time()/sec_per_hour) > 0) {
         h = nint(fcst_wd.get_accum_time()/sec_per_hour);
         sprintf(field_name, "%s/%.2i", abbr_str, h);
      }
      else {
         strcpy(field_name, abbr_str);
      }

      if(verbosity > 1 && gc_info[i].lvl_str != (char *) 0) {
         cout << "\n----------------------------------------\n\n"
              << "Processing Grib Code " << gc_info[i].code
              << " with accumulation/level indicator of " << gc_info[i].lvl_str
              << "\n" << flush;
      }
      else if(verbosity > 1) {
         cout << "\n----------------------------------------\n\n"
              << "Processing Grib Code " << gc_info[i].code
              << " with no accumulation/level indicator\n" << flush;
      }

      // Loop through and apply each of the smoothing operations
      for(j=0; j<n_interp; j++) {

         // Smooth the forecast and observation fields based on the
         // interp_mthd and interp_wdth
         fcst_wd_smooth = smooth_field(&fcst_wd, interp_mthd[j], interp_wdth[j]);
         obs_wd_smooth  = smooth_field(&obs_wd,  interp_mthd[j], interp_wdth[j]);

         // Loop through the masks to be applied
         for(k=0; k<n_masks; k++) {

            // Apply the current mask to the fields
            apply_mask(fcst_wd_smooth, obs_wd_smooth, k, f_na, o_na);

            // Compute CNT and CTS scores for each threshold specified
            if(conf.output_flag(i_fho).ival() || conf.output_flag(i_ctc).ival() ||
               conf.output_flag(i_ctp).ival() || conf.output_flag(i_cfp).ival() ||
               conf.output_flag(i_cop).ival() || conf.output_flag(i_cts).ival() ||
               conf.output_flag(i_cnt).ival() || conf.output_flag(i_sl1l2).ival()) {

               // Allocate memory for CTSInfo objects
               cts_info = new CTSInfo [gc_ti[i].n_thresh];

               // Initialize
               cnt_info.clear();

               do_cnt_cts(cnt_info, cts_info, i, j, k, f_na, o_na);

               // Write out the CNT statistics
               if(conf.output_flag(i_cnt).ival()) {
                  write_cnt(cnt_info, i, fcst_wd_smooth);
               } // end write CNT

               // Write out the SL1L2 statistics
               if(conf.output_flag(i_sl1l2).ival()) {
                  write_sl1l2(cnt_info, i, fcst_wd_smooth);
               } // end write SL1L2

               // Write out the CTS statistics
               if(conf.output_flag(i_fho).ival() || conf.output_flag(i_ctc).ival() ||
                  conf.output_flag(i_ctp).ival() || conf.output_flag(i_cfp).ival() ||
                  conf.output_flag(i_cop).ival() || conf.output_flag(i_cts).ival()) {

                  for(m=0; m<gc_ti[i].n_thresh; m++) {
                     write_cts(cts_info[m], i, fcst_wd_smooth);
                  }
               } // end write CTS

               // Deallocate memory for CTSInfo objects
               if(cts_info) { delete [] cts_info; cts_info = (CTSInfo *) 0; }

            } // end compute CNT and CTS
         } // end for k

         // Write out the smoothed forecast, observation, and difference
         // fields in netCDF format for each grib code if requested in
         // the config file
         if(conf.output_flag(i_nc).ival())
            write_nc(fcst_wd_smooth, obs_wd_smooth, i,
                     interp_mthd[j], interp_wdth[j]);

      } // end for j

      // Loop through and apply the Neighborhood methods for each of the
      // neighborhood widths if requested in the config file
      if(conf.output_flag(i_nbrcts).ival() ||
         conf.output_flag(i_nbrcnt).ival()) {

         // Allocate memory for NBRCTSInfo objects
         nbrcts_info = new NBRCTSInfo [frac_ti.n_thresh];

         // Loop through and apply each of the neighborhood widths
         for(j=0; j<n_nbr_wdth; j++) {

            // Loop through and apply each of the raw threshold values
            for(k=0; k<gc_ti[i].n_thresh; k++) {

               // Compute the fractional coverage based on the threshold
               fcst_wd_smooth = fractional_coverage(fcst_wd, conf.nbr_width(j).ival(),
                                                    gc_ti[i].thresh[k],
                                                    conf.nbr_threshold().dval());

               obs_wd_smooth  = fractional_coverage(obs_wd,conf.nbr_width(j).ival(),
                                                    gc_ti[i].thresh[k],
                                                    conf.nbr_threshold().dval());

               // Loop through the masks to be applied
               for(m=0; m<n_masks; m++) {

                  // Apply the current mask to the fields
                  apply_mask(fcst_wd_smooth, obs_wd_smooth, m, f_na, o_na);

                  // Initialize
                  nbrcnt_info.clear();
                  for(n=0; n<frac_ti.n_thresh; n++) nbrcts_info[n].clear();

                  do_nbrcnt_nbrcts(nbrcnt_info, nbrcts_info, i, j, k, m, f_na, o_na);

                  // Write out the NBRCNT statistics
                  if(conf.output_flag(i_nbrcnt).ival()) {
                     write_nbrcnt(nbrcnt_info, i, fcst_wd_smooth);
                  } // end write NBRCNT

                  // Write out the CTS statistics
                  if(conf.output_flag(i_nbrcts).ival()) {

                     for(n=0; n<frac_ti.n_thresh; n++) {
                        write_nbrcts(nbrcts_info[n], i, fcst_wd_smooth);
                     }
                  } // end write NBRCTS

               } // end for m
            } // end for k
         } // end for j

         // Deallocate memory for NBRCTSInfo objects
         if(nbrcts_info) { delete [] nbrcts_info; nbrcts_info = (NBRCTSInfo *) 0; }

      } // end if
   } // end for i

   if(verbosity > 1) {
      cout << "\n----------------------------------------\n\n" << flush;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void process_masks() {
   int i, n_mask_grids, n_mask_polys;

   // Get the number of masks
   n_mask_grids = conf.n_mask_grids_elements();
   n_mask_polys = conf.n_mask_polys_elements();
   n_masks = n_mask_grids + n_mask_polys;

   // Check that at least one verification masking region is provided
   if(n_masks == 0) {
      cerr << "\n\nERROR: process_masks() -> "
           << "At least one grid or polyline verification masking "
           << "region must be provided.\n\n" << flush;
      exit(1);
   }

   // Allocate space to store the masking information
   mask_wd    = new WrfData [n_masks];
   mask_names = new char *[n_masks];

   // Parse out the masking grids
   for(i=0; i<n_mask_grids; i++)
      parse_grid_mask(conf.mask_grids(i).sval(), grid,
                      mask_wd[i], mask_names[i]);

   // Parse out the masking polys
   for(i=0; i<n_mask_polys; i++)
      parse_poly_mask(conf.mask_polys(i).sval(), grid,
                      mask_wd[i+n_mask_grids], mask_names[i+n_mask_grids]);

   return;
}

////////////////////////////////////////////////////////////////////////

void clean_up() {
   int i;

   // Close the output text files that were open for writing
   close_txt_files();

   // Close the output NetCDF file as long as it was opened
   if(nc_out && conf.output_flag(i_nc).ival()) {

      // List the NetCDF file after it is finished
      if(verbosity > 0) {
         cout << "Output NetCDF file: "
              << out_nc_file << "\n" << flush;
      }
      nc_out->close();
      delete nc_out;
      nc_out = (NcFile *) 0;
   }

   // Close the forecast file
   if(fcst_fmt_flag == 0) fcst_nc_file->close();
   else                   fcst_gb_file.close();

   // Close the observation file
   if(obs_fmt_flag == 0)  obs_nc_file->close();
   else                   obs_gb_file.close();

   // Clean up allocated memory
   if(gc_info)          { delete [] gc_info;       gc_info = (GCInfo *) 0; }
   if(gc_ti)            { delete [] gc_ti;         gc_ti = (ThreshInfo *) 0; }
   if(interp_mthd)      { delete [] interp_mthd;   interp_mthd = (int *) 0; }
   if(interp_wdth)      { delete [] interp_wdth;   interp_wdth = (int *) 0; }
   if(mask_wd)          { delete [] mask_wd;       mask_wd = (WrfData *) 0; }

   for(i=0; i<n_masks; i++) {
      if(mask_names[i]) { delete [] mask_names[i]; mask_names[i] = (char *) 0; }
   }
   if(mask_names)       { delete [] mask_names;    mask_names = (char **) 0; }

   // Deallocate memory for the random number generator
   rng_free(rng_ptr);

   return;
}

////////////////////////////////////////////////////////////////////////

void setup_first_pass(const WrfData &wd) {
   int j, max_n_thresh;

   // Store the grid to be used through the verification
   grid = fcst_grid;

   // Process masks Grids and Polylines in the config file
   process_masks();

   // Figure out the maximum number of thresholds being applied
   // across Grib Codes
   max_n_thresh = 0;
   for(j=0; j<n_gc; j++) {
      if(gc_ti[j].n_thresh > max_n_thresh)
         max_n_thresh = gc_ti[j].n_thresh;
   }

   // Create output text files as requested in the config file
   setup_txt_files(wd.get_valid_time(), wd.get_lead_time());

   // If requested, create a NetCDF file to store the matched pairs and
   // difference fields for each grib code and masking region
   if(conf.output_flag(i_nc).ival()) {
      setup_nc_file(wd.get_valid_time(), wd.get_lead_time());
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void setup_txt_files(unixtime valid_ut, int lead_sec) {
   int mon, day, yr, hr, min, sec;
   int l_hr, l_min, l_sec;
   char header[max_line_len];

   // Initialize output file streams
   ctc_out   = ctp_out    = cfp_out    = cop_out    = (ofstream *) 0;
   vsdb_out  = fho_out    = cts_out    = cnt_out    = (ofstream *) 0;
   sl1l2_out = nbrctc_out = nbrcts_out = nbrcnt_out = (ofstream *) 0;

   // Create output VSDB file names for the VSDB and optional text files
   sec_to_hms(lead_sec, l_hr, l_min, l_sec);
   unix_to_mdyhms(valid_ut, mon, day, yr, hr, min, sec);
   sprintf(vsdb_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV.vsdb",
          out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(fho_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_fho.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(ctc_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_ctc.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(ctp_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_ctp.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(cfp_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_cfp.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(cop_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_cop.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(cts_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_cts.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(cnt_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_cnt.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(sl1l2_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_sl1l2.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(nbrctc_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_nbrctc.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(nbrcts_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_nbrcts.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);
   sprintf(nbrcnt_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_nbrcnt.txt",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);

   // Create and open the VSDB output file stream
   vsdb_out = new ofstream;
   vsdb_out->open(vsdb_file);

   if(!vsdb_out) {
      cerr << "\n\nERROR: setup_txt_files() -> "
           << "unable to open output VSDB file \""
           << vsdb_file << "\"\n\n";
      exit(1);
   }
   vsdb_out->setf(ios::fixed);

   // List VSDB file as it is being created
   if(verbosity > 1) {
      cout << "Creating Output VSDB file: "
           << vsdb_file << "\n" << flush;
   }

   // Create the FHO text file
   if(conf.output_flag(i_fho).ival() >= 2) {

      // Create and open the FHO output file stream
      fho_out = new ofstream;
      fho_out->open(fho_file);

      if(!fho_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output FHO file \""
              << fho_file << "\"\n\n";
         exit(1);
      }
      fho_out->setf(ios::fixed);

      // List FHO file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output FHO file: "
              << fho_file << "\n" << flush;
      }

      get_line_names(fho_columns, n_fho_columns, header);
      *fho_out << header << "\n" << flush;
   }

   // Create the CTC text file
   if(conf.output_flag(i_ctc).ival() >= 2) {

      // Create and open the CTC output file stream
      ctc_out = new ofstream;
      ctc_out->open(ctc_file);

      if(!ctc_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output CTC file \""
              << ctc_file << "\"\n\n";
         exit(1);
      }
      ctc_out->setf(ios::fixed);

      // List CTC file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output CTC file: "
              << ctc_file << "\n" << flush;
      }

      get_line_names(ctc_columns, n_ctc_columns, header);
      *ctc_out << header << "\n" << flush;
   }

   // Create the CTP text file
   if(conf.output_flag(i_ctp).ival() >= 2) {

      // Create and open the CTP output file stream
      ctp_out = new ofstream;
      ctp_out->open(ctp_file);

      if(!ctp_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output CTP file \""
              << ctp_file << "\"\n\n";
         exit(1);
      }
      ctp_out->setf(ios::fixed);

      // List CTP file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output CTP file: "
              << ctp_file << "\n" << flush;
      }

      get_line_names(ctp_columns, n_ctp_columns, header);
      *ctp_out << header << "\n" << flush;
   }

   // Create the CFP text file
   if(conf.output_flag(i_cfp).ival() >= 2) {

      // Create and open the CFP output file stream
      cfp_out = new ofstream;
      cfp_out->open(cfp_file);

      if(!cfp_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output CFP file \""
              << cfp_file << "\"\n\n";
         exit(1);
      }
      cfp_out->setf(ios::fixed);

      // List CFP file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output CFP file: "
              << cfp_file << "\n" << flush;
      }

      get_line_names(cfp_columns, n_cfp_columns, header);
      *cfp_out << header << "\n" << flush;
   }

   // Create the COP text file
   if(conf.output_flag(i_cop).ival() >= 2) {

      // Create and open the COP output file stream
      cop_out = new ofstream;
      cop_out->open(cop_file);

      if(!cop_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output COP file \""
              << cop_file << "\"\n\n";
         exit(1);
      }
      cop_out->setf(ios::fixed);

      // List COP file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output COP file: "
              << cop_file << "\n" << flush;
      }

      get_line_names(cop_columns, n_cop_columns, header);
      *cop_out << header << "\n" << flush;
   }

   // Create the CTS text file
   if(conf.output_flag(i_cts).ival() >= 2) {

      // Create and open the CTS output file stream
      cts_out = new ofstream;
      cts_out->open(cts_file);

      if(!cts_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output CTS file \""
              << cts_file << "\"\n\n";
         exit(1);
      }
      cts_out->setf(ios::fixed);

      // List CTS file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output CTS file: "
              << cts_file << "\n" << flush;
      }

      get_line_names(cts_columns, n_cts_columns, header);
      *cts_out << header << "\n" << flush;
   }

   // Create the CNT text file
   if(conf.output_flag(i_cnt).ival() >= 2) {

      // Create and open the CNT output file stream
      cnt_out = new ofstream;
      cnt_out->open(cnt_file);

      if(!cnt_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output CNT file \""
              << cnt_file << "\"\n\n";
         exit(1);
      }
      cnt_out->setf(ios::fixed);

      // List CNT file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output CNT file: "
              << cnt_file << "\n" << flush;
      }

      get_line_names(cnt_columns, n_cnt_columns, header);
      *cnt_out << header << "\n" << flush;
   }

   // Create the SL1L2 text file
   if(conf.output_flag(i_sl1l2).ival() >= 2) {

      // Create and open the SL1L2 output file stream
      sl1l2_out = new ofstream;
      sl1l2_out->open(sl1l2_file);

      if(!sl1l2_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output SL1L2 file \""
              << sl1l2_file << "\"\n\n";
         exit(1);
      }
      sl1l2_out->setf(ios::fixed);

      // List SL1L2 file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output SL1L2 file: "
              << sl1l2_file << "\n" << flush;
      }

      get_line_names(sl1l2_columns, n_sl1l2_columns, header);
      *sl1l2_out << header << "\n" << flush;
   }

   // Create the Neighborhood Methods Contingency Table Counts
   // text file
   if(conf.output_flag(i_nbrctc).ival() >= 2) {

      // Create and open the NBRCTC output file stream
      nbrctc_out = new ofstream;
      nbrctc_out->open(nbrctc_file);

      if(!nbrctc_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output NBRCTC file \""
              << nbrctc_file << "\"\n\n";
         exit(1);
      }
      nbrctc_out->setf(ios::fixed);

      // List NBRCNT file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output NBRCTC file: "
              << nbrctc_file << "\n" << flush;
      }

      get_line_names(nbrctc_columns, n_nbrctc_columns, header);
      *nbrctc_out << header << "\n" << flush;
   }

   // Create the Neighborhood Methods Contingency Table Statistics
   // text file
   if(conf.output_flag(i_nbrcts).ival() >= 2) {

      // Create and open the NBRCTS output file stream
      nbrcts_out = new ofstream;
      nbrcts_out->open(nbrcts_file);

      if(!nbrcts_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output NBRCTS file \""
              << nbrcts_file << "\"\n\n";
         exit(1);
      }
      nbrcts_out->setf(ios::fixed);

      // List NBRCNT file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output NBRCTS file: "
              << nbrcts_file << "\n" << flush;
      }

      get_line_names(nbrcts_columns, n_nbrcts_columns, header);
      *nbrcts_out << header << "\n" << flush;
   }

   // Create the Continuous Neighborhood Methods text file
   if(conf.output_flag(i_nbrcnt).ival() >= 2) {

      // Create and open the NBRCNT output file stream
      nbrcnt_out = new ofstream;
      nbrcnt_out->open(nbrcnt_file);

      if(!nbrcnt_out) {
         cerr << "\n\nERROR: setup_txt_files() -> "
              << "unable to open output NBRCNT file \""
              << nbrcnt_file << "\"\n\n";
         exit(1);
      }
      nbrcnt_out->setf(ios::fixed);

      // List NBRCNT file as it is being created
      if(verbosity > 1) {
         cout << "Creating Output NBRCNT file: "
              << nbrcnt_file << "\n" << flush;
      }

      get_line_names(nbrcnt_columns, n_nbrcnt_columns, header);
      *nbrcnt_out << header << "\n" << flush;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void setup_nc_file(unixtime valid_ut, int lead_sec) {
   int yr, mon, day, hr, min, sec;
   int l_hr, l_min, l_sec;
   int x, y;
   char attribute_str[max_str_len], time_str[max_str_len];
   char hostname_str[max_str_len];
   double lat, lon;
   float *lat_data, *lon_data;
   unixtime ut;

   // Create output NetCDF file name
   sec_to_hms(lead_sec, l_hr, l_min, l_sec);
   unix_to_mdyhms(valid_ut, mon, day, yr, hr, min, sec);
   sprintf(out_nc_file, "%s/grid_stat_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_pairs.nc",
           out_dir, l_hr, l_min, l_sec, yr, mon, day, hr, min, sec);

   // Create a new NetCDF file and open it
   nc_out = new NcFile(out_nc_file, NcFile::Replace);

   if(!nc_out || !nc_out->is_valid()) {
      cerr << "\n\nERROR: setup_nc_file() -> "
           << "trouble opening output NetCDF file "
           << out_nc_file << "\n\n" << flush;
      exit(1);
   }

   // List NetCDF file as it is being created
   if(verbosity > 1) {
      cout << "Creating Output NetCDF file: "
           << out_nc_file << "\n" << flush;
   }

   // Add global attributes
   nc_out->add_att("Conventions", "COARDS");

   ut = time(NULL);
   unix_to_mdyhms(ut, mon, day, yr, hr, min, sec);
   sprintf(time_str, "%.4i-%.2i-%.2i %.2i:%.2i:%.2i", yr, mon, day, hr, min, sec);

   gethostname(hostname_str, max_str_len);

   sprintf(attribute_str, "File %s generated %s UTC on host %s",
           out_nc_file, time_str, hostname_str);
   nc_out->add_att("FileOrigins", attribute_str);
   nc_out->add_att("Projection", grid.name());
   nc_out->add_att("Difference", "Forecast Value - Observation Value");

   // Define Dimensions
   lat_dim = nc_out->add_dim("lat", (long) grid.ny());
   lon_dim = nc_out->add_dim("lon", (long) grid.nx());

   // Define Variables
   lat_var = nc_out->add_var("lat", ncFloat, lat_dim, lon_dim);
   lon_var = nc_out->add_var("lon", ncFloat, lat_dim, lon_dim);

   // Add variable attributes
   lat_var->add_att("units", "degrees_north");
   lat_var->add_att("long_name", "Latitude");
   lat_var->add_att("_FillValue", bad_data_float);

   lon_var->add_att("units", "degrees_east");
   lon_var->add_att("long_name", "Longitude");
   lon_var->add_att("_FillValue", bad_data_float);

   // Allocate memory to store the Lat/Lon values for each grid point
   lat_data = lon_data = (float *) 0;
   lat_data = new float [grid.nx()*grid.ny()];
   lon_data = new float [grid.nx()*grid.ny()];

   for(x=0; x<grid.nx(); x++) {
      for(y=0; y<grid.ny(); y++) {

         // Lat/Lon values for each grid point
         grid.xy_to_latlon(x, y, lat, lon);

         lat_data[two_to_one(grid.nx(), grid.ny(), x, y)] = (float) lat;
         lon_data[two_to_one(grid.nx(), grid.ny(), x, y)] = (float) -1.0*lon;
      }
   }
   if( !lat_var->put(&lat_data[0], grid.ny(), grid.nx())
    || !lon_var->put(&lon_data[0], grid.ny(), grid.nx()) ) {

      cerr << "\n\nERROR: setup_nc_file() -> "
           << "error with the lat_var->put or lon_var->put\n\n" << flush;
      exit(1);
   }

   // Dealloate memory
   delete [] lat_data;
   lat_data = (float *) 0;
   delete [] lon_data;
   lon_data = (float *) 0;

   return;
}

////////////////////////////////////////////////////////////////////////

void write_cnt(CNTInfo &cnt_info, int i_gc, const WrfData &fcst_wd) {
   char hdr_str[max_line_len], col_str[max_line_len];
   char type_str[max_str_len];
   char line[max_line_len];
   int i;

   //
   // Grib Field abbreviation
   //
   get_grib_code_abbr(gc_info[i_gc].code, conf.ncep_defaults().ival(),
                      abbr_str);

   //
   // CNT line type
   //
   if(conf.output_flag(i_cnt).ival() != 0 &&
      cnt_info.n > 0) {

      for(i=0; i<n_ci_alpha; i++) {

         sprintf(type_str, "CNT/%.3f", cnt_info.alpha[i]);

         build_header_str(hdr_str,
            conf.model().sval(),
            fcst_wd.get_lead_time(),
            fcst_wd.get_valid_time(),
            msg_type,
            cnt_info.name,
            type_str,
            field_name,
            level_name);

         build_cnt_str(col_str, cnt_info, i);

         sprintf(line, "%s %s", hdr_str, col_str);

         *vsdb_out << line << "\n";

         if(conf.output_flag(i_cnt).ival() >= 2) *cnt_out << line << "\n";
      } // end for i
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void write_sl1l2(CNTInfo &cnt_info, int i_gc, const WrfData &fcst_wd) {
   char hdr_str[max_line_len], col_str[max_line_len];
   char type_str[max_str_len];
   char line[max_line_len];

   //
   // Grib Field abbreviation
   //
   get_grib_code_abbr(gc_info[i_gc].code, conf.ncep_defaults().ival(),
                      abbr_str);

   //
   // SL1L2 line type
   //
   if(conf.output_flag(i_sl1l2).ival() != 0 &&
      cnt_info.n > 0) {

      strcpy(type_str, "SL1L2");

      build_header_str(hdr_str,
         conf.model().sval(),
         fcst_wd.get_lead_time(),
         fcst_wd.get_valid_time(),
         msg_type,
         cnt_info.name,
         type_str,
         field_name,
         level_name);

      build_sl1l2_str(col_str, cnt_info);

      sprintf(line, "%s %s", hdr_str, col_str);

      *vsdb_out << line << "\n";

      if(conf.output_flag(i_sl1l2).ival() >= 2) *sl1l2_out << line << "\n";
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void write_cts(CTSInfo &cts_info, int i_gc, const WrfData &fcst_wd) {
   char hdr_str[max_line_len], col_str[max_line_len];
   char thresh_str[max_str_len], type_str[max_str_len];
   char line[max_line_len];
   int i;

   //
   // Type of thresholding performed
   //
   cts_info.cts_thresh.get_str(thresh_str);

   //
   // Grib Field abbreviation
   //
   get_grib_code_abbr(gc_info[i_gc].code, conf.ncep_defaults().ival(),
                      abbr_str);

   //
   // FHO line type
   //
   if(conf.output_flag(i_fho).ival() != 0 &&
      cts_info.cts.n() > 0   ) {

      sprintf(type_str, "FHO%s", thresh_str);

      build_header_str(hdr_str,
         conf.model().sval(),
         fcst_wd.get_lead_time(),
         fcst_wd.get_valid_time(),
         msg_type,
         cts_info.name,
         type_str,
         field_name,
         level_name);

      build_fho_str(col_str, cts_info);

      sprintf(line, "%s %s", hdr_str, col_str);

      *vsdb_out << line << "\n";

      if(conf.output_flag(i_fho).ival() >= 2) *fho_out << line << "\n";
   }

   //
   // CTC line type
   //
   if(conf.output_flag(i_ctc).ival() != 0 &&
      cts_info.cts.n() > 0) {

      sprintf(type_str, "CTC%s", thresh_str);

      build_header_str(hdr_str,
         conf.model().sval(),
         fcst_wd.get_lead_time(),
         fcst_wd.get_valid_time(),
         msg_type,
         cts_info.name,
         type_str,
         field_name,
         level_name);

      build_ctc_str(col_str, cts_info);

      sprintf(line, "%s %s", hdr_str, col_str);

      *vsdb_out << line << "\n";

      if(conf.output_flag(i_ctc).ival() >= 2) *ctc_out << line << "\n";
   }

   //
   // CTP line type
   //
   if(conf.output_flag(i_ctp).ival() != 0 &&
      cts_info.cts.n() > 0) {

      sprintf(type_str, "CTP%s", thresh_str);

      build_header_str(hdr_str,
         conf.model().sval(),
         fcst_wd.get_lead_time(),
         fcst_wd.get_valid_time(),
         msg_type,
         cts_info.name,
         type_str,
         field_name,
         level_name);

      build_ctp_str(col_str, cts_info);

      sprintf(line, "%s %s", hdr_str, col_str);

      *vsdb_out << line << "\n";

      if(conf.output_flag(i_ctp).ival() >= 2) *ctp_out << line << "\n";
   }

   //
   // CFP line type
   //
   if(conf.output_flag(i_cfp).ival() != 0 &&
      cts_info.cts.n() > 0) {

      sprintf(type_str, "CFP%s", thresh_str);

      build_header_str(hdr_str,
         conf.model().sval(),
         fcst_wd.get_lead_time(),
         fcst_wd.get_valid_time(),
         msg_type,
         cts_info.name,
         type_str,
         field_name,
         level_name);

      build_cfp_str(col_str, cts_info);

      sprintf(line, "%s %s", hdr_str, col_str);

      *vsdb_out << line << "\n";

      if(conf.output_flag(i_cfp).ival() >= 2) *cfp_out << line << "\n";
   }

   //
   // COP line type
   //
   if(conf.output_flag(i_cop).ival() != 0 &&
      cts_info.cts.n() > 0) {

      sprintf(type_str, "COP%s", thresh_str);

      build_header_str(hdr_str,
         conf.model().sval(),
         fcst_wd.get_lead_time(),
         fcst_wd.get_valid_time(),
         msg_type,
         cts_info.name,
         type_str,
         field_name,
         level_name);

      build_cop_str(col_str, cts_info);

      sprintf(line, "%s %s", hdr_str, col_str);

      *vsdb_out << line << "\n";

      if(conf.output_flag(i_cop).ival() >= 2) *cop_out << line << "\n";
   }

   //
   // CTS line type
   //
   if(conf.output_flag(i_cts).ival() != 0 &&
      cts_info.cts.n() > 0) {

      for(i=0; i<cts_info.n_alpha; i++) {

         sprintf(type_str, "CTS%s/%.3f", thresh_str, cts_info.alpha[i]);

         build_header_str(hdr_str,
            conf.model().sval(),
            fcst_wd.get_lead_time(),
            fcst_wd.get_valid_time(),
            msg_type,
            cts_info.name,
            type_str,
            field_name,
            level_name);

         build_cts_str(col_str, cts_info, i);

         sprintf(line, "%s %s", hdr_str, col_str);

         *vsdb_out << line << "\n";

         if(conf.output_flag(i_cts).ival() >= 2) *cts_out << line << "\n";
      } // end for i
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void write_nbrcnt(NBRCNTInfo &nbrcnt_info, int i_gc,
                  const WrfData &fcst_wd) {
   char hdr_str[max_line_len], col_str[max_line_len];
   char type_str[max_str_len], thresh_str[max_str_len];
   char line[max_line_len];
   int i;

   //
   // Grib Field abbreviation
   //
   get_grib_code_abbr(gc_info[i_gc].code, conf.ncep_defaults().ival(),
                      abbr_str);

   //
   // Type of thresholding performed
   //
   nbrcnt_info.raw_thresh.get_str(thresh_str);

   //
   // NBRCNT line type
   //
   if(conf.output_flag(i_nbrcnt).ival() != 0 &&
      nbrcnt_info.cnt_info.n > 0) {

      for(i=0; i<n_ci_alpha; i++) {

         sprintf(type_str, "NBRCNT%s/%.3f", thresh_str,
                 nbrcnt_info.cnt_info.alpha[i]);

         build_header_str(hdr_str,
            conf.model().sval(),
            fcst_wd.get_lead_time(),
            fcst_wd.get_valid_time(),
            msg_type,
            nbrcnt_info.cnt_info.name,
            type_str,
            field_name,
            level_name);

         build_nbrcnt_str(col_str, nbrcnt_info, i);

         sprintf(line, "%s %s", hdr_str, col_str);

         *vsdb_out << line << "\n";

         if(conf.output_flag(i_nbrcnt).ival() >= 2)
            *nbrcnt_out << line << "\n";
      } // end for i
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void write_nbrcts(NBRCTSInfo &nbrcts_info, int i_gc,
                  const WrfData &fcst_wd) {
   char hdr_str[max_line_len], col_str[max_line_len];
   char type_str[max_str_len];
   char thresh_str[max_str_len], frac_thresh_str[max_str_len];
   char line[max_line_len];
   int i;

   //
   // Grib Field abbreviation
   //
   get_grib_code_abbr(gc_info[i_gc].code, conf.ncep_defaults().ival(),
                      abbr_str);

   //
   // Type of thresholding performed
   //
   nbrcts_info.raw_thresh.get_str(thresh_str);
   nbrcts_info.frac_thresh.get_str(frac_thresh_str);

   //
   // NBRCTC line type
   //
   if(conf.output_flag(i_nbrctc).ival() != 0 &&
      nbrcts_info.cts_info.cts.n() > 0) {

      sprintf(type_str, "NBRCTC%s%s", thresh_str, frac_thresh_str);

      build_header_str(hdr_str,
         conf.model().sval(),
         fcst_wd.get_lead_time(),
         fcst_wd.get_valid_time(),
         msg_type,
         nbrcts_info.cts_info.name,
         type_str,
         field_name,
         level_name);

      build_nbrctc_str(col_str, nbrcts_info);

      sprintf(line, "%s %s", hdr_str, col_str);

      *vsdb_out << line << "\n";

      if(conf.output_flag(i_nbrctc).ival() >= 2) *nbrctc_out << line << "\n";
   }

   //
   // NBRCTS line type
   //
   if(conf.output_flag(i_nbrcts).ival() != 0 &&
      nbrcts_info.cts_info.cts.n() > 0) {

      for(i=0; i<n_ci_alpha; i++) {

         sprintf(type_str, "NBRCTS%s%s/%.3f", thresh_str,
                 frac_thresh_str, nbrcts_info.cts_info.alpha[i]);

         build_header_str(hdr_str,
            conf.model().sval(),
            fcst_wd.get_lead_time(),
            fcst_wd.get_valid_time(),
            msg_type,
            nbrcts_info.cts_info.name,
            type_str,
            field_name,
            level_name);

         build_nbrcts_str(col_str, nbrcts_info, i);

         sprintf(line, "%s %s", hdr_str, col_str);

         *vsdb_out << line << "\n";

         if(conf.output_flag(i_nbrcts).ival() >= 2)
            *nbrcts_out << line << "\n";
      } // end for i
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void close_txt_files() {

   // Close the output VSDB file
   if(vsdb_out) {

      // List VSDB file after it is finished
      if(verbosity > 0) {
         cout << "Output VSDB file: "
              << vsdb_file << "\n" << flush;
      }

      // Close the output VSDB file
      vsdb_out->close();
      delete vsdb_out;
      vsdb_out = (ofstream *) 0;
   }

   // Close the output FHO txt file
   if(conf.output_flag(i_fho).ival() >= 2 && fho_out) {

      // List FHO file after it is finished
      if(verbosity > 0) {
         cout << "Output FHO file: "
              << fho_file << "\n" << flush;
      }

      // Close the output FHO file
      fho_out->close();
      delete fho_out;
      fho_out = (ofstream *) 0;
   }

   // Close the output CTC txt file
   if(conf.output_flag(i_ctc).ival() >= 2 && ctc_out) {

      // List CTC file after it is finished
      if(verbosity > 0) {
         cout << "Output CTC file: "
              << ctc_file << "\n" << flush;
      }

      // Close the output CTC file
      ctc_out->close();
      delete ctc_out;
      ctc_out = (ofstream *) 0;
   }

   // Close the output CTP txt file
   if(conf.output_flag(i_ctp).ival() >= 2 && ctp_out) {

      // List CTP file after it is finished
      if(verbosity > 0) {
         cout << "Output CTP file: "
              << ctp_file << "\n" << flush;
      }

      // Close the output CTP file
      ctp_out->close();
      delete ctp_out;
      ctp_out = (ofstream *) 0;
   }

   // Close the output CFP txt file
   if(conf.output_flag(i_cfp).ival() >= 2 && cfp_out) {

      // List CFP file after it is finished
      if(verbosity > 0) {
         cout << "Output CFP file: "
              << cfp_file << "\n" << flush;
      }

      // Close the output CFP file
      cfp_out->close();
      delete cfp_out;
      cfp_out = (ofstream *) 0;
   }

   // Close the output COP txt file
   if(conf.output_flag(i_cop).ival() >= 2 && cop_out) {

      // List COP file after it is finished
      if(verbosity > 0) {
         cout << "Output COP file: "
              << cop_file << "\n" << flush;
      }

      // Close the output COP file
      cop_out->close();
      delete cop_out;
      cop_out = (ofstream *) 0;
   }

   // Close the output CTS txt file
   if(conf.output_flag(i_cts).ival() >= 2 && cts_out) {

      // List CTS file after it is finished
      if(verbosity > 0) {
         cout << "Output CTS file: "
              << cts_file << "\n" << flush;
      }

      // Close the output CTS file
      cts_out->close();
      delete cts_out;
      cts_out = (ofstream *) 0;
   }

   // Close the output CNT txt file
   if(conf.output_flag(i_cnt).ival() >= 2 && cnt_out) {

      // List CNT file after it is finished
      if(verbosity > 0) {
         cout << "Output CNT file: "
              << cnt_file << "\n" << flush;
      }

      // Close the output CNT file
      cnt_out->close();
      delete cnt_out;
      cnt_out = (ofstream *) 0;
   }

   // Close the output SL1L2 txt file
   if(conf.output_flag(i_sl1l2).ival() >= 2 && sl1l2_out) {

      // List SL1L2 file after it is finished
      if(verbosity > 0) {
         cout << "Output SL1L2 file: "
              << sl1l2_file << "\n" << flush;
      }

      // Close the output SL1L2 file
      sl1l2_out->close();
      delete sl1l2_out;
      sl1l2_out = (ofstream *) 0;
   }

   // Close the output NBRCTC txt file
   if(conf.output_flag(i_nbrctc).ival() >= 2 && nbrctc_out) {

      // List NBRCTC file after it is finished
      if(verbosity > 0) {
         cout << "Output NBRCTC file: "
              << nbrctc_file << "\n" << flush;
      }

      // Close the output NBRCTC file
      nbrctc_out->close();
      delete nbrctc_out;
      nbrctc_out = (ofstream *) 0;
   }

   // Close the output NBRCTS txt file
   if(conf.output_flag(i_nbrcts).ival() >= 2 && nbrcts_out) {

      // List NBRCTS file after it is finished
      if(verbosity > 0) {
         cout << "Output NBRCTS file: "
              << nbrcts_file << "\n" << flush;
      }

      // Close the output NBRCTS file
      nbrcts_out->close();
      delete nbrcts_out;
      nbrcts_out = (ofstream *) 0;
   }

   // Close the output NBRCNT txt file
   if(conf.output_flag(i_nbrcnt).ival() >= 2 && nbrcnt_out) {

      // List NBRCNT file after it is finished
      if(verbosity > 0) {
         cout << "Output NBRCNT file: "
              << nbrcnt_file << "\n" << flush;
      }

      // Close the output NBRCNT file
      nbrcnt_out->close();
      delete nbrcnt_out;
      nbrcnt_out = (ofstream *) 0;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void apply_mask(const WrfData &fcst_wd, const WrfData &obs_wd,
                int i_mask, NumArray &f_na, NumArray &o_na) {
   int x, y;

   //
   // Initialize the NumArray objects
   //
   f_na.clear();
   o_na.clear();

   //
   // Store the pairs in NumArray objects
   //
   for(x=0; x<fcst_wd.get_nx(); x++) {
      for(y=0; y<fcst_wd.get_ny(); y++) {

         // Skip any grid points containing bad data values for
         // either of the raw fields or where the verification
         // masking region is turned off
         if(fcst_wd.is_bad_data(x, y) ||
            obs_wd.is_bad_data(x, y) ||
            !mask_wd[i_mask].s_is_on(x, y) ) continue;

         f_na.add(fcst_wd.get_xy_double(x, y));
         o_na.add(obs_wd.get_xy_double(x, y));
      } // end for y
   } // end for x

   return;
}

////////////////////////////////////////////////////////////////////////

void do_cnt_cts(CNTInfo &cnt_info, CTSInfo *&cts_info,
                int i_gc, int i_interp, int i_mask,
                const NumArray &f_na, const NumArray &o_na) {
   int i, j, n_cts;

   //
   // Set up the CNTInfo alpha values
   //
   cnt_info.allocate_n_alpha(n_ci_alpha);
   for(i=0; i<n_ci_alpha; i++) {
      cnt_info.alpha[i] = conf.ci_alpha(i).dval();
   }

   //
   // Set up the CTSInfo thresholds and alpha values
   //
   n_cts = gc_ti[i_gc].n_thresh;
   for(i=0; i<n_cts; i++) {
      cts_info[i].cts_thresh = gc_ti[i_gc].thresh[i];
      cts_info[i].allocate_n_alpha(n_ci_alpha);

      for(j=0; j<n_ci_alpha; j++) {
         cts_info[i].alpha[j] = conf.ci_alpha(j).dval();
      }
   }

   //
   // Keep track of the interpolation method and width
   //
   cnt_info.interp_mthd = interp_mthd[i_interp];
   cnt_info.interp_wdth = interp_wdth[i_interp];
   for(i=0; i<n_cts; i++) {
      cts_info[i].interp_mthd = interp_mthd[i_interp];
      cts_info[i].interp_wdth = interp_wdth[i_interp];
   }

   //
   // Keep track of the masking region
   //
   cnt_info.set_name(mask_names[i_mask]);
   for(i=0; i<n_cts; i++) {
      cts_info[i].set_name(mask_names[i_mask]);
   }

   if(verbosity > 1) {

      get_grib_code_abbr(gc_info[i_gc].code, conf.ncep_defaults().ival(),
                         abbr_str);

      cout << "Computing CNT, CTS, and SL1L2 for Grib Field " << abbr_str
           << " at " << gc_info[i_gc].lvl_str
           << ", for interpolation method " << mthd_str[cnt_info.interp_mthd]
           << "(" << cnt_info.interp_wdth << "), over region "
           << cnt_info.name << ", using "
           << f_na.n_elements() << " pairs\n" << flush;
   }

   //
   // Compute the stats, normal confidence intervals, and bootstrap
   // bootstrap confidence intervals
   //
   if(conf.boot_interval().ival() == boot_bca_flag) {
      compute_stats_ci_bca(rng_ptr, f_na, o_na, gc_info[i_gc].code,
         conf.n_boot_rep().ival(),
         cnt_info, cts_info, n_cts, conf.rank_corr_flag().ival(),
         conf.tmp_dir().sval());
   }
   else {
      compute_stats_ci_perc(rng_ptr, f_na, o_na, gc_info[i_gc].code,
         conf.n_boot_rep().ival(), conf.boot_rep_prop().dval(),
         cnt_info, cts_info, n_cts, conf.rank_corr_flag().ival(),
         conf.tmp_dir().sval());
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void do_nbrcnt_nbrcts(NBRCNTInfo &nbrcnt_info, NBRCTSInfo *&nbrcts_info,
                      int i_gc, int i_wdth, int i_thresh, int i_mask,
                      const NumArray &f_na, const NumArray &o_na) {
   int i, j, n_nbrcts;
   char thresh_str[max_str_len];

   //
   // Set up the NBRCNTInfo threshold and alpha values
   //
   nbrcnt_info.raw_thresh = gc_ti[i_gc].thresh[i_thresh];
   nbrcnt_info.allocate_n_alpha(n_ci_alpha);
   for(i=0; i<n_ci_alpha; i++) {
      nbrcnt_info.cnt_info.alpha[i] = conf.ci_alpha(i).dval();
   }

   //
   // Set up the NBRCTSInfo thresholds and alpha values
   //
   n_nbrcts = frac_ti.n_thresh;
   for(i=0; i<n_nbrcts; i++) {

      nbrcts_info[i].raw_thresh = gc_ti[i_gc].thresh[i_thresh];
      nbrcts_info[i].frac_thresh = frac_ti.thresh[i];

      nbrcts_info[i].cts_info.cts_thresh = frac_ti.thresh[i];
      nbrcts_info[i].allocate_n_alpha(n_ci_alpha);

      for(j=0; j<n_ci_alpha; j++) {
         nbrcts_info[i].cts_info.alpha[j] = conf.ci_alpha(j).dval();
      }
   }

   //
   // Keep track of the neighborhood width
   //
   nbrcnt_info.nbr_wdth = conf.nbr_width(i_wdth).ival();
   for(i=0; i<n_nbrcts; i++) {
      nbrcts_info[i].nbr_wdth = conf.nbr_width(i_wdth).ival();
   }

   //
   // Keep track of the masking region
   //
   nbrcnt_info.cnt_info.set_name(mask_names[i_mask]);
   for(i=0; i<n_nbrcts; i++) {
      nbrcts_info[i].cts_info.set_name(mask_names[i_mask]);
   }

   if(verbosity > 1) {

      get_grib_code_abbr(gc_info[i_gc].code, conf.ncep_defaults().ival(),
                         abbr_str);

      nbrcnt_info.raw_thresh.get_str(thresh_str);

      cout << "Computing NBRCNT and NBRCTS for Grib Field " << abbr_str
           << " at " << gc_info[i_gc].lvl_str
           << ", for neighborhood width of " << nbrcnt_info.nbr_wdth
           << ", raw threshold of " << thresh_str
           << ", over region " << nbrcnt_info.cnt_info.name
           << " using " << f_na.n_elements() << " pairs\n" << flush;
   }

   //
   // Compute the stats, normal confidence intervals, and bootstrap
   // bootstrap confidence intervals
   //
   if(conf.boot_interval().ival() == boot_bca_flag) {
      compute_nbr_stats_ci_bca(rng_ptr, f_na, o_na, gc_info[i_gc].code,
         conf.n_boot_rep().ival(),
         nbrcnt_info, nbrcts_info, n_nbrcts, conf.tmp_dir().sval());
   }
   else {
      compute_nbr_stats_ci_perc(rng_ptr, f_na, o_na, gc_info[i_gc].code,
         conf.n_boot_rep().ival(), conf.boot_rep_prop().dval(),
         nbrcnt_info, nbrcts_info, n_nbrcts, conf.tmp_dir().sval());
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void write_nc(const WrfData &fcst_wd, const WrfData &obs_wd,
              int lev, int mthd, int wdth) {
   int i, n, x, y, mon, day, yr, hr, min, sec;
   float *fcst_data = (float *) 0;
   float *obs_data  = (float *) 0;
   float *diff_data = (float *) 0;
   double fval, oval;
   unixtime ut;
   NcVar *fcst_var, *obs_var, *diff_var;
   char fcst_var_name[max_str_len], obs_var_name[max_str_len];
   char diff_var_name[max_str_len];
   char time_str[max_str_len];
   char tmp_str[max_str_len];

   // Allocate memory for the forecast, observation, and difference fields
   fcst_data = new float [grid.nx()*grid.ny()];
   obs_data  = new float [grid.nx()*grid.ny()];
   diff_data = new float [grid.nx()*grid.ny()];

   // Get the grib code abbreviation string
   get_grib_code_abbr(gc_info[lev].code, conf.ncep_defaults().ival(),
                      abbr_str);

   // Compute the difference field for each of the masking regions
   for(i=0; i<n_masks; i++) {

      // If the neighborhood width is greater than 1, name the variables
      // using the smoothing method
      if(wdth > 1) {
         sprintf(fcst_var_name, "FCST_%s_%s_%s_%s_%i",
                 abbr_str, gc_info[lev].lvl_str, mask_names[i],
                 mthd_str[mthd], wdth*wdth);
         sprintf(obs_var_name, "OBS_%s_%s_%s_%s_%i",
                 abbr_str, gc_info[lev].lvl_str, mask_names[i],
                 mthd_str[mthd], wdth*wdth);
         sprintf(diff_var_name, "DIFF_%s_%s_%s_%s_%i",
                 abbr_str, gc_info[lev].lvl_str, mask_names[i],
                 mthd_str[mthd], wdth*wdth);
      }
      // If the neighborhood width is equal to 1, no smoothing was done
      else {

         sprintf(fcst_var_name, "FCST_%s_%s_%s",
                 abbr_str, gc_info[lev].lvl_str, mask_names[i]);
         sprintf(obs_var_name, "OBS_%s_%s_%s",
                 abbr_str, gc_info[lev].lvl_str, mask_names[i]);
         sprintf(diff_var_name, "DIFF_%s_%s_%s",
                 abbr_str, gc_info[lev].lvl_str, mask_names[i]);
      }

      // Define the forecast and difference variables
      fcst_var = nc_out->add_var(fcst_var_name, ncFloat, lat_dim, lon_dim);
      obs_var  = nc_out->add_var(obs_var_name,  ncFloat, lat_dim, lon_dim);
      diff_var = nc_out->add_var(diff_var_name, ncFloat, lat_dim, lon_dim);

      // Add variable attributes for the forecast field
      fcst_var->add_att("name", field_name);
      get_grib_code_unit(gc_info[lev].code, conf.ncep_defaults().ival(), tmp_str);
      fcst_var->add_att("units", tmp_str);
      get_grib_code_name(gc_info[lev].code, conf.ncep_defaults().ival(), tmp_str);
      fcst_var->add_att("long_name", tmp_str);
      fcst_var->add_att("level", level_name);

      fcst_var->add_att("_FillValue", bad_data_float);

      ut = fcst_wd.get_valid_time();
      unix_to_mdyhms(ut, mon, day, yr, hr, min, sec);
      sprintf(time_str, "%.4i-%.2i-%.2i %.2i:%.2i:%.2i", yr, mon, day, hr, min, sec);
      fcst_var->add_att("valid_time", time_str);
      fcst_var->add_att("valid_time_ut", (long int) ut);

      fcst_var->add_att("masking_region", mask_names[i]);
      fcst_var->add_att("smoothing_method", mthd_str[mthd]);
      fcst_var->add_att("smoothing_neighborhood", wdth*wdth);

      // Add variable attributes for the observation field
      obs_var->add_att("name", field_name);
      get_grib_code_unit(gc_info[lev].code, conf.ncep_defaults().ival(), tmp_str);
      obs_var->add_att("units", tmp_str);
      get_grib_code_name(gc_info[lev].code, conf.ncep_defaults().ival(), tmp_str);
      obs_var->add_att("long_name", tmp_str);
      obs_var->add_att("level", level_name);

      obs_var->add_att("_FillValue", bad_data_float);

      ut = obs_wd.get_valid_time();
      unix_to_mdyhms(ut, mon, day, yr, hr, min, sec);
      sprintf(time_str, "%.4i-%.2i-%.2i %.2i:%.2i:%.2i", yr, mon, day, hr, min, sec);
      obs_var->add_att("valid_time", time_str);
      obs_var->add_att("valid_time_ut", (long int) ut);

      obs_var->add_att("masking_region", mask_names[i]);
      obs_var->add_att("smoothing_method", mthd_str[mthd]);
      obs_var->add_att("smoothing_neighborhood", wdth*wdth);

      // Add variable attributes for the difference field
      diff_var->add_att("name", field_name);
      get_grib_code_unit(gc_info[lev].code, conf.ncep_defaults().ival(), tmp_str);
      diff_var->add_att("units", tmp_str);
      get_grib_code_name(gc_info[lev].code, conf.ncep_defaults().ival(), tmp_str);
      diff_var->add_att("long_name", tmp_str);
      diff_var->add_att("level", level_name);

      diff_var->add_att("_FillValue", bad_data_float);

      ut = fcst_wd.get_valid_time();
      unix_to_mdyhms(ut, mon, day, yr, hr, min, sec);
      sprintf(time_str, "%.4i-%.2i-%.2i %.2i:%.2i:%.2i", yr, mon, day, hr, min, sec);
      diff_var->add_att("fcst_valid_time", time_str);
      fcst_var->add_att("fcst_valid_time_ut", (long int) ut);

      ut = obs_wd.get_valid_time();
      unix_to_mdyhms(ut, mon, day, yr, hr, min, sec);
      sprintf(time_str, "%.4i-%.2i-%.2i %.2i:%.2i:%.2i", yr, mon, day, hr, min, sec);
      diff_var->add_att("obs_valid_time", time_str);
      fcst_var->add_att("obs_valid_time_ut", (long int) ut);

      diff_var->add_att("masking_region", mask_names[i]);
      diff_var->add_att("smoothing_method", mthd_str[mthd]);
      diff_var->add_att("smoothing_neighborhood", wdth*wdth);

      // Store the forecast, observation, and difference values
      for(x=0; x<grid.nx(); x++) {
         for(y=0; y<grid.ny(); y++) {

            n = two_to_one(grid.nx(), grid.ny(), x, y);

            // Check whether the mask is on or off for this point
            if(!mask_wd[i].s_is_on(x, y)) {
               fcst_data[n] = obs_data[n] = diff_data[n] = bad_data_float;
               continue;
            }

            // Retrieve the forecast and observation values
            fval = fcst_wd.get_xy_double(x, y);
            oval = obs_wd.get_xy_double(x, y);

            // Set the forecast data
            if(fval == bad_data_double) fcst_data[n] = bad_data_float;
            else                        fcst_data[n] = fval;

            // Set the observation data
            if(oval == bad_data_double) obs_data[n]  = bad_data_float;
            else                        obs_data[n]  = oval;

            // Set the difference data
            if(fval == bad_data_double ||
               oval == bad_data_double) diff_data[n] = bad_data_float;
            else                        diff_data[n] = fval - oval;
         } // end for y
      } // end for x

      // Write out the forecast field
      if(!fcst_var->put(&fcst_data[0], grid.ny(), grid.nx())) {
         cerr << "\n\nERROR: write_nc() -> "
              << "error with the fcst_var->put for grib code " << field_name
              << " and masking region " << mask_names[i] << "\n\n" << flush;
         exit(1);
      }

      // Write out the observation field
      if(!obs_var->put(&obs_data[0], grid.ny(), grid.nx())) {
         cerr << "\n\nERROR: write_nc() -> "
              << "error with the obs_var->put for grib code " << field_name
              << " and masking region " << mask_names[i] << "\n\n" << flush;
         exit(1);
      }

      // Write out the difference field
      if(!diff_var->put(&diff_data[0], grid.ny(), grid.nx())) {
         cerr << "\n\nERROR: write_nc() -> "
              << "error with the diff_var->put for grib code " << field_name
              << " and masking region " << mask_names[i] << "\n\n" << flush;
         exit(1);
      }

   } // end for i

   // Deallocate and clean up
   if(fcst_data) { delete [] fcst_data; fcst_data = (float *) 0; }
   if(obs_data)  { delete [] obs_data;  obs_data  = (float *) 0; }
   if(diff_data) { delete [] diff_data; diff_data = (float *) 0; }

   return;
}

////////////////////////////////////////////////////////////////////////

WrfData smooth_field(WrfData *wd, int mthd, int wdth) {
   WrfData smooth_wd;
   double v;
   int x, y, x_ll, y_ll;

   // Initialize the smoothed field to the raw field
   smooth_wd = *wd;

   // Check for a width value of 1 for which no smoothing should
   // be performed
   if(wdth <= 1) return(smooth_wd);

   // Otherwise, apply smoothing to each grid point
   for(x=0; x<wd->get_nx(); x++) {
      for(y=0; y<wd->get_ny(); y++) {

         // The neighborhood width must be odd, find the lower-left
         // corner of the neighborhood
         x_ll = x - (wdth - 1)/2;
         y_ll = y - (wdth - 1)/2;

         // Compute the smoothed value based on the interpolation method
         switch(mthd) {

            case i_min:     // Minimum
               v = interp_min(wd, x_ll, y_ll, wdth,
                              conf.interp_threshold().dval());
               break;

            case i_max:     // Maximum
               v = interp_max(wd, x_ll, y_ll, wdth,
                              conf.interp_threshold().dval());
               break;

            case i_median:  // Median
               v = interp_median(wd, x_ll, y_ll, wdth,
                                 conf.interp_threshold().dval());
               break;

            case i_uw_mean: // Unweighted Mean
               v = interp_uw_mean(wd, x_ll, y_ll, wdth,
                                  conf.interp_threshold().dval());
               break;

            // Distance-weighted mean is omitted here since it is not
            // an option for grid_stat

            // Least-squares fit is omitted here since it is not
            // an option for grid_stat

            default:
               cerr << "\n\nERROR: smooth_field() -> "
                    << "unexpected interpolation method encountered: "
                    << mthd << "\n\n" << flush;
               exit(1);
               break;
         }

         // Store the smoothed value
         smooth_wd.put_xy_double(v, x, y);
      } // end for y
   } // end for x

   return(smooth_wd);
}

////////////////////////////////////////////////////////////////////////

void usage(int argc, char *argv[]) {

   cout << "\nUsage: "
        << program_name << "\n"
        << "\tfcst_file\n"
        << "\tobs_file\n"
        << "\tconfig_file\n"
        << "\t[-outdir path]\n"
        << "\t[-v level]\n\n"

        << "\twhere\t\"fcst_file\" is a forecast file in either Grib "
        << "or netCDF format (output of pcp_combine) containing the "
        << "field(s) to be verified (required).\n"

        << "\t\t\"obs_file\" is an observation file in either Grib "
        << "or netCDF format (output of pcp_combine) containing the "
        << "verifying field(s) (required).\n"

        << "\t\t\"config_file\" is a GridStatConfig file containing "
        << "the desired configuration settings (required).\n"

        << "\t\t\"-outdir path\" overrides the default output directory ("
        << out_dir << ") (optional).\n"

        << "\t\t\"-v level\" overrides the default level of logging ("
        << verbosity << ") (optional).\n"

        << "\n\tNOTE: The forecast and observation fields must be "
        << "on the same grid.\n\n" << flush;

   return;
}

////////////////////////////////////////////////////////////////////////
