// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*
// ** Copyright UCAR (c) 1992 - 2007
// ** University Corporation for Atmospheric Research (UCAR)
// ** National Center for Atmospheric Research (NCAR)
// ** Research Applications Lab (RAL)
// ** P.O.Box 3000, Boulder, Colorado, 80307-3000, USA
// *=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*=*

///////////////////////////////////////////////////////////////////////
//
//   Filename:   mode.cc
//
//   Description:
//      Based on user specified parameters, this tool derives objects
//      within two gridded datasets using a convolution-thresholding
//      approach.  It then compares objects within the same field and
//      across fields, and calculates a total interest value for each
//      pair of objects using a fuzzy logic approach.  The interest
//      values are thresholded, and object pairs with a high enough
//      interest value are associated with one another.
//
//   Mod#   Date      Name           Description
//   ----   ----      ----           -----------
//   000    10-11-06  Halley Gotway  New
//   001    12-20-06  Halley Gotway  Write check_xy_ll to reduce
//          duplicated code.
//   002    01/11/08  Halley Gotway  Modify sprintf statements which
//          use the grib code abbreviation string
//   003    02/06/08  Halley Gotway  Modify to read the updated NetCDF
//          output of PCP-Combine
//
////////////////////////////////////////////////////////////////////////

using namespace std;

#include <cstdio>
#include <cstdlib>
#include <ctime>
#include <ctype.h>
#include <dirent.h>
#include <iostream>
#include <fstream>
#include <math.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "netcdf.hh"
#include "vx_grib_classes/grib_classes.h"

#include "vx_wrfdata/vx_wrfdata.h"
#include "vx_wrfmode/vx_wrfmode.h"
#include "vx_met_util/vx_met_util.h"
#include "vx_data_grids/grid.h"
#include "vx_util/vx_util.h"
#include "vx_cal/vx_cal.h"
#include "vx_math/vx_math.h"
#include "vx_contable/vx_contable.h"

#include "vx_color/vx_color.h"
#include "vx_ps/vx_ps.h"
#include "vx_pxm/vx_pxm.h"
#include "vx_render/vx_render.h"
#include "vx_plot_util/vx_plot_util.h"

////////////////////////////////////////////////////////////////////////
//
// Create output files using the following naming convention:
//
//    mode_YYYYMMDDI_FHH_HHA.ps/.txt
//
//    Where I indicates initilization time, L indicates lead time,
//    and A indicates accumulation time.
//
////////////////////////////////////////////////////////////////////////

static const char * program_name = "mode";

static int model_width = default_model_width;

static int verbosity = 1;
static const int n_cts = 3;
static const char *cts_str[n_cts] = {"RAW", "FILTER", "OBJECT"};
static TTContingencyTable cts[n_cts];

// Grid definition data structures
static PlateCarreeData   pc_data;
static LambertData       lc_data;
static StereographicData st_data;

static Engine engine;
static Grid grid;
static BoundingBox xy_bb;
static BoundingBox ll_bb;
static char out_dir[max_str_len];
static char met_data_dir[max_str_len];

static char fcst_field_name[max_str_len], obs_field_name[max_str_len];
static char fcst_level_name[max_str_len], obs_level_name[max_str_len];
static char fcst_unit_name[max_str_len], obs_unit_name[max_str_len];

// Flags to indicate the format of the input files.
//    0 = netCDF (output of pcp_combine)
//    1 = Grib Version 1
static int fcst_fmt_flag = 0;
static int obs_fmt_flag  = 0;

// Grib Codes to be verified for the forecast and observation fields
static GCInfo fcst_gc, obs_gc;
static double data_min, data_max;

///////////////////////////////////////////////////////////////////////////////
//
// Plotting Info
//
///////////////////////////////////////////////////////////////////////////////

static ColorTable fcst_ct, obs_ct;
static int stride = 1;
static int plot_flag = 1;
static int obj_stat_flag = 1;
static int obj_plot_flag = 1;
static int ct_stat_flag = 1;
static const Color c_map(25, 25, 25);
static const Color c_hull(0, 0, 0);
static const Color c_bndy(0, 0, 255);
static const double l_width = 0.50;

static int n_page;

static const double page_width = 8.5*72.0;
static const double page_height = 11.0*72.0;

static const double h_margin = 20.0;
static const double v_margin = 20.0;

static const double plot_text_sep = 15.0;


static const BoundingBox sm_pane_bb = { h_margin, 235.0,
                                        450.0, 720.0,
                                        450.0-h_margin, 720.0-235.0 };
static const BoundingBox lg_pane_bb = { h_margin, v_margin,
                                        page_width-3.0*h_margin, page_height-4.0*v_margin,
                                        page_width-4.0*h_margin, page_height-5.0*v_margin };

static const double h_tab_cen = page_width/2.0;

static double sm_plot_height, lg_plot_height;

static double h_tab_1, h_tab_2, h_tab_3;
static double v_tab_1, v_tab_2, v_tab_3;

static Color c_fcst_fill(150, 150, 150);
static Color c_obs_fill(150, 150, 150);

///////////////////////////////////////////////////////////////////////////////

static void parse_masks();
static void check_engine_config();
static void compute_ct_stats();

static void set_plot_dims(int, int);
static void plot_engine();
static void plot_engine(PSfile &, Engine &, const char *, int);
static void plot_threshold_merging(PSfile &, Engine &, const char *, int);
static void draw_border(PSfile &, BoundingBox &);
static void draw_map(PSfile &, BoundingBox &);
static void render_image(PSfile &, Engine &, const WrfData &, BoundingBox &, int, int);
static void plot_shape_numbers(PSfile &, Engine &, BoundingBox &, int);
static void draw_convex_hulls(PSfile &, Engine &, const BoundingBox &, int);
static void draw_boundaries(PSfile &, WrfData &, int, const BoundingBox &);
static void draw_polyline(PSfile &, Polyline &, const Color &, BoundingBox, bool);
static void plot_colorbar(PSfile &, BoundingBox &, int);
static void composite_mask(WrfData &, Engine &, int, int);
static void set_dim(BoundingBox &, double, double, double);
static void set_xy_bb();
static void valid_xy_bb(const WrfData *, BoundingBox &);
static void set_ll_bb();
static void check_xy_ll(int, int);
static void write_obj_stats();
static void write_obj_netcdf();
static void write_ct_stats();
static void usage(int, char **);

////////////////////////////////////////////////////////////////////////

int main(int argc, char *argv[]) {
   int i, h, status;
   char fcst_file[max_str_len], obs_file[max_str_len];
   char match_config_file[max_str_len], merge_config_file[max_str_len];
   char merge_str[max_str_len];
   char fcst_abbr_str[max_str_len], obs_abbr_str[max_str_len];
   Grid fcst_grid, obs_grid;
   GribFile fcst_gb_file, obs_gb_file;
   GribRecord fcst_r, obs_r;
   NcFile *fcst_nc_file, *obs_nc_file;
   WrfData fcst_wd, obs_wd;

   sprintf(out_dir, "%s/out/mode", MET_BASE);

   if(argc < 4) {
      usage(argc, argv);
      return(0);
   }

   strcpy(fcst_file, argv[1]);
   strcpy(obs_file, argv[2]);
   strcpy(match_config_file, argv[3]);

   //
   // Set the merge config file to the match config file by default
   // but it may be overridden using the optional arguments
   //
   strcpy(merge_config_file, match_config_file);

   //
   // Parse command line arguments
   //
   for(i=0; i<argc; i++) {

      if(strcmp(argv[i], "-config_merge") == 0) {
         strcpy(merge_config_file, argv[i+1]);
         i++;
      }
      else if(strcmp(argv[i], "-outdir") == 0) {
         strcpy(out_dir, argv[i+1]);
         i++;
      }
      else if(strcmp(argv[i], "-v") == 0) {
         verbosity = atoi(argv[i+1]);
         i++;
      }
      else if(strcmp(argv[i], "-plot") == 0) {
         plot_flag = 0;
      }
      else if(strcmp(argv[i], "-obj_stat") == 0) {
         obj_stat_flag = 0;
      }
      else if(strcmp(argv[i], "-obj_plot") == 0) {
         obj_plot_flag = 0;
      }
      else if(strcmp(argv[i], "-ct_stat") == 0) {
         ct_stat_flag = 0;
      }
   }

   //
   // Read the match config file
   //
   engine.wconf.read(match_config_file);

   //
   // Process the engine configuration
   //
   engine.process_engine_config();

   //
   // Check the configuration of the engine
   //
   check_engine_config();

   //
   // Parse out the grib code and level information for the forecast
   // and observation grib codes.
   //
   parse_gcinfo(engine.wconf.fcst_grib_code().sval(), fcst_gc);
   parse_gcinfo(engine.wconf.obs_grib_code().sval(), obs_gc);

   //
   // Check to make sure that MODE can be run on the grib codes specified
   //
   if(fcst_gc.code == wdir_grib_code || obs_gc.code == wdir_grib_code) {
      cerr << "\n\nERROR: main() -> "
           << "mode may not be run on a wind direction field.\n\n"
           << flush;
      exit(1);
   }

   // List the input files
   if(verbosity > 0) {
      cout << "Forecast File: " << fcst_file << "\n"
           << "Observation File: " << obs_file << "\n"
           << "Match Config File: " << match_config_file << "\n"
           << "Merge Config File: " << merge_config_file << "\n" << flush;
   }

   //
   // Attempt to open the forecast file as a NetCDF file.  If it
   // fails, attempt to read it as Grib.
   //
   fcst_nc_file = new NcFile(fcst_file);

   if(!fcst_nc_file || !(fcst_nc_file->is_valid())) {
      fcst_nc_file->close();
      delete fcst_nc_file;
      fcst_nc_file = (NcFile *) 0;
      fcst_fmt_flag = 1;
   }
   else {
      fcst_fmt_flag = 0;
   }

   //
   // Attempt to open the forecast file as a Grib file.
   //
   if(fcst_fmt_flag == 1) {

      if( !(fcst_gb_file.open(fcst_file)) ) {
         cerr << "\n\nERROR: main() -> "
              << "can't open grib file: "
              << fcst_file << "\n\n" << flush;
         exit(1);
      }
   }

   //
   // Attempt to open the observation file as a NetCDF file.  If it
   // fails, attempt to read it as Grib.
   //
   obs_nc_file = new NcFile(obs_file);

   if(!obs_nc_file || !(obs_nc_file->is_valid())) {
      obs_nc_file->close();
      delete obs_nc_file;
      obs_nc_file = (NcFile *) 0;
      obs_fmt_flag = 1;
   }
   else {
      obs_fmt_flag = 0;
   }

   //
   // Attempt to open the observation file as a Grib file.
   //
   if(obs_fmt_flag == 1) {

      if(!(obs_gb_file.open(obs_file)) ) {
         cerr << "\n\nERROR: main() -> "
              << "can't open grib file: "
              << obs_file << "\n\n" << flush;
         exit(1);
      }
   }

   //
   // Get the grib code abbreviation strings
   //
   get_grib_code_abbr(fcst_gc.code, engine.wconf.ncep_defaults().ival(),
                      fcst_abbr_str);
   get_grib_code_abbr(obs_gc.code, engine.wconf.ncep_defaults().ival(),
                      obs_abbr_str);

   //
   // Process the forecast file in NetCDF format (output of pcp_combine)
   //
   if(fcst_fmt_flag == 0) {
      read_pcp_combine_netcdf(fcst_nc_file, fcst_abbr_str, fcst_level_name,
                              fcst_wd, fcst_grid, verbosity,
                              pc_data, lc_data, st_data);

      //
      // Check that the accumulation interval matches what was requested
      // in the config file
      //
      if(fcst_wd.get_accum_time() != fcst_gc.lvl_1*sec_per_hour) {
         cerr << "\n\nERROR: main() -> "
              << "accumulation time of " << fcst_wd.get_accum_time()/sec_per_hour
              << " hours in the forecast file (" << fcst_file
              << ") does not match the " << fcst_gc.lvl_1
              << " hour accumulation time requested in the config file.\n\n"
              << flush;
         exit(1);
      }
   }
   //
   // Process the forecast file in Grib format
   //
   else {

      if(verbosity > 1) {
         cout << "Searching grib forecast file for grib code "
              << fcst_gc.code << " with accumulation/level indicator of "
              << fcst_gc.lvl_str << ".\n" << flush;
      }

      status = get_grib_record(fcst_gb_file, fcst_r,
                               fcst_gc.code, fcst_gc.lvl_1, fcst_gc.lvl_2,
                               fcst_wd, fcst_grid, verbosity,
                               pc_data, lc_data, st_data);

      if(status != 0) {
         cerr << "\n\nERROR: main() -> "
              << "grib code " << fcst_gc.code
              << " with accumulation/level indicator of "
              << fcst_gc.lvl_str << " not found in grib file: "
              << fcst_file << "\n" << flush;
         exit(1);
      }

      //
      // Get the forecast level name
      //
      get_grib_level_str(fcst_r.pds->type, engine.wconf.ncep_defaults().ival(),
                         fcst_r.pds->level_info, fcst_level_name);
   }

   //
   // Process the observation file in NetCDF format (output of pcp_combine)
   //
   if(obs_fmt_flag == 0) {
      read_pcp_combine_netcdf(obs_nc_file, obs_abbr_str, obs_level_name,
                              obs_wd, obs_grid, verbosity,
                              pc_data, lc_data, st_data);

      //
      // Check that the accumulation interval matches what was requested
      // in the config file
      //
      if(obs_wd.get_accum_time() != obs_gc.lvl_1*sec_per_hour) {
         cerr << "\n\nERROR: main() -> "
              << "accumulation time of " << obs_wd.get_accum_time()/sec_per_hour
              << " hours in the observation file (" << obs_file
              << ") does not match the " << obs_gc.lvl_1
              << " hour accumulation time requested in the config file.\n\n"
              << flush;
         exit(1);
      }
   }
   //
   // Process the observation file in Grib format
   //
   else {

      if(verbosity > 1) {
         cout << "Searching grib observation file for grib code "
              << obs_gc.code << " with accumulation/level indicator of "
              << obs_gc.lvl_str << ".\n" << flush;
      }

      status = get_grib_record(obs_gb_file, obs_r,
                               obs_gc.code, obs_gc.lvl_1, obs_gc.lvl_2,
                               obs_wd, obs_grid, verbosity,
                               pc_data, lc_data, st_data);

      if(status != 0) {
         cerr << "\n\nERROR: main() -> "
              << "grib code " << obs_gc.code
              << " with accumulation/level indicator of "
              << obs_gc.lvl_str << " not found in grib file: "
              << obs_file << "\n" << flush;
         exit(1);
      }

      //
      // Get the observation level name
      //
      get_grib_level_str(obs_r.pds->type, engine.wconf.ncep_defaults().ival(),
                         obs_r.pds->level_info, obs_level_name);
   }

   //
   // Check that the grid dimensions match
   //
   if(fcst_grid.nx() != obs_grid.nx() ||
      fcst_grid.ny() != obs_grid.ny()) {
      cerr << "\n\nERROR: main() -> "
           << "forecast and observation grid dimensions "
           << "do not match (" << fcst_grid.nx() << ", "
           << fcst_grid.ny() << ") != (" << obs_grid.nx()
           << ", " << obs_grid.ny() << ")\n\n" << flush;
      exit(1);
   }
   grid = fcst_grid;

   //
   // Print a warning if the valid times do not match
   //
   if(fcst_wd.get_valid_time() != obs_wd.get_valid_time()) {
      cout << "\n***WARNING***: main() -> "
           << "forecast and observation valid times do not match "
           << fcst_wd.get_valid_time() << " != "
           << obs_wd.get_valid_time() << ".\n\n" << flush;
   }

   //
   // Print a warning if the accumulation intervals do not match
   //
   if(fcst_wd.get_accum_time() != obs_wd.get_accum_time()) {
      cout << "\n***WARNING***: main() -> "
           << "forecast and observation accumulation times do not match "
           << fcst_wd.get_accum_time() << " != "
           << obs_wd.get_accum_time() << ".\n\n" << flush;
   }

   //
   // Setup the forecast field and unit strings
   //
   if(nint(fcst_wd.get_accum_time()/sec_per_hour) > 0) {
      h = nint(fcst_wd.get_accum_time()/sec_per_hour);
      sprintf(fcst_field_name, "%s_%.2i", fcst_abbr_str, h);
   }
   else {
      strcpy(fcst_field_name, fcst_abbr_str);
   }
   get_grib_code_unit(fcst_gc.code, engine.wconf.ncep_defaults().ival(),
                      fcst_unit_name);

   //
   // Setup the observation field and unit strings
   //
   if(nint(obs_wd.get_accum_time()/sec_per_hour) > 0) {
      h = nint(obs_wd.get_accum_time()/sec_per_hour);
      sprintf(obs_field_name, "%s_%.2i", obs_abbr_str, h);
   }
   else {
      strcpy(obs_field_name, obs_abbr_str);
   }
   get_grib_code_unit(obs_gc.code, engine.wconf.ncep_defaults().ival(),
                      obs_unit_name);

   //
   // Store the fcst/obs_field_name, and fcst/obs_level_name values
   // in the engine itself
   //
   strcpy(engine.fcst_field_name, fcst_field_name);
   strcpy(engine.obs_field_name,  obs_field_name);
   strcpy(engine.fcst_level_name, fcst_level_name);
   strcpy(engine.obs_level_name,  obs_level_name);

   if(verbosity > 0) {
      cout << "Forecast Field: " << fcst_field_name << " at "
           << fcst_level_name << "\n"
           << "Observation Field: " << obs_field_name << " at "
           << obs_level_name << "\n" << flush;
   }

   //
   // Mask out the missing data in one field with the other if requested
   //
   if(engine.wconf.mask_missing_flag().ival() == 1 ||
      engine.wconf.mask_missing_flag().ival() == 3)   mask_bad_data(fcst_wd, obs_wd);
   if(engine.wconf.mask_missing_flag().ival() == 2 ||
      engine.wconf.mask_missing_flag().ival() == 3)   mask_bad_data(obs_wd, fcst_wd);

   //
   // Parse the grid and/or polyline masks from the configuration
   //
   parse_masks();

   //
   // Apply the grid mask to the forecast field if requested
   //
   if(engine.wconf.mask_grid_flag().ival() == 1 ||
      engine.wconf.mask_grid_flag().ival() == 3) {
      apply_grid_mask(&grid, engine.grid_mask, fcst_wd, bad_data_flag);
   }

   //
   // Apply the grid mask to the observation field if requested
   //
   if(engine.wconf.mask_grid_flag().ival() == 2 ||
      engine.wconf.mask_grid_flag().ival() == 3) {
      apply_grid_mask(&grid, engine.grid_mask, obs_wd, bad_data_flag);
   }

   //
   // Apply the Lat/Lon polyline mask to the forecast field if requested
   //
   if(engine.wconf.mask_poly_flag().ival() == 1 ||
      engine.wconf.mask_poly_flag().ival() == 3) {

      apply_poly_mask_latlon(engine.poly_mask, &grid, fcst_wd, bad_data_flag);
   }

   //
   // Apply the Lat/Lon polyline mask to the observation field if requested
   //
   if(engine.wconf.mask_poly_flag().ival() == 2 ||
      engine.wconf.mask_poly_flag().ival() == 3) {

      apply_poly_mask_latlon(engine.poly_mask, &grid, obs_wd, bad_data_flag);
   }

   //
   // Compute the min and max data values across both raw fields for use
   // in setting up the color table
   //
   data_min = min(fcst_wd.int_to_double(0),
                  obs_wd.int_to_double(0));
   data_max = max(fcst_wd.int_to_double(wrfdata_int_data_max),
                  obs_wd.int_to_double(wrfdata_int_data_max));

   //
   // Set up the engine with these raw fields
   //
   if(verbosity > 0) {
         cout << "Identifying objects in the forecast and observation fields...\n";
   }
   engine.set(fcst_wd, obs_wd);

   //
   // Compute the contingency table statistics for the fields
   //
   if(ct_stat_flag) {
      compute_ct_stats();
   }

   if(plot_flag || obj_stat_flag || obj_plot_flag) {
      if(verbosity > 0) {
         cout << "Identified: " << engine.n_fcst << " forecast objects "
              << "and " << engine.n_obs << " observation objects.\n"
              << flush;
      }

      if(verbosity > 0) {

         if(engine.wconf.fcst_merge_flag().ival() == 1)
            strcpy(merge_str, "threshold merging");
         else if(engine.wconf.fcst_merge_flag().ival() == 2)
            strcpy(merge_str, "engine merging");
         else if(engine.wconf.fcst_merge_flag().ival() == 3)
            strcpy(merge_str, "threshold merging and engine merging");
         else
            strcpy(merge_str, "no merging");

         cout << "Performing merging (" << merge_str << ") in the forecast field.\n" << flush;
      }
      if(strcmp(merge_config_file, "default") == 0) engine.do_fcst_merging((char *) 0);
      else                                          engine.do_fcst_merging(merge_config_file);

      if(verbosity > 0) {

         if(engine.wconf.obs_merge_flag().ival() == 1)
            strcpy(merge_str, "threshold merging");
         else if(engine.wconf.obs_merge_flag().ival() == 2)
            strcpy(merge_str, "engine merging");
         else if(engine.wconf.obs_merge_flag().ival() == 3)
            strcpy(merge_str, "threshold merging and engine merging");
         else
            strcpy(merge_str, "no merging");

         cout << "Performing merging (" << merge_str << ") in the observation field.\n" << flush;
      }
      if(strcmp(merge_config_file, "default") == 0) engine.do_obs_merging((char *) 0);
      else                                          engine.do_obs_merging(merge_config_file);

      if(verbosity > 0) {
         cout << "Remaining: " << engine.n_fcst << " forecast objects "
              << "and " << engine.n_obs << " observation objects.\n"
              << flush;
      }

      if(verbosity > 0) {
         cout << "Performing matching between the forecast and observation fields.\n"
              << flush;
      }

      //
      // Do the matching of objects between fields
      //
      engine.do_matching();
   }

   //
   // Create output stats files and plots
   //
   if(ct_stat_flag)  write_ct_stats();
   if(obj_stat_flag) write_obj_stats();
   if(obj_plot_flag) write_obj_netcdf();
   if(plot_flag)     plot_engine();

   return(0);
}

////////////////////////////////////////////////////////////////////////

void parse_masks() {
   ifstream in;

   //
   // Check that the masking grid requested is a valid grid
   //
   if(strlen(engine.wconf.mask_grid().sval()) != 0) {
      if(find_grid(engine.wconf.mask_grid().sval(), *engine.grid_mask) != 0) {
         cerr << "\n\nERROR: parse_masks() -> "
              << "the mask_grid requested \""
              << engine.wconf.mask_grid().sval()
              << "\" is not defined.\n\n" << flush;
         exit(1);
      }
   }

   //
   // Process the mask_poly string as a file containing a lat/lon
   // masking polygon.
   //
   if(engine.wconf.mask_poly_flag().ival() != 0) {
      parse_latlon_poly_file(engine.wconf.mask_poly().sval(),
                             *engine.poly_mask);
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void check_engine_config() {
   ColorTable ct_test;
   char tmp_str[max_str_len];

   // Check that the model name is not empty
   if(strlen(engine.wconf.model().sval()) == 0 ||
      check_reg_exp(non_ws_reg_exp,
                    engine.wconf.model().sval()) != true) {

      cerr << "\n\nERROR: check_engine_config() -> "
           << "The model name must be set to a non-empty "
           << "string value.\n\n" << flush;
      exit(1);
   }

   // Check that grid_res is > 0
   if(engine.wconf.grid_res().dval() <= 0) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "grid_res (" << engine.wconf.grid_res().dval()
           << ") must be set > 0 in the configuration file\n\n"
           << flush;
      exit(1);
   }

   // Check that mask_missing_flag is set between 0 and 3
   if(engine.wconf.mask_missing_flag().ival() < 0 ||
      engine.wconf.mask_missing_flag().ival() > 3) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "The mask_missing_flag (" << engine.wconf.mask_missing_flag().ival()
           << ") must be set between 0 and 3\n\n" << flush;
      exit(1);
   }

   // Check that mask_grid_flag is set between 0 and 3
   if(engine.wconf.mask_grid_flag().ival() < 0 ||
      engine.wconf.mask_grid_flag().ival() > 3) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "The mask_grid_flag (" << engine.wconf.mask_grid_flag().ival()
           << ") must be set between 0 and 3\n\n" << flush;
      exit(1);
   }

   // Check that mask_poly_flag is set between 0 and 3
   if(engine.wconf.mask_poly_flag().ival() < 0 ||
      engine.wconf.mask_poly_flag().ival() > 3) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "The mask_poly_flag (" << engine.wconf.mask_poly_flag().ival()
           << ") must be set between 0 and 3\n\n" << flush;
      exit(1);
   }

   // Check that fcst_conv_radius and obs_conv_radius are non-negative
   if(engine.wconf.fcst_conv_radius().ival() < 0 ||
      engine.wconf.obs_conv_radius().ival() < 0) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "fcst_conv_radius (" << engine.wconf.fcst_conv_radius().ival()
           << ") and obs_conv_radius (" << engine.wconf.obs_conv_radius().ival()
           << ") must be non-negative\n\n" << flush;
      exit(1);
   }

   // Check that fcst_area_threshold and obs_area threshold are non-negative
   if(verbosity > 0 &&
      (engine.fcst_area_threshold.thresh < 0 ||
       engine.obs_area_threshold.thresh < 0)) {
      cout << "***WARNING*** check_engine_config() -> "
           << "fcst_area_threshold (" << engine.wconf.fcst_area_threshold().ival()
           << ") and obs_area_threshold (" << engine.wconf.obs_area_threshold().ival()
           << ") should be non-negative\n" << flush;
   }

   // Check that fcst_merge_flag is set between 0 and 3
   if(engine.wconf.fcst_merge_flag().ival() < 0 ||
      engine.wconf.fcst_merge_flag().ival() > 3) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "The fcst_merge_flag (" << engine.wconf.fcst_merge_flag().ival()
           << ") must be set between 0 and 3\n\n" << flush;
      exit(1);
   }

   // Check that obs_merge_flag is set between 0 and 3
   if(engine.wconf.obs_merge_flag().ival() < 0 ||
      engine.wconf.obs_merge_flag().ival() > 3) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "The obs_merge_flag (" << engine.wconf.obs_merge_flag().ival()
           << ") must be set between 0 and 3\n\n" << flush;
      exit(1);
   }

   // Check that match_flag is set between 0 and 3
   if(engine.wconf.match_flag().ival() < 0 ||
      engine.wconf.match_flag().ival() > 3) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "The match_flag (" << engine.wconf.match_flag().ival()
           << ") must be set between 0 and 3\n\n" << flush;
      exit(1);
   }

   // Check that match_flag is set between 0 and 3
   if(engine.wconf.match_flag().ival() == 0 &&
      (engine.wconf.fcst_merge_flag().ival() != 0 ||
       engine.wconf.obs_merge_flag().ival() != 0) ) {
      cout << "***WARNING*** check_engine_config() -> "
           << "when matching is disabled (match_flag = " << engine.wconf.match_flag().ival()
           << ") but merging is requested (fcst_merge_flag = " << engine.wconf.fcst_merge_flag().ival()
           << ", obs_merge_flag = " << engine.wconf.obs_merge_flag().ival()
           << ") any merging information will be discarded.\n" << flush;
   }

   // Check that max_centroid_dist is > 0
   if(verbosity > 0 &&
      engine.wconf.max_centroid_dist().dval() <= 0) {
      cout << "***WARNING*** check_engine_config() -> "
           << "max_centroid_dist (" << engine.wconf.max_centroid_dist().dval()
           << ") should be set > 0\n" << flush;
   }

   // Check that the fuzzy engine weights are non-negative
   if(engine.wconf.centroid_dist_weight().dval() < 0 ||
      engine.wconf.boundary_dist_weight().dval() < 0 ||
      engine.wconf.convex_hull_dist_weight().dval() < 0 ||
      engine.wconf.angle_diff_weight().dval() < 0 ||
      engine.wconf.area_ratio_weight().dval() < 0 ||
      engine.wconf.int_area_ratio_weight().dval() < 0 ||
      engine.wconf.complexity_ratio_weight().dval() < 0 ||
      engine.wconf.intensity_ratio_weight().dval() < 0) {

      cerr << "\n\nERROR: check_engine_config() -> "
           << "All of the fuzzy engine weights must be >= 0\n\n"
           << flush;
      exit(1);
   }

   // Check that the sum of the weights is non-zero if matching is
   // requested.
   if(engine.wconf.match_flag().ival() != 0 &&
      (engine.wconf.centroid_dist_weight().dval()
       + engine.wconf.boundary_dist_weight().dval()
       + engine.wconf.convex_hull_dist_weight().dval()
       + engine.wconf.angle_diff_weight().dval()
       + engine.wconf.area_ratio_weight().dval()
       + engine.wconf.int_area_ratio_weight().dval()
       + engine.wconf.complexity_ratio_weight().dval()
       + engine.wconf.intensity_ratio_weight().dval()) == 0) {

      cerr << "\n\nERROR: check_engine_config() -> "
           << "When matching is requested, the sum of the fuzzy engine "
           << "weights cannot be 0\n\n"
           << flush;
      exit(1);
   }

   // Check that intensity_percentile >= 0 and <= 100
   if(engine.wconf.intensity_percentile().ival() < 0 ||
      engine.wconf.intensity_percentile().ival() > 100) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "intensity_percentile (" << engine.wconf.intensity_percentile().ival()
           << ") must be >= 0 and <= 100\n\n" << flush;
      exit(1);
   }

   // Check that total_interest_threshold >= 0 and <= 1
   if(engine.wconf.total_interest_threshold().dval() < 0 ||
      engine.wconf.total_interest_threshold().dval() > 1) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "total_interest_threshold (" << engine.wconf.total_interest_threshold().dval()
           << ") must be >= 0 and <= 1\n\n" << flush;
      exit(1);
   }

   // Check that print_interest_threshold >= 0 and <= 1
   if(engine.wconf.print_interest_threshold().dval() < 0 ||
      engine.wconf.print_interest_threshold().dval() > 1) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "print_interest_threshold (" << engine.wconf.print_interest_threshold().dval()
           << ") must be >= 0 and <= 1\n\n" << flush;
      exit(1);
   }

   // Check that zero_border_size >= 1
   if(engine.wconf.zero_border_size().ival() < 1) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "zero_border_size (" << engine.wconf.zero_border_size().ival()
           << ") must be >= 1\n\n" << flush;
      exit(1);
   }

   // Get the MET Data directory from which to read the plotting data files
   replace_string("MET_BASE", MET_BASE, engine.wconf.met_data_dir().sval(), met_data_dir);

   // Check that fcst_raw_color_table can be read
   ct_test.clear();
   replace_string("MET_BASE", MET_BASE, engine.wconf.fcst_raw_color_table().sval(), tmp_str);
   if(!ct_test.read(tmp_str)) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "cannot read fcst_raw_color_table file ("
           << tmp_str
           << ")\n\n" << flush;
      exit(1);
   }

   // Check that obs_raw_color_table can be read
   ct_test.clear();
   replace_string("MET_BASE", MET_BASE, engine.wconf.obs_raw_color_table().sval(), tmp_str);
   if(!ct_test.read(tmp_str)) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "cannot read obs_raw_color_table file ("
           << tmp_str
           << ")\n\n" << flush;
      exit(1);
   }

   // Check that the stride length is set >= 1
   if((stride = engine.wconf.stride_length().ival()) < 1) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "stride_length must be set >= 1: " << stride
           << "\n\n" << flush;
      exit(1);
   }

   // Check that mode_color_table can be read
   ct_test.clear();
   replace_string("MET_BASE", MET_BASE, engine.wconf.mode_color_table().sval(), tmp_str);
   if(!ct_test.read(tmp_str)) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "cannot read mode_color_table file ("
           << tmp_str
           << ")\n\n" << flush;
      exit(1);
   }

   // Check that plot_valid_flag is set between 0 and 1
   if(engine.wconf.plot_valid_flag().ival() < 0 ||
      engine.wconf.plot_valid_flag().ival() > 1) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "The plot_valid_flag (" << engine.wconf.plot_valid_flag().ival()
           << ") must be set to 0 or 1\n\n" << flush;
      exit(1);
   }

   // Check that plot_gcarc_flag is set between 0 and 1
   if(engine.wconf.plot_gcarc_flag().ival() < 0 ||
      engine.wconf.plot_gcarc_flag().ival() > 1) {
      cerr << "\n\nERROR: check_engine_config() -> "
           << "The plot_gcarc_flag (" << engine.wconf.plot_gcarc_flag().ival()
           << ") must be set to 0 or 1\n\n" << flush;
      exit(1);
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void compute_ct_stats() {
   int i, x, y;
   WrfData fcst_mask, obs_mask;

   if(verbosity > 0) {
      cout << "Computing contingency table statistics...\n" << flush;
   }

   for(i=0; i<n_cts; i++) {

      cts[i].set_name(cts_str[i]);

      // Raw fields
      if(i == 0) {
         fcst_mask = *engine.fcst_raw;
         obs_mask  = *engine.obs_raw;

         // Apply the thresholds specified in the config file
         fcst_mask.threshold_double(engine.fcst_conv_threshold);
         obs_mask.threshold_double(engine.obs_conv_threshold);
      }
      // Filtered fields
      else if(i == 1) {
         fcst_mask = *engine.fcst_filter;
         obs_mask  = *engine.obs_filter;

         // Apply the thresholds specified in the config file
         fcst_mask.threshold_double(engine.fcst_conv_threshold);
         obs_mask.threshold_double(engine.obs_conv_threshold);
      }
      // Object fields
      else if(i == 2) {
         fcst_mask = *engine.fcst_mask;
         obs_mask  = *engine.obs_mask;
      }

      // Compute contingency table counts
      for(x=0; x<fcst_mask.get_nx(); x++) {
         for(y=0; y<fcst_mask.get_ny(); y++) {

            // Key off of the bad data values in the raw field
            if(engine.fcst_raw->is_bad_data(x, y) ||
               engine.obs_raw->is_bad_data(x, y)) continue;

            else if( fcst_mask.s_is_on(x, y) &&  obs_mask.s_is_on(x, y)) cts[i].inc_fy_oy();
            else if( fcst_mask.s_is_on(x, y) && !obs_mask.s_is_on(x, y)) cts[i].inc_fy_on();
            else if(!fcst_mask.s_is_on(x, y) &&  obs_mask.s_is_on(x, y)) cts[i].inc_fn_oy();
            else if(!fcst_mask.s_is_on(x, y) && !obs_mask.s_is_on(x, y)) cts[i].inc_fn_on();
         }
      }
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void set_plot_dims(int nx, int ny) {
   double grid_ar, sm_plot_ar, lg_plot_ar;

   grid_ar    = (double) nx/ny;
   sm_plot_ar = (sm_pane_bb.width/2.0)/(sm_pane_bb.height/3.0);
   lg_plot_ar = (lg_pane_bb.width/1.0)/(lg_pane_bb.height/2.0);

   if(grid_ar > sm_plot_ar) sm_plot_height = sm_pane_bb.height/3.0*sm_plot_ar/grid_ar;
   else                     sm_plot_height = sm_pane_bb.height/3.0;

   if(grid_ar > lg_plot_ar) lg_plot_height = lg_pane_bb.height/2.0*lg_plot_ar/grid_ar;
   else                     lg_plot_height = lg_pane_bb.height/2.0;

   h_tab_1 = sm_pane_bb.x_ll + sm_pane_bb.width/4.0;
   h_tab_2 = sm_pane_bb.x_ll + sm_pane_bb.width/4.0 * 3.0;
   h_tab_3 = sm_pane_bb.x_ur + plot_text_sep;

   v_tab_1 = sm_pane_bb.y_ur - sm_plot_height;
   v_tab_2 = v_tab_1 - sm_plot_height;
   v_tab_3 = v_tab_2 - sm_plot_height;

   return;
}

////////////////////////////////////////////////////////////////////////

void plot_engine() {
   PSfile plot;
   char ps_file[max_str_len], tmp_str[max_str_len];
   int mon, day, yr, hr, min, sec;
   int l_hr, l_min, l_sec;
   int a_hr, a_min, a_sec;

   //
   // Setup the x/y bounding box for the data to be plotted.
   // If requested in the config file, only plot the regions of valid data.
   // Otherwise, plot the entire grid.
   //
   set_xy_bb();

   //
   // Setup the lat/lon bounding box based on the x/y bounding box.
   //
   set_ll_bb();

   //
   // Setup the plotting dimensions based on the x/y bounding box
   //
   set_plot_dims(nint(xy_bb.width), nint(xy_bb.height));

   //
   // Create output PSfile name
   //
   sec_to_hms(engine.fcst_raw->get_lead_time(), l_hr, l_min, l_sec);
   unix_to_mdyhms(engine.fcst_raw->get_valid_time(),
                  mon, day, yr, hr, min, sec);
   sec_to_hms(engine.fcst_raw->get_accum_time(), a_hr, a_min, a_sec);
   sprintf(ps_file,
      "%s/mode_%s_%s_vs_%s_%s_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_%.2i%.2i%.2iA.ps",
      out_dir,
      fcst_field_name, fcst_level_name,
      obs_field_name, obs_level_name,
      l_hr, l_min, l_sec,
      yr, mon, day, hr, min, sec,
      a_hr, a_min, a_sec);

   //
   // Load the raw forecast color table
   //
   replace_string("MET_BASE", MET_BASE,
                  engine.wconf.fcst_raw_color_table().sval(), tmp_str);
   if(verbosity > 0) {
      cout << "Loading forecast raw color table: "
           << tmp_str << "\n" << flush;
   }
   fcst_ct.read(tmp_str);

   //
   // Load the raw observation color table
   //
   replace_string("MET_BASE", MET_BASE,
                  engine.wconf.obs_raw_color_table().sval(), tmp_str);
   if(verbosity > 0) {
      cout << "Loading observation raw color table: "
           << tmp_str << "\n" << flush;
   }
   obs_ct.read(tmp_str);

   //
   // If the forecast and observation fields are the same and if the range
   // of both colortables is [0, 1], rescale both colortables to the
   // data_min and data_max values
   //
   if(fcst_gc == obs_gc &&
      fcst_ct.data_min(bad_data_double) == 0.0 &&
      fcst_ct.data_max(bad_data_double) == 1.0 &&
      obs_ct.data_min(bad_data_double) == 0.0 &&
      obs_ct.data_max(bad_data_double) == 1.0) {

      fcst_ct.rescale(data_min, data_max, bad_data_double);
      obs_ct.rescale(data_min, data_max, bad_data_double);
   }
   //
   // Otherwise, if the range of either colortable is [0, 1], rescale
   // the field using the min/max values in the field
   //
   else {
      if(fcst_ct.data_min(bad_data_double) == 0.0 &&
         fcst_ct.data_max(bad_data_double) == 1.0) {

         fcst_ct.rescale(engine.fcst_raw->int_to_double(0),
                         engine.fcst_raw->int_to_double(wrfdata_int_data_max),
                         bad_data_double);
      }
      if(obs_ct.data_min(bad_data_double) == 0.0 &&
         obs_ct.data_max(bad_data_double) == 1.0) {

         obs_ct.rescale(engine.obs_raw->int_to_double(0),
                        engine.obs_raw->int_to_double(wrfdata_int_data_max),
                        bad_data_double);
      }
   }

   //
   // If the fcst_raw_plot_min or fcst_raw_plot_max value is set in the
   // config file, rescale the forecast colortable to the requested range
   //
   if(engine.wconf.fcst_raw_plot_min().dval() != 0.0 ||
      engine.wconf.fcst_raw_plot_max().dval() != 0.0) {
      fcst_ct.rescale(engine.wconf.fcst_raw_plot_min().dval(),
                      engine.wconf.fcst_raw_plot_max().dval(),
                      bad_data_double);
   }

   //
   // If the obs_raw_plot_min or obs_raw_plot_max value is set in the
   // config file, rescale the observation colortable to the requested range
   //
   if(engine.wconf.obs_raw_plot_min().dval() != 0.0 ||
      engine.wconf.obs_raw_plot_max().dval() != 0.0) {
      obs_ct.rescale(engine.wconf.obs_raw_plot_min().dval(),
                     engine.wconf.obs_raw_plot_max().dval(),
                     bad_data_double);
   }

   //
   // Set the fill colors.  If a fill value is not specified in the range
   // of the color table, use the default color.  Otherwise, use the
   // color specified in the color table.
   //
   if(bad_data_double >= fcst_ct.data_min() &&
      bad_data_double <= fcst_ct.data_max()) c_fcst_fill = fcst_ct.nearest(bad_data_double);
   if(bad_data_double >= obs_ct.data_min() &&
      bad_data_double <= obs_ct.data_max())  c_obs_fill = obs_ct.nearest(bad_data_double);

   if(verbosity > 0) {
      cout << "Creating postscript file: " << ps_file << "\n" << flush;
   }

   plot.open(ps_file);
   n_page = 1;

   sprintf(tmp_str, "MODE: %s at %s vs %s at %s",
           fcst_field_name, fcst_level_name,
           obs_field_name, obs_level_name);
   plot_engine(plot, engine, tmp_str, 2);

   if(engine.wconf.fcst_merge_flag().ival() == 1 ||
      engine.wconf.fcst_merge_flag().ival() == 3 ) {
      plot_threshold_merging(plot, engine, "Forecast: Threshold Merging", 1);
   }

   if(engine.wconf.fcst_merge_flag().ival() == 2 ||
      engine.wconf.fcst_merge_flag().ival() == 3) {
      plot_engine(plot, *engine.fcst_engine, "Forecast: Engine Merging", 1);
   }

   if(engine.wconf.obs_merge_flag().ival() == 1 ||
      engine.wconf.obs_merge_flag().ival() == 3) {
      plot_threshold_merging(plot, engine, "Observation: Threshold Merging", 0);
   }

   if(engine.wconf.obs_merge_flag().ival() == 2 ||
      engine.wconf.obs_merge_flag().ival() == 3) {
      plot_engine(plot, *engine.obs_engine, "Observation: Engine Merging", 0);
   }

   return;
}

////////////////////////////////////////////////////////////////////////
//
// Flag value of 2 indicates that both fcst and obs should be enlarged
// Flag value of 1 indicates that only the fcst should be enlarged
// Flag value of 0 indicates that only the obs should be enlarged
//
////////////////////////////////////////////////////////////////////////

void plot_engine(PSfile &p, Engine &eng, const char *title, int flag) {
   BoundingBox dim;
   char label[max_str_len];
   char fcst_str[max_str_len], fcst_short_str[max_str_len];
   char obs_str[max_str_len], obs_short_str[max_str_len];
   char tmp_str[max_str_len], thresh_str[max_str_len];
   double v_tab, h_tab_a, h_tab_b, h_tab_c;
   bool draw_line;
   int i, mon, day, yr, hr, min, sec;

   ////////////////////////////////////////////////////////////////////////////
   //
   // First Page: create a 6 plot
   //
   ////////////////////////////////////////////////////////////////////////////

   p.pagenumber(n_page);

   p.choose_font(31, 24.0, met_data_dir);
   p.write_centered_text(1, 1, 306.0, 752.0, 0.5, 0.5, title);

   if(flag == 2) {
      strcpy(fcst_str, "Forecast");
      strcpy(fcst_short_str, "Fcst");
      strcpy(obs_str, "Observation");
      strcpy(obs_short_str, "Obs");
   }
   else if(flag == 1) {
      strcpy(fcst_str, "Forecast");
      strcpy(fcst_short_str, "Fcst");
      strcpy(obs_str, "Forecast");
      strcpy(obs_short_str, "Fcst");
   }
   else if(flag == 0) {
      strcpy(fcst_str, "Observation");
      strcpy(fcst_short_str, "Obs");
      strcpy(obs_str, "Observation");
      strcpy(obs_short_str, "Obs");
   }

   p.choose_font(31, 18.0, met_data_dir);
   p.write_centered_text(1, 1, h_tab_1, 727.0, 0.5, 0.5, fcst_str);
   p.write_centered_text(1, 1, h_tab_2, 727.0, 0.5, 0.5, obs_str);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Draw raw forecast field
   //
   ////////////////////////////////////////////////////////////////////////////

   set_dim(dim, v_tab_1, v_tab_1 + sm_plot_height, h_tab_1);
   render_image(p, eng, *eng.fcst_raw, dim, 1, 0);
   draw_border(p, dim);
   draw_map(p, dim);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Draw raw observation field
   //
   ////////////////////////////////////////////////////////////////////////////

   set_dim(dim, v_tab_1, v_tab_1 + sm_plot_height, h_tab_2);
   render_image(p, eng, *eng.obs_raw, dim, 0, 0);
   draw_border(p, dim);
   draw_map(p, dim);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Draw fcst split field
   //
   ////////////////////////////////////////////////////////////////////////////

   set_dim(dim, v_tab_2, v_tab_2 + sm_plot_height, h_tab_1);
   render_image(p, eng, *eng.fcst_split, dim, 1, 1);
   draw_border(p, dim);
   draw_map(p, dim);
   draw_convex_hulls(p, eng, dim, 1);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Draw obs split field
   //
   ////////////////////////////////////////////////////////////////////////////

   set_dim(dim, v_tab_2, v_tab_2 + sm_plot_height, h_tab_2);
   render_image(p, eng, *eng.obs_split, dim, 0, 1);
   draw_border(p, dim);
   draw_map(p, dim);
   draw_convex_hulls(p, eng, dim, 0);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Draw fcst shape numbers
   //
   ////////////////////////////////////////////////////////////////////////////

   set_dim(dim, v_tab_3, v_tab_3 + sm_plot_height, h_tab_1);
   draw_border(p, dim);
   draw_map(p, dim);
   plot_shape_numbers(p, eng, dim, 1);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Draw obs shape numbers
   //
   ////////////////////////////////////////////////////////////////////////////

   set_dim(dim, v_tab_3, v_tab_3 + sm_plot_height, h_tab_2);
   draw_border(p, dim);
   draw_map(p, dim);
   plot_shape_numbers(p, eng, dim, 0);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Plot interest values
   //
   ////////////////////////////////////////////////////////////////////////////

   p.choose_font(31, 12.0, met_data_dir);

   p.write_centered_text(1, 1, h_tab_3 + 1.5*plot_text_sep, 727.0, 0.5, 0.5, fcst_short_str);
   p.write_centered_text(1, 1, h_tab_3 + 4.5*plot_text_sep, 727.0, 0.5, 0.5, obs_short_str);
   p.write_centered_text(1, 1, h_tab_3 + 7.5*plot_text_sep, 727.0, 0.5, 0.5, "Interest");

   v_tab = 727.0 - 1.5*plot_text_sep;
   draw_line = false;

   for(i=0; i<(eng.n_fcst*eng.n_obs) && v_tab >= v_margin; i++) {

      if(eng.info[i].interest_value < eng.wconf.total_interest_threshold().dval()
      && draw_line == false) {

         sprintf(label, "%s", "----------------------------------");
         p.write_centered_text(1, 1, h_tab_3 + 4.5*plot_text_sep,
                               v_tab, 0.5, 0.5, label);
         draw_line = true;
         v_tab -= plot_text_sep;
      }

      sprintf(label, "%i", eng.info[i].fcst_number);
      p.write_centered_text(1, 1, h_tab_3 + 1.5*plot_text_sep,
                            v_tab, 0.5, 0.5, label);

      sprintf(label, "%i", eng.info[i].obs_number);
      p.write_centered_text(1, 1, h_tab_3 + 4.5*plot_text_sep,
                            v_tab, 0.5, 0.5, label);

      if(eng.info[i].interest_value < 0) sprintf(label, "N/A");
      else sprintf(label, "%.4f", eng.info[i].interest_value);
      p.write_centered_text(1, 1, h_tab_3 + 7.5*plot_text_sep,
                            v_tab, 0.5, 0.5, label);

      v_tab -= plot_text_sep;
   }

   ////////////////////////////////////////////////////////////////////////////
   //
   // Model Name, Initialization, Valid, Lead, and Accumulation Times
   //
   ////////////////////////////////////////////////////////////////////////////

   v_tab = v_tab_3 - 1.0*plot_text_sep;
   h_tab_a = h_tab_1 - 0.5*dim.width;
   h_tab_b = h_tab_a + 5.0*plot_text_sep;
   h_tab_c = h_tab_a + 10.0*plot_text_sep;

   //
   // Model Name
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Model Name:");
   p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, eng.wconf.model().sval());
   v_tab -= plot_text_sep;

   //
   // Initialization Time
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Init Time:");
   unix_to_mdyhms(eng.fcst_raw->get_valid_time() - eng.fcst_raw->get_lead_time(),
                  mon, day, yr, hr, min, sec);
   sprintf(label, "%s %i, %i %.2i:%.2i:%.2i",
           short_month_name[mon], day, yr, hr, min, sec);
   p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Valid time
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Valid Time:");
   unix_to_mdyhms(eng.fcst_raw->get_valid_time(),
                  mon, day, yr, hr, min, sec);
   sprintf(label, "%s %i, %i %.2i:%.2i:%.2i",
           short_month_name[mon], day, yr, hr, min, sec);
   p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Lead time
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Lead Time:");
   sec_to_hms(eng.fcst_raw->get_lead_time(), hr, min, sec);
   sprintf(label, "%.2i:%.2i:%.2i", hr, min, sec);
   p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Accumulation time
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Accum Time:");
   sec_to_hms(eng.fcst_raw->get_accum_time(), hr, min, sec);
   sprintf(label, "%.2i:%.2i:%.2i", hr, min, sec);
   p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   ////////////////////////////////////////////////////////////////////////////
   //
   // Engine Weights
   //
   ////////////////////////////////////////////////////////////////////////////

   //
   // Centroid Distance Weight
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Centroid Distance:");
   sprintf(label, "%.2f", eng.wconf.centroid_dist_weight().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Boundary Distance Weight
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Boundary Distance:");
   sprintf(label, "%.2f", eng.wconf.boundary_dist_weight().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Convex Hull Distance Weight
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Convex Hull Distance:");
   sprintf(label, "%.2f", eng.wconf.convex_hull_dist_weight().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Angle Difference Weight
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Angle Difference:");
   sprintf(label, "%.2f", eng.wconf.angle_diff_weight().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Area Ratio Weight
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Area Ratio:");
   sprintf(label, "%.2f", eng.wconf.area_ratio_weight().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Intersection Over Minimum Area Weight
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Intersection/Area:");
   sprintf(label, "%.2f", eng.wconf.int_area_ratio_weight().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Complexity Ratio Weight
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Complexity Ratio:");
   sprintf(label, "%.2f", eng.wconf.complexity_ratio_weight().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Intensity Ratio Weight
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Intensity Ratio:");
   sprintf(label, "%.2f", eng.wconf.intensity_ratio_weight().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Total Interest Threshold
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Total Interest Thresh:");
   sprintf(label, "%.2f", eng.wconf.total_interest_threshold().dval());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   ////////////////////////////////////////////////////////////////////////////
   //
   // Engine configuration
   //
   ////////////////////////////////////////////////////////////////////////////

   v_tab = v_tab_3 - 1.0*plot_text_sep;
   h_tab_a = h_tab_2 - 0.5*dim.width;
   h_tab_b = h_tab_a + 5.0*plot_text_sep;
   h_tab_c = h_tab_a + 10.0*plot_text_sep;

   if(flag == 2) {

      //
      // Field name
      //
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, fcst_str);
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, obs_str);
      v_tab -= plot_text_sep;

      //
      // Mask bad data flag
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Mask Missing:");
      if(eng.wconf.mask_missing_flag().ival() == 1 ||
         eng.wconf.mask_missing_flag().ival() == 3)    sprintf(label, "on");
      else                                             sprintf(label, "off");
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);

      if(eng.wconf.mask_missing_flag().ival() == 2 ||
         eng.wconf.mask_missing_flag().ival() == 3)    sprintf(label, "on");
      else                                             sprintf(label, "off");
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Mask Grid Flag and Mask Polygon Flags
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Mask Gr/Ply:");
      if(eng.wconf.mask_grid_flag().ival() == 1 ||
         eng.wconf.mask_grid_flag().ival() == 3)       sprintf(tmp_str, "on");
      else                                             sprintf(tmp_str, "off");
      if(eng.wconf.mask_poly_flag().ival() == 1 ||
         eng.wconf.mask_poly_flag().ival() == 3)       sprintf(label, "%s/on", tmp_str);
      else                                             sprintf(label, "%s/off", tmp_str);
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);

      if(eng.wconf.mask_grid_flag().ival() == 2 ||
         eng.wconf.mask_grid_flag().ival() == 3)       sprintf(tmp_str, "on");
      else                                             sprintf(tmp_str, "off");
      if(eng.wconf.mask_poly_flag().ival() == 2 ||
         eng.wconf.mask_poly_flag().ival() == 3)       sprintf(label, "%s/on", tmp_str);
      else                                             sprintf(label, "%s/off", tmp_str);
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Raw threshold
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Raw Thresh:");
      eng.fcst_raw_threshold.get_str(thresh_str);
      sprintf(label, "%s %s", thresh_str, fcst_unit_name);
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
      eng.obs_raw_threshold.get_str(thresh_str);
      sprintf(label, "%s %s", thresh_str, obs_unit_name);
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Convolution Radius
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Conv Radius:");
      sprintf(label, "%.0i gs", eng.wconf.fcst_conv_radius().ival());
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
      sprintf(label, "%.0i gs", eng.wconf.obs_conv_radius().ival());
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Convolution Threshold
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Conv Thresh:");
      eng.fcst_conv_threshold.get_str(thresh_str);
      sprintf(label, "%s %s", thresh_str, fcst_unit_name);
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
      eng.obs_conv_threshold.get_str(thresh_str);
      sprintf(label, "%s %s", thresh_str, obs_unit_name);
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Area Threshold
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Area Thresh:");
      eng.fcst_area_threshold.get_str(thresh_str);
      sprintf(label, "%s gs", thresh_str);
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
      eng.obs_area_threshold.get_str(thresh_str);
      sprintf(label, "%s gs", thresh_str);
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Intensity Percentile, check for the mean (101) and sum (102)
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab,
                           0.0, 0.5, "Inten Perc:");
      if(nint(eng.wconf.fcst_inten_perc().ival()) == 101)      sprintf(label, "mean");
      else if(nint(eng.wconf.fcst_inten_perc().ival()) == 102) sprintf(label, "sum");
      else {
         sprintf(label, "%.0i th", eng.wconf.fcst_inten_perc().ival());
      }
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);

      if(nint(eng.wconf.obs_inten_perc().ival()) == 101)      sprintf(label, "mean");
      else if(nint(eng.wconf.obs_inten_perc().ival()) == 102) sprintf(label, "sum");
      else {
         sprintf(label, "%.0i th", eng.wconf.obs_inten_perc().ival());
      }
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Intensity Threshold
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab,
                           0.0, 0.5, "Inten Thresh:");
      eng.fcst_inten_perc_threshold.get_str(thresh_str);
      sprintf(label, "%s %s", thresh_str, fcst_unit_name);
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
      eng.obs_inten_perc_threshold.get_str(thresh_str);
      sprintf(label, "%s %s", thresh_str, obs_unit_name);
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Merge Threshold
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Merge Thresh:");
      eng.fcst_merge_threshold.get_str(thresh_str);
      sprintf(label, "%s %s", thresh_str, fcst_unit_name);
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
      eng.obs_merge_threshold.get_str(thresh_str);
      sprintf(label, "%s %s", thresh_str, obs_unit_name);
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      ////////////////////////////////////////////////////////////////////////////
      //
      // Matching/Merging Criteria
      //
      ////////////////////////////////////////////////////////////////////////////

      //
      // Merging flag
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Merging:");
      if(eng.wconf.fcst_merge_flag().ival() == 0)       sprintf(label, "none");
      else if(eng.wconf.fcst_merge_flag().ival() == 1)  sprintf(label, "thresh");
      else if(eng.wconf.fcst_merge_flag().ival() == 2)  sprintf(label, "engine");
      else if(eng.wconf.fcst_merge_flag().ival() == 3)  sprintf(label, "thresh/engine");
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);

      if(eng.wconf.obs_merge_flag().ival() == 0)        sprintf(label, "none");
      else if(eng.wconf.obs_merge_flag().ival() == 1)   sprintf(label, "thresh");
      else if(eng.wconf.obs_merge_flag().ival() == 2)   sprintf(label, "engine");
      else if(eng.wconf.obs_merge_flag().ival() == 3)   sprintf(label, "thresh/engine");
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;

      //
      // Matching scheme
      //
      p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Matching:");
      if(eng.wconf.match_flag().ival() == 0)      sprintf(label, "none");
      else if(eng.wconf.match_flag().ival() == 1) sprintf(label, "match/merge");
      else if(eng.wconf.match_flag().ival() == 2) sprintf(label, "match/fcst merge");
      else if(eng.wconf.match_flag().ival() == 3) sprintf(label, "match/no merge");
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
      v_tab -= plot_text_sep;
   }
   else {
      v_tab -= 11.0*plot_text_sep;

      //
      // Field name
      //
      p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, fcst_str);
      p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, obs_str);
      v_tab -= plot_text_sep;
   }

   ////////////////////////////////////////////////////////////////////////////
   //
   // Object Counts
   //
   ////////////////////////////////////////////////////////////////////////////

   //
   // Simple objects (Matched Simples/Unmatched Simples)
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Simple(M/U):");
   sprintf(label, "%i (%i/%i)", eng.n_fcst, eng.n_fcst - eng.get_unmatched_fcst(),
           eng.get_unmatched_fcst());
   p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
   sprintf(label, "%i (%i/%i)", eng.n_obs, eng.n_obs - eng.get_unmatched_obs(),
           eng.get_unmatched_obs());
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   //
   // Composite objects
   //
   p.write_centered_text(1, 1, h_tab_a, v_tab, 0.0, 0.5, "Composites:");
   sprintf(label, "%i", eng.collection.n_sets);
   p.write_centered_text(1, 1, h_tab_b, v_tab, 0.0, 0.5, label);
   sprintf(label, "%i", eng.collection.n_sets);
   p.write_centered_text(1, 1, h_tab_c, v_tab, 0.0, 0.5, label);
   v_tab -= plot_text_sep;

   ////////////////////////////////////////////////////////////////////////////
   //
   // Finished with this page
   //
   ////////////////////////////////////////////////////////////////////////////

   p.showpage();
   n_page++;

   if(flag == 2 || flag == 1) {

      ////////////////////////////////////////////////////////////////////////////
      //
      // Second Page: englargement of forecast fields
      //
      ////////////////////////////////////////////////////////////////////////////

      p.pagenumber(n_page);

      p.choose_font(31, 24.0, met_data_dir);
      p.write_centered_text(1, 1, h_tab_cen, 752.0, 0.5, 0.5, title);
      p.write_centered_text(1, 1, h_tab_cen, 722.0, 0.5, 0.5, fcst_str);

      ////////////////////////////////////////////////////////////////////////////
      //
      // Draw raw forecast field
      //
      ////////////////////////////////////////////////////////////////////////////

      v_tab = page_height - 4.0*v_margin;
      set_dim(dim, v_tab - lg_plot_height, v_tab, h_tab_cen);
      render_image(p, eng, *eng.fcst_raw, dim, 1, 0);
      draw_border(p, dim);
      draw_map(p, dim);
      plot_colorbar(p, dim, 1);

      ////////////////////////////////////////////////////////////////////////////
      //
      // Draw fcst split field
      //
      ////////////////////////////////////////////////////////////////////////////

      v_tab -= lg_plot_height;
      set_dim(dim, v_tab - lg_plot_height, v_tab, h_tab_cen);
      render_image(p, eng, *eng.fcst_split, dim, 1, 1);
      draw_border(p, dim);
      draw_map(p, dim);
      draw_convex_hulls(p, eng, dim, 1);

      ////////////////////////////////////////////////////////////////////////////
      //
      // Finished with this page
      //
      ////////////////////////////////////////////////////////////////////////////

      p.showpage();
      n_page++;
   }

   if(flag == 2 || flag == 0) {

      ////////////////////////////////////////////////////////////////////////////
      //
      // Third Page: englargement of observation fields
      //
      ////////////////////////////////////////////////////////////////////////////

      p.pagenumber(n_page);

      p.choose_font(31, 24.0, met_data_dir);
      p.write_centered_text(1, 1, h_tab_cen, 752.0, 0.5, 0.5, title);
      p.write_centered_text(1, 1, h_tab_cen, 722.0, 0.5, 0.5, obs_str);


      ////////////////////////////////////////////////////////////////////////////
      //
      // Draw raw observation field
      //
      ////////////////////////////////////////////////////////////////////////////

      v_tab = page_height - 4.0*v_margin;
      set_dim(dim, v_tab - lg_plot_height, v_tab, h_tab_cen);
      render_image(p, eng, *eng.obs_raw, dim, 0, 0);
      draw_border(p, dim);
      draw_map(p, dim);
      plot_colorbar(p, dim, 0);

      ////////////////////////////////////////////////////////////////////////////
      //
      // Draw observation split field
      //
      ////////////////////////////////////////////////////////////////////////////

      v_tab -= lg_plot_height;
      set_dim(dim, v_tab - lg_plot_height, v_tab, h_tab_cen);
      render_image(p, eng, *eng.obs_split, dim, 0, 1);
      draw_border(p, dim);
      draw_map(p, dim);
      draw_convex_hulls(p, eng, dim, 0);

      ////////////////////////////////////////////////////////////////////////////
      //
      // Finished with this page
      //
      ////////////////////////////////////////////////////////////////////////////

      p.showpage();
      n_page++;
   }

   if(flag == 2) {

      ////////////////////////////////////////////////////////////////////////////
      //
      // Fourth Page: overlap of object fields
      //
      ////////////////////////////////////////////////////////////////////////////

      p.pagenumber(n_page);

      ////////////////////////////////////////////////////////////////////////////
      //
      // Draw the forecast object field and observation boundaries
      //
      ////////////////////////////////////////////////////////////////////////////

      v_tab = page_height - 2.0*v_margin;
      set_dim(dim, v_tab - lg_plot_height, v_tab, h_tab_cen);

      sprintf(tmp_str, "%s Objects with %s Outlines", fcst_str, obs_str);
      p.choose_font(31, 24.0, met_data_dir);
      p.write_centered_text(1, 1, h_tab_cen, dim.y_ur + plot_text_sep/2.0,
                            0.5, 0.5, tmp_str);

      render_image(p, eng, *eng.fcst_split, dim, 1, 1);
      draw_border(p, dim);
      draw_map(p, dim);
      draw_boundaries(p, *eng.obs_split, eng.n_obs, dim);

      ////////////////////////////////////////////////////////////////////////////
      //
      // Draw the observation object field and forecast boundaries
      //
      ////////////////////////////////////////////////////////////////////////////

      v_tab = v_tab - lg_plot_height - 2.0*v_margin;
      set_dim(dim, v_tab - lg_plot_height, v_tab, h_tab_cen);

      sprintf(tmp_str, "%s Objects with %s Outlines", obs_str, fcst_str);
      p.choose_font(31, 24.0, met_data_dir);
      p.write_centered_text(1, 1, h_tab_cen, dim.y_ur + plot_text_sep/2.0,
                            0.5, 0.5, tmp_str);

      render_image(p, eng, *eng.obs_split, dim, 0, 1);
      draw_border(p, dim);
      draw_map(p, dim);
      draw_boundaries(p, *eng.fcst_split, eng.n_fcst, dim);

      ////////////////////////////////////////////////////////////////////////////
      //
      // Finished with this page
      //
      ////////////////////////////////////////////////////////////////////////////

      p.showpage();
      n_page++;
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void plot_threshold_merging(PSfile &p, Engine &eng, const char *title, int flag) {
   BoundingBox dim;
   WrfData merge_mask, merge_split, merge_shape;
   Polyline poly;
   int n_merge;
   double v_tab;

   if(flag == 1) {
      merge_mask = *(eng.fcst_conv);
      merge_mask.threshold_double(eng.fcst_merge_threshold);
   }
   else if(flag == 0) {
      merge_mask = *(eng.obs_conv);
      merge_mask.threshold_double(eng.obs_merge_threshold);
   }
   merge_split = split(merge_mask, n_merge);

   p.pagenumber(n_page);

   p.choose_font(31, 24.0, met_data_dir);
   p.write_centered_text(1, 1, h_tab_cen, 752.0, 0.5, 0.5, title);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Draw raw field
   //
   ////////////////////////////////////////////////////////////////////////////

   v_tab = page_height - 3.0*v_margin;
   set_dim(dim, v_tab - lg_plot_height, v_tab, h_tab_cen);
   if(flag == 1) render_image(p, eng, *eng.fcst_raw, dim, 1, 0);
   if(flag == 0) render_image(p, eng, *eng.obs_raw, dim, 0, 0);
   draw_border(p, dim);
   draw_map(p, dim);
   plot_colorbar(p, dim, flag);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Draw split field
   //
   ////////////////////////////////////////////////////////////////////////////

   v_tab -= lg_plot_height;
   set_dim(dim, v_tab - lg_plot_height, v_tab, h_tab_cen);
   if(flag == 1) render_image(p, eng, *eng.fcst_split, dim, 1, 2);
   if(flag == 0) render_image(p, eng, *eng.obs_split, dim, 0, 2);
   draw_border(p, dim);
   draw_map(p, dim);
   draw_boundaries(p, merge_split, n_merge, dim);

   ////////////////////////////////////////////////////////////////////////////
   //
   // Finished with this page
   //
   ////////////////////////////////////////////////////////////////////////////

   p.showpage();
   n_page++;

   return;
}

////////////////////////////////////////////////////////////////////////

void draw_border(PSfile &p, BoundingBox &dim) {

   p.gsave();
   p.setlinewidth(l_width);
   p.newpath();
   p.moveto(dim.x_ll, dim.y_ll);
   p.lineto(dim.x_ur, dim.y_ll);
   p.lineto(dim.x_ur, dim.y_ur);
   p.lineto(dim.x_ll, dim.y_ur);
   p.closepath();
   p.stroke();
   p.grestore();

   return;
}

////////////////////////////////////////////////////////////////////////

void draw_map(PSfile &p, BoundingBox &dim) {

   p.gsave();
   p.setlinewidth(l_width);
   draw_world(grid, xy_bb, p, ll_bb, dim, c_map, met_data_dir);
   draw_states(grid, xy_bb, p, ll_bb, dim, c_map, met_data_dir);
   p.grestore();

   return;
}

////////////////////////////////////////////////////////////////////////

void render_image(PSfile &p, Engine &eng, const WrfData &wd,
                  BoundingBox &dim, int fcst, int split) {
   RenderInfo render_info;
   Ppm ppm_image;
   int x, y, v_int;
   double mag;
   Color c;
   Color *c_fill_ptr;
   ColorTable *ct_ptr;

   //
   // Set up pointers to the appropriate colortable and fill color values
   //
   if(fcst == 1) {
      ct_ptr = &fcst_ct;
      c_fill_ptr = &c_fcst_fill;
   }
   else {
      ct_ptr = &obs_ct;
      c_fill_ptr = &c_obs_fill;
   }

   //
   // Convert the WrfData object to PPM
   //
   ppm_image.set_size_xy((int) xy_bb.width, (int) xy_bb.height);

   for(x=nint(xy_bb.x_ll); x<xy_bb.x_ur; x++) {
      for(y=nint(xy_bb.y_ll); y<xy_bb.y_ur; y++) {

         v_int = wd.get_xy_int(x, y);

         //
         // Single object field
         //
         if(split == 1) {
            if(v_int == bad_data_flag) { // Should be no bad data left at this point
               c = *c_fill_ptr;
            }
            else if(v_int == 0) {        // Set to white
               c.set_rgb(255, 255, 255);
            }
            // Single fcst object field
            else if(v_int > 0 && fcst) {
               c = eng.fcst_color[v_int-1];
            }
            // Single obs object field
            else if(v_int > 0 && !fcst) {
               c = eng.obs_color[v_int-1];
            }
         }
         //
         // Composite object field
         //
         else if(split == 2) {
            if(v_int == bad_data_flag) { // Should be no bad data left at this point
               c = *c_fill_ptr;
            }
            else if(v_int == 0) {        // Set to white
               c.set_rgb(255, 255, 255);
            }
            else {
               c = c_bndy;
            }
         }
         //
         // Raw data field
         //
         else {
            if(v_int == bad_data_flag) {
               c = *c_fill_ptr;
            }
            else {
               c = ct_ptr->nearest(wd.get_xy_double(x, y));
            }
         }

         ppm_image.putxy(c, x-nint(xy_bb.x_ll), y-nint(xy_bb.y_ll));
      } // end for y
   } // end for x

   mag = (dim.x_ur - dim.x_ll)/xy_bb.width;

   render_info.x_ll = dim.x_ll;
   render_info.y_ll = dim.y_ll;
   render_info.magnification = mag;
   render_info.bw = 0;
   render_info.add_filter(RunLengthEncode);
   render_info.add_filter(ASCII85Encode);
   render(p, ppm_image, render_info);

   return;
}

////////////////////////////////////////////////////////////////////////

void plot_shape_numbers(PSfile &p, Engine &eng, BoundingBox &dim, int fcst) {
   int i, n_shapes;
   double grid_x, grid_y, page_x, page_y;
   Color c;
   char label[max_str_len];

   p.gsave();

   if(fcst) n_shapes = eng.n_fcst;
   else     n_shapes = eng.n_obs;

   p.setlinewidth(0.0);

   for(i=0; i<n_shapes; i++) {

      if(fcst) {
         grid_x = eng.fcst_single[i].xbar;
         grid_y = eng.fcst_single[i].ybar;
      }
      else {
         grid_x = eng.obs_single[i].xbar;
         grid_y = eng.obs_single[i].ybar;
      }

      gridxy_to_pagexy(grid, xy_bb, grid_x, grid_y, page_x, page_y, dim);

      if(fcst) c = eng.fcst_color[i];
      else     c = eng.obs_color[i];

      p.setrgbcolor(c.red()/255.0, c.green()/255.0, c.blue()/255.0);
      p.choose_font(28, 16.0, met_data_dir);
      sprintf(label, "%i", i+1);
      p.write_centered_text(2, 1, page_x, page_y, 0.5, 0.5, label);

      // Draw outline in black
      p.setrgbcolor(0.0, 0.0, 0.0);
      p.write_centered_text(2, 0, page_x, page_y, 0.5, 0.5, label);
   }

   p.grestore();

   return;
}

///////////////////////////////////////////////////////////////////////////////

void draw_convex_hulls(PSfile &p, Engine &eng, const BoundingBox &dim, int fcst) {
   int i;
   Polyline poly;
   WrfData comp;

   //
   // Draw convex hulls around collection of shapes
   //
   for(i=0; i<eng.collection.n_sets; i++) {

      composite_mask(comp, eng, i, fcst);
      poly = comp.convex_hull();
      draw_polyline(p, poly, c_hull, dim, false);
   }

   return;
}

///////////////////////////////////////////////////////////////////////////////

void draw_boundaries(PSfile &p, WrfData &split_wd, int n_shapes, const BoundingBox &dim) {
   int i;
   Polyline poly;
   WrfData wd_shape;

   //
   // Draw boundary for each shape in the split field
   //
   for(i=0; i<n_shapes; i++) {
      wd_shape = select(split_wd, i+1);
      poly = wd_shape.single_boundary();
      draw_polyline(p, poly, c_bndy, dim, false);
   }

   return;
}

///////////////////////////////////////////////////////////////////////////////

void draw_polyline(PSfile &p, Polyline &poly, const Color &c,
                   BoundingBox dim, bool latlon) {
   int i;
   double lat, lon;
   double grid_x_prev, grid_y_prev, grid_x_cur, grid_y_cur;
   double page_x, page_y;

   // Nothing to draw for empty polylines
   if(poly.n_points <= 0) {

      return;
   }

   // Set to specified values
   p.gsave();
   p.setlinewidth(l_width);
   p.setrgbcolor(c.red()/255.0, c.green()/255.0, c.blue()/255.0);
   p.newpath();

   //
   // Set the starting Grid x,y to the first vertex of
   // the polyline
   //
   if(latlon) {
      lat = poly.u[0];
      lon = poly.v[0];
      grid.latlon_to_xy(lat, lon, grid_x_prev, grid_y_prev);
   }
   // Grid coordinates
   else {
      grid_x_prev = poly.u[0];
      grid_y_prev = poly.v[0];
   }

   // Convert the starting Grid x/y to page x,y and move to it
   gridxy_to_pagexy(grid, xy_bb, grid_x_prev, grid_y_prev, page_x, page_y, dim);
   p.moveto(page_x, page_y);

   // Loop through the vertices and back to the starting vertex
   for(i=1; i<poly.n_points; i++) {

      //
      // Draw an arc between the previous Grid x,y point and the
      // current point.
      //
      if(latlon) {
         lat = poly.u[i];
         lon = poly.v[i];
         grid.latlon_to_xy(lat, lon, grid_x_cur, grid_y_cur);
      }
      // Grid coordinates
      else {
         grid_x_cur = poly.u[i];
         grid_y_cur = poly.v[i];
      }

      gridxy_to_pagexy(grid, xy_bb, grid_x_cur, grid_y_cur, page_x, page_y, dim);

      // Plot polylines as straight lines in the grid
      if(engine.wconf.plot_gcarc_flag().ival() == 0) {
         p.lineto(page_x, page_y);
      }
      // Plot polylines using great circle arcs
      else {
         gc_arcto(grid, xy_bb, p, grid_x_prev, grid_y_prev,
                  grid_x_cur, grid_y_cur, 10, dim);
      }

      // Reset the previous Grid x,y to the current Grid x,y
      grid_x_prev = grid_x_cur;
      grid_y_prev = grid_y_cur;
   }

   p.closepath();
   // Stroke the path
   p.stroke();

   p.grestore();

   return;
}

////////////////////////////////////////////////////////////////////////

void plot_colorbar(PSfile &p, BoundingBox &dim, int fcst) {
   int i, n_colors;
   char label[max_str_len];
   double bar_width, bar_height, x_ll, y_ll;
   ColorTable *ct_ptr;
   Color c;

   //
   // Set up the pointer to the appropriate colortable
   //
   if(fcst == 1) ct_ptr = &fcst_ct;
   else          ct_ptr = &obs_ct;

   //
   // Get the number of colortable entries
   //
   n_colors = ct_ptr->n_entries();

   //
   // Draw colorbar in the bottom right corner of the Bounding Box
   //
   p.gsave();
   p.setlinewidth(l_width);
   p.choose_font(28, 8.0, met_data_dir);

   bar_width = h_margin/2.0;
   bar_height = (dim.y_ur - dim.y_ll)/n_colors;

   x_ll = dim.x_ur;
   y_ll = dim.y_ll;

   for(i=0; i<n_colors; i++) {

     c = (*ct_ptr)[i].color();

     //
     // Color box
     //
     p.setrgbcolor(c.red()/255.0, c.green()/255.0, c.blue()/255.0);
     p.newpath();
     p.moveto(x_ll,             y_ll);
     p.lineto(x_ll,             y_ll + bar_height);
     p.lineto(x_ll + bar_width, y_ll + bar_height);
     p.lineto(x_ll + bar_width, y_ll);
     p.closepath();
     p.fill();

     //
     // Outline color box
     //
     p.setrgbcolor(0.0, 0.0, 0.0);
     p.newpath();
     p.moveto(x_ll,             y_ll);
     p.lineto(x_ll,             y_ll + bar_height);
     p.lineto(x_ll + bar_width, y_ll + bar_height);
     p.lineto(x_ll + bar_width, y_ll);
     p.closepath();
     p.stroke();

     //
     // Add text
     //
     if(i > 0 && i%stride == 0) {

        //
        // Choose the label format based on whether the colortable
        // has been rescaled
        //
        if(ct_ptr->rescale_flag) {
           sprintf(label, "%.2f", (*ct_ptr)[i].value_low());
        }
        else {
           sprintf(label, "%g", (*ct_ptr)[i].value_low());
        }

        p.write_centered_text(2, 1,  x_ll + bar_width + 2.0,
                              y_ll, 0.0, 0.5, label);
     }

     y_ll += bar_height;
   }

   p.grestore();

  return;
}

////////////////////////////////////////////////////////////////////////

void composite_mask(WrfData &comp, Engine &eng, int n_set, int fcst) {
   int i, x, y, n_shapes;
   WrfData junk;

   comp.set_size(grid.nx(), grid.ny());

   if(fcst) n_shapes = eng.n_fcst;
   else     n_shapes = eng.n_obs;

   for(i=0; i<n_shapes; i++) {

      if(fcst && !(eng.collection.set[n_set].has_fcst(i+1)) ) {
         continue;
      }

      if(!fcst && !(eng.collection.set[n_set].has_obs(i+1)) ) {
         continue;
      }

      if(fcst) junk = select(*(eng.fcst_split), i+1);
      else     junk = select(*(eng.obs_split), i+1);

      for(x=0; x<(comp.get_nx()); x++) {
         for(y=0; y<(comp.get_ny()); y++) {

            if(junk.get_xy_int(x, y)) {
               comp.put_xy_int(wrfdata_int_data_max, x, y);
            }
         }
      }
   }
   comp.calc_moments();

   return;
}

////////////////////////////////////////////////////////////////////////

void write_obj_stats() {
   char stat_file[max_str_len];
   ofstream out;
   int mon, day, yr, hr, min, sec;
   int l_hr, l_min, l_sec;
   int a_hr, a_min, a_sec;

   //
   // Create output stat file name
   //
   sec_to_hms(engine.fcst_raw->get_lead_time(), l_hr, l_min, l_sec);
   unix_to_mdyhms(engine.fcst_raw->get_valid_time(), mon, day, yr, hr, min, sec);
   sec_to_hms(engine.fcst_raw->get_accum_time(), a_hr, a_min, a_sec);
   sprintf(stat_file,
      "%s/mode_%s_%s_vs_%s_%s_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_%.2i%.2i%.2iA_obj.txt",
      out_dir,
      fcst_field_name, fcst_level_name,
      obs_field_name, obs_level_name,
      l_hr, l_min, l_sec,
      yr, mon, day, hr, min, sec,
      a_hr, a_min, a_sec);

   //
   // Open output stat file
   //
   out.open(stat_file);

   if(!out) {
      cerr << "\n\nERROR: write_obj_stats() -> "
           << "unable to open stats output file \""
           << stat_file << "\"\n\n";
      exit(1);
   }
   out.setf(ios::fixed);

   //
   // List stat file as it is being created
   //
   if(verbosity > 0) {
      cout << "Creating Fcst-Obs Object Statistics file: "
           << stat_file << "\n" << flush;
   }

   write_engine_stats(engine, grid, out);

   out << "\n" << flush;

   out.close();

   if(engine.wconf.fcst_merge_flag().ival() == 2 ||
      engine.wconf.fcst_merge_flag().ival() == 3) {

      sprintf(stat_file,
         "%s/mode_%s_%s_vs_%s_%s_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_%.2i%.2i%.2iA_fcst_merge.txt",
         out_dir,
         fcst_field_name, fcst_level_name,
         obs_field_name, obs_level_name,
         l_hr, l_min, l_sec,
         yr, mon, day, hr, min, sec,
         a_hr, a_min, a_sec);

      out.open(stat_file);

      if(!out) {
         cerr << "\n\nERROR: write_obj_stats() -> "
              << "unable to open stats output file \""
              << stat_file << "\"\n\n";
         exit(1);
      }
      out.setf(ios::fixed);

      if(verbosity > 0) {
         cout << "Creating Fcst-Fcst Object Statistics file: "
              << stat_file << "\n" << flush;
      }

      write_engine_stats(*engine.fcst_engine, grid, out);

      out << "\n" << flush;

      out.close();
   }

   if(engine.wconf.obs_merge_flag().ival() == 2 ||
      engine.wconf.obs_merge_flag().ival() == 3) {

      sprintf(stat_file,
         "%s/mode_%s_%s_vs_%s_%s_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_%.2i%.2i%.2iA_obs_merge.txt",
         out_dir,
         fcst_field_name, fcst_level_name,
         obs_field_name, obs_level_name,
         l_hr, l_min, l_sec,
         yr, mon, day, hr, min, sec,
         a_hr, a_min, a_sec);

      out.open(stat_file);

      if(!out) {
         cerr << "\n\nERROR: write_obj_stats() -> "
              << "unable to open stats output file \""
              << stat_file << "\"\n\n";
         exit(1);
      }
      out.setf(ios::fixed);

      if(verbosity > 0) {
         cout << "Creating Obs-Obs Object Statistics file: "
              << stat_file << "\n" << flush;
      }

      write_engine_stats(*engine.obs_engine, grid, out);

      out << "\n" << flush;

      out.close();
   }

   return;
}

////////////////////////////////////////////////////////////////////////

void write_obj_netcdf() {
   int n, x, y, yr, mon, day, hr, min, sec;
   int l_hr, l_min, l_sec;
   int a_hr, a_min, a_sec;
   double lat, lon;
   unixtime ut;

   char out_file[max_str_len];
   char attribute_str[max_str_len];
   char time_str[max_str_len];
   char hostname_str[max_str_len];

   float *lat_data = (float *) 0;
   float *lon_data = (float *) 0;
   int *fcst_obj_data = (int *) 0;
   int *fcst_comp_data = (int *) 0;
   int *obs_obj_data = (int *) 0;
   int *obs_comp_data = (int *) 0;

   NcFile *f_out = (NcFile *) 0;
   NcDim *lat_dim = (NcDim *) 0;
   NcDim *lon_dim = (NcDim *) 0;
   NcVar *lat_var = (NcVar *) 0;
   NcVar *lon_var = (NcVar *) 0;
   NcVar *fcst_obj_var = (NcVar *) 0;
   NcVar *fcst_comp_var = (NcVar *) 0;
   NcVar *obs_obj_var = (NcVar *) 0;
   NcVar *obs_comp_var = (NcVar *) 0;

   //
   // Create output NetCDF file name
   //
   sec_to_hms(engine.fcst_raw->get_lead_time(), l_hr, l_min, l_sec);
   unix_to_mdyhms(engine.fcst_raw->get_valid_time(),
                  mon, day, yr, hr, min, sec);
   sec_to_hms(engine.fcst_raw->get_accum_time(), a_hr, a_min, a_sec);
   sprintf(out_file,
      "%s/mode_%s_%s_vs_%s_%s_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_%.2i%.2i%.2iA_obj.nc",
      out_dir,
      fcst_field_name, fcst_level_name,
      obs_field_name, obs_level_name,
      l_hr, l_min, l_sec,
      yr, mon, day, hr, min, sec,
      a_hr, a_min, a_sec);

   if(verbosity > 0) {
      cout << "Creating Object NetCDF file: "
           << out_file << "\n" << flush;
   }

   //
   // Create a new NetCDF file and open it
   // NOTE: must multiply longitudes throughout by -1 to convert from
   // degrees_west to degree_east
   //
   f_out = new NcFile(out_file, NcFile::Replace);

   if(!f_out || !f_out->is_valid()) {
      cout << "\n\nERROR: write_obj_netcdf() -> trouble opening output file "
           << out_file << "\n\n";
      f_out->close();
      delete f_out;
      f_out = (NcFile *) 0;

      exit(1);
   }

   // Add global attributes
   f_out->add_att("Conventions", "COARDS");

   ut = time(NULL);
   unix_to_mdyhms(ut, mon, day, yr, hr, min, sec);
   sprintf(time_str, "%.4i-%.2i-%.2i %.2i:%.2i:%.2i", yr, mon, day, hr, min, sec);

   gethostname(hostname_str, max_str_len);

   sprintf(attribute_str, "File %s generated %s UTC on host %s",
      out_file, time_str, hostname_str);
   f_out->add_att("FileOrigins", attribute_str);

   f_out->add_att("Projection", grid.name());

   // Define Dimensions
   lat_dim = f_out->add_dim("lat", (long) grid.ny());
   lon_dim = f_out->add_dim("lon", (long) grid.nx());

   // Define Variables
   lat_var = f_out->add_var("lat", ncFloat, lat_dim, lon_dim);
   lon_var = f_out->add_var("lon", ncFloat, lat_dim, lon_dim);
   fcst_obj_var = f_out->add_var("fcst_obj_id", ncInt, lat_dim, lon_dim);
   fcst_comp_var = f_out->add_var("fcst_comp_id", ncInt, lat_dim, lon_dim);
   obs_obj_var = f_out->add_var("obs_obj_id", ncInt, lat_dim, lon_dim);
   obs_comp_var = f_out->add_var("obs_comp_id", ncInt, lat_dim, lon_dim);

   // Add variable attributes
   lat_var->add_att("units", "degrees_north");
   lat_var->add_att("long_name", "Latitude");
   lat_var->add_att("_FillValue", bad_data_float);

   lon_var->add_att("units", "degrees_east");
   lon_var->add_att("long_name", "Longitude");
   lon_var->add_att("_FillValue", bad_data_float);

   fcst_obj_var->add_att("long_name", "Forecast Object ID");
   fcst_obj_var->add_att("_FillValue", bad_data_int);

   fcst_comp_var->add_att("long_name", "Forecast Composite Object ID");
   fcst_comp_var->add_att("_FillValue", bad_data_int);

   obs_obj_var->add_att("long_name", "Observation Object ID");
   obs_obj_var->add_att("_FillValue", bad_data_int);

   obs_comp_var->add_att("long_name", "Observation Composite Object ID");
   obs_comp_var->add_att("_FillValue", bad_data_int);

   //
   // Allocate memory to store the Lat/Lon values for each grid point
   //
   lat_data = new float [grid.nx()*grid.ny()];
   lon_data = new float [grid.nx()*grid.ny()];

   //
   // Allocate memory for the object ID's for each grid box
   //
   fcst_obj_data  = new int [grid.nx()*grid.ny()];
   fcst_comp_data = new int [grid.nx()*grid.ny()];
   obs_obj_data   = new int [grid.nx()*grid.ny()];
   obs_comp_data  = new int [grid.nx()*grid.ny()];

   for(x=0; x<grid.nx(); x++) {
      for(y=0; y<grid.ny(); y++) {

         //
         // Lat/Lon values for each grid point
         //
         grid.xy_to_latlon(x, y, lat, lon);

         n = two_to_one(grid.nx(), grid.ny(), x, y);

         lat_data[n] = (float) lat;
         lon_data[n] = (float) -1.0*lon;

         //
         // Object ID's for each grid box
         //
         if(engine.fcst_split->get_xy_int(x, y) > 0) {
            fcst_obj_data[n] = engine.fcst_split->get_xy_int(x, y);
         }
         else {
            fcst_obj_data[n] = bad_data_int;
         }

         if(engine.fcst_comp_split->get_xy_int(x, y) > 0) {
            fcst_comp_data[n] = engine.fcst_comp_split->get_xy_int(x, y);
         }
         else {
            fcst_comp_data[n] = bad_data_int;
         }

         if(engine.obs_split->get_xy_int(x, y) > 0) {
            obs_obj_data[n] = engine.obs_split->get_xy_int(x, y);
         }
         else {
            obs_obj_data[n] = bad_data_int;
         }

         if(engine.obs_comp_split->get_xy_int(x, y) > 0) {
            obs_comp_data[n] = engine.obs_comp_split->get_xy_int(x, y);
         }
         else {
            obs_comp_data[n] = bad_data_int;
         }
      }
   }

   if( !lat_var->put(&lat_data[0], grid.ny(), grid.nx())
    || !lon_var->put(&lon_data[0], grid.ny(), grid.nx()) ) {

      cerr << "\n\nERROR: write_obj_netcdf() -> "
           << "error with the lat_var->put or lon_var->put\n\n" << flush;
      exit(1);
   }

   if( !fcst_obj_var->put(&fcst_obj_data[0], grid.ny(), grid.nx())
    || !fcst_comp_var->put(&fcst_comp_data[0], grid.ny(), grid.nx())
    || !obs_obj_var->put(&obs_obj_data[0], grid.ny(), grid.nx())
    || !obs_comp_var->put(&obs_comp_data[0], grid.ny(), grid.nx()) ) {

      cerr << "\n\nERROR: write_obj_netcdf() -> "
           << "error with the fcst_obj_var->put, fcst_comp_var->put, "
           << "obs_obj_var, or obs_comp_var\n\n" << flush;
      exit(1);
   }

   if(lat_data)       { delete lat_data;       lat_data = (float *) 0; }
   if(lon_data)       { delete lon_data;       lon_data = (float *) 0; }
   if(fcst_obj_data)  { delete fcst_obj_data;  fcst_obj_data = (int *) 0; }
   if(fcst_comp_data) { delete fcst_comp_data; fcst_comp_data = (int *) 0; }
   if(obs_obj_data)   { delete obs_obj_data;   obs_obj_data = (int *) 0; }
   if(obs_comp_data)  { delete obs_comp_data;  obs_comp_data = (int *) 0; }

   f_out->close();
   delete f_out;
   f_out = (NcFile *) 0;

   return;
}

////////////////////////////////////////////////////////////////////////

void write_ct_stats() {
   char stat_file[max_str_len], line[max_line_len];
   char sep_char = ' ';
   ofstream out;
   int i, mon, day, yr, hr, min, sec, status;
   int l_hr, l_min, l_sec;
   int a_hr, a_min, a_sec;
   double v;

   model_width = max(strlen("MODEL"), strlen(engine.wconf.model().sval()));

   //
   // Create output contingency table stat file name
   //
   sec_to_hms(engine.fcst_raw->get_lead_time(), l_hr, l_min, l_sec);
   unix_to_mdyhms(engine.fcst_raw->get_valid_time(),
                  mon, day, yr, hr, min, sec);
   sec_to_hms(engine.fcst_raw->get_accum_time(), a_hr, a_min, a_sec);
   sprintf(stat_file,
      "%s/mode_%s_%s_vs_%s_%s_%.2i%.2i%.2iL_%.4i%.2i%.2i_%.2i%.2i%.2iV_%.2i%.2i%.2iA_cts.txt",
      out_dir,
      fcst_field_name, fcst_level_name,
      obs_field_name, obs_level_name,
      l_hr, l_min, l_sec,
      yr, mon, day, hr, min, sec,
      a_hr, a_min, a_sec);

   //
   // Open output stat file
   //
   out.open(stat_file);

   if(!out) {
      cerr << "\n\nERROR: write_ct_stats() -> "
           << "unable to open stats output file \""
           << stat_file << "\"\n\n";
      exit(1);
   }
   out.setf(ios::fixed);

   //
   // List stat file as it is being created
   //
   if(verbosity > 0) {
      cout << "Creating Contingency Table Statistics file: "
           << stat_file << "\n" << flush;
   }

   //
   // Write out a header line
   //
   out.precision(default_precision);
   out.setf(ios::left);

   // Model Name
   out.width(model_width); out << "MODEL" << sep;

   // Forecast lead time
   out.width(fcst_lead_width); out << "FCST_LEAD" << sep_char;
   // Forecast valid time
   out.width(fcst_valid_width); out << "FCST_VALID" << sep_char;
   // Forecast accumulation time
   out.width(fcst_accum_width); out << "FCST_ACCUM" << sep_char;

   // Observation (lead2) lead time
   out.width(lead2_width); out << "LEAD2" << sep_char;
   // Observation valid time
   out.width(obs_valid_width); out << "OBS_VALID" << sep_char;
   // Observation accumulation time
   out.width(obs_accum_width); out << "OBS_ACCUM" << sep_char;

   // Forecast convolution radius in grid squares
   out.width(rad_width); out << "FCST_RAD" << sep_char;
   // Forecast convolution threshold
   out.width(thresh_width); out << "FCST_THR" << sep_char;
   // Observation convolution radius
   out.width(rad_width); out << "OBS_RAD" << sep_char;
   // Observation convolution threshold
   out.width(thresh_width); out << "OBS_THR" << sep_char;
   // Forecast variable being analyzed
   out.width(cts_width); out << "FCST_VAR" << sep_char;
   // Forecast variable level being analyzed
   out.width(cts_width); out << "FCST_LEV" << sep_char;
   // Observation variable being analyzed
   out.width(cts_width); out << "OBS_VAR" << sep_char;
   // Observation variable level being analyzed
   out.width(cts_width); out << "OBS_LEV" << sep_char;
   // Type of field stats (raw, filter, object)
   out.width(cts_width); out << "FIELD" << sep_char;

   out << "=" << sep_char;

   // Total contingency table count
   out.width(cts_width); out << "TOTAL" << sep_char;
   // Count of Forecast Yes/Observation Yes
   out.width(cts_width); out << "FY_OY" << sep_char;
   // Count of Forecast Yes/Observation No
   out.width(cts_width); out << "FY_ON" << sep_char;
   // Count of Forecast No/Observation Yes
   out.width(cts_width); out << "FN_OY" << sep_char;
   // Count of Forecast No/Observation No
   out.width(cts_width); out << "FN_ON" << sep_char;
   // Base Rate
   out.width(cts_width); out << "BASER" << sep_char;
   // Forecast Mean
   out.width(cts_width); out << "FMEAN" << sep_char;
   // Accuracy
   out.width(cts_width); out << "ACC" << sep_char;
   // Frequency Bias
   out.width(cts_width); out << "FBIAS" << sep_char;
   // Probability of Detecting Yes
   out.width(cts_width); out << "PODY" << sep_char;
   // Probability of Detecting No
   out.width(cts_width); out << "PODN" << sep_char;
   // Probability of False Detection
   out.width(cts_width); out << "POFD" << sep_char;
   // False Alarm Ratio
   out.width(cts_width); out << "FAR" << sep_char;
   // Critical Success Index
   out.width(cts_width); out << "CSI" << sep_char;
   // Gilbert Skill Score
   out.width(cts_width); out << "GSS" << sep_char;
   // Hanssen and Kuipers Discriminant
   out.width(cts_width); out << "HK" << sep_char;
   // Heidke Skill Score
   out.width(cts_width); out << "HSS" << sep_char;
   // Odds Ratio
   out.width(cts_width); out << "ODDS" << sep_char;

   out << "\n";

   //
   // Write out contingency table counts
   //
   for(i=0; i<n_cts; i++) {

      // Model Name
      out.width(model_width);
      out << engine.wconf.model().sval() << sep;

      // Forecast lead time
      sec_to_hms(engine.fcst_raw->get_lead_time(), hr, min, sec);
      sprintf(line, hms_fmt, hr, min, sec);
      out.width(fcst_lead_width); out << line << sep;

      // Forecast valid time
      unix_to_mdyhms(engine.fcst_raw->get_valid_time(),
                     mon, day, yr, hr, min, sec);
      sprintf(line, ymd_hms_fmt, yr, mon, day, hr, min, sec);
      out.width(fcst_valid_width); out << line << sep;

      // Forecast accumulation time
      sec_to_hms(engine.fcst_raw->get_accum_time(), hr, min, sec);
      sprintf(line, hms_fmt, hr, min, sec);
      out.width(fcst_accum_width); out << line << sep;

      // Observation lead time
      sec_to_hms(engine.obs_raw->get_lead_time(), hr, min, sec);
      sprintf(line, hms_fmt, hr, min, sec);
      out.width(lead2_width); out << line << sep;

      // Observation valid time
      unix_to_mdyhms(engine.obs_raw->get_valid_time(),
                     mon, day, yr, hr, min, sec);
      sprintf(line, ymd_hms_fmt, yr, mon, day, hr, min, sec);
      out.width(obs_valid_width); out << line << sep;

      // Observation accumulation time
      sec_to_hms(engine.obs_raw->get_accum_time(), hr, min, sec);
      sprintf(line, hms_fmt, hr, min, sec);
      out.width(obs_accum_width); out << line << sep;

      // Forecast convolution radius
      out.width(rad_width); out.precision(integer_precision);
      out << engine.wconf.fcst_conv_radius().ival() << sep_char;

      // Forecast convolution threshold
      engine.fcst_conv_threshold.get_str(line);
      out.width(thresh_width); out << line << sep_char;

      // Observation convolution radius
      out.width(rad_width); out.precision(integer_precision);
      out << engine.wconf.obs_conv_radius().ival() << sep_char;

      // Observation convolution threshold
      engine.obs_conv_threshold.get_str(line);
      out.width(thresh_width); out << line << sep_char;

      // Forecast Variable Name
      out.width(cts_width);
      out << fcst_field_name << sep_char;

      // Forecast Variable Level
      out.width(cts_width);
      out << fcst_level_name << sep_char;

      // Observation Variable Name
      out.width(cts_width);
      out << obs_field_name << sep_char;

      // Observation Variable Level
      out.width(cts_width);
      out << obs_level_name << sep_char;

      // Name of the Field
      out.width(cts_width); out << cts[i].name() << sep_char;

      out << "=" << sep_char;

      out.precision(default_precision);

      out.width(cts_width); out << cts[i].n() << sep_char;
      out.width(cts_width); out << cts[i].fy_oy() << sep_char;
      out.width(cts_width); out << cts[i].fy_on() << sep_char;
      out.width(cts_width); out << cts[i].fn_oy() << sep_char;
      out.width(cts_width); out << cts[i].fn_on() << sep_char;

      // Base Rate
      v = cts[i].oy_tp(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      // Forecast Mean
      v = cts[i].fy_tp(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].accuracy(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].fbias(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].pod_yes(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].pod_no(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].pofd(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].far(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].csi(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].gss(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].hk(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].hss(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      v = cts[i].odds(status);
      if(status != 0) v = bad_data_double;
      out.width(cts_width); out << v << sep_char;

      out << "\n";
   }

   out << "\n" << flush;

   out.close();

   return;
}

////////////////////////////////////////////////////////////////////////

void set_dim(BoundingBox &dim, double y_ll, double y_ur, double x_cen) {
   double mag;

   dim.y_ll   = y_ll;
   dim.y_ur   = y_ur;
   dim.height = dim.y_ur - dim.y_ll;
   mag        = dim.height/xy_bb.height;
   dim.width  = mag*xy_bb.width;
   dim.x_ll   = x_cen - 0.5*dim.width;
   dim.x_ur   = x_cen + 0.5*dim.width;

   return;
}

////////////////////////////////////////////////////////////////////////

void set_xy_bb() {
   BoundingBox fcst_xy_bb, obs_xy_bb;

   //
   // Check the value of the plot_valid_flag in the config file.
   // If not set, reset the xy_bb to the entire grid.
   //
   if(engine.wconf.plot_valid_flag().ival() == 0) {
      xy_bb.x_ll   = 0;
      xy_bb.y_ll   = 0;
      xy_bb.x_ur   = grid.nx();
      xy_bb.y_ur   = grid.ny();
      xy_bb.width  = xy_bb.x_ur - xy_bb.x_ll;
      xy_bb.height = xy_bb.y_ur - xy_bb.y_ll;
   }
   //
   // Search the raw forecast and raw observation fields to define a
   // bounding box around the regions of valid data
   //
   else {

      //
      // Compute the x/y bounding box for valid data in each field
      //
      valid_xy_bb(engine.fcst_raw, fcst_xy_bb);
      valid_xy_bb(engine.obs_raw, obs_xy_bb);

      //
      // Compute the x/y bounding box as the union of the
      // fcst and obs x/y bounding boxes
      //
      xy_bb.x_ll = min(fcst_xy_bb.x_ll, obs_xy_bb.x_ll);
      xy_bb.y_ll = min(fcst_xy_bb.y_ll, obs_xy_bb.y_ll);
      xy_bb.x_ur = max(fcst_xy_bb.x_ur, obs_xy_bb.x_ur);
      xy_bb.y_ur = max(fcst_xy_bb.y_ur, obs_xy_bb.y_ur);

      xy_bb.width  = xy_bb.x_ur - xy_bb.x_ll;
      xy_bb.height = xy_bb.y_ur - xy_bb.y_ll;
   }

   return;
}

////////////////////////////////////////////////////////////////////////
//
// Compute the smallest rectangle around the region of valid data in
// the field.
//
////////////////////////////////////////////////////////////////////////

void valid_xy_bb(const WrfData *wd_ptr, BoundingBox &bb) {
   int x, y;

   //
   // Initialize the x/y bounding box
   //
   bb.x_ll = grid.nx();
   bb.y_ll = grid.ny();
   bb.x_ur = 0;
   bb.y_ur = 0;

   for(x=0; x<wd_ptr->get_nx(); x++) {
      for(y=0; y<wd_ptr->get_ny(); y++) {

         if(wd_ptr->is_valid_data(x, y)) {

            if(x < bb.x_ll) bb.x_ll = x;
            if(x > bb.x_ur) bb.x_ur = x;
            if(y < bb.y_ll) bb.y_ll = y;
            if(y > bb.y_ur) bb.y_ur = y;
         }
      } // end for y
   } // end for x

   bb.width  = bb.x_ur - bb.x_ll;
   bb.height = bb.y_ur - bb.y_ll;

   return;
}
////////////////////////////////////////////////////////////////////////

void set_ll_bb() {
   int x, y;

   ll_bb.x_ll = -180.0; // Initialize the max lon
   ll_bb.y_ll = 90.0;   // Initialize the min lat
   ll_bb.x_ur = 180.0;  // Initialize the min lon
   ll_bb.y_ur = -90.0;  // Initialize the max lat

   // Search the bottom edge of the grid
   for(x=nint(xy_bb.x_ll); x<xy_bb.x_ur; x++) check_xy_ll(x, nint(xy_bb.y_ll));

   // Search the top edge of the grid
   for(x=nint(xy_bb.x_ll); x<xy_bb.x_ur; x++) check_xy_ll(x, nint(xy_bb.y_ur-1));

   // Search the left side of the grid
   for(y=nint(xy_bb.y_ll); y<xy_bb.y_ur; y++) check_xy_ll(nint(xy_bb.x_ll), y);

   // Search the right side of the grid
   for(y=nint(xy_bb.y_ll); y<xy_bb.y_ur; y++) check_xy_ll(nint(xy_bb.x_ur-1), y);

   return;
}

////////////////////////////////////////////////////////////////////////

void check_xy_ll(int x, int y) {
   double lat, lon;

   // Convert to lat/lon
   grid.xy_to_latlon(x, y, lat, lon);

   // Check limits
   if(lat > ll_bb.y_ur) ll_bb.y_ur = lat;
   if(lat < ll_bb.y_ll) ll_bb.y_ll = lat;
   if(lon > ll_bb.x_ll) ll_bb.x_ll = lon;
   if(lon < ll_bb.x_ur) ll_bb.x_ur = lon;

   return;
}

////////////////////////////////////////////////////////////////////////

void usage(int argc, char *argv[]) {

   cout << "\nUsage: "
        << program_name << "\n"
        << "\tfcst_file\n"
        << "\tobs_file\n"
        << "\tconfig_file\n"
        << "\t[-config_merge merge_config_file]\n"
        << "\t[-outdir path]\n"
        << "\t[-plot]\n"
        << "\t[-obj_plot]\n"
        << "\t[-obj_stat]\n"
        << "\t[-ct_stat]\n"
        << "\t[-v level]\n\n"

        << "\twhere\t\"fcst_file\" is a forecast file in either Grib "
        << "or netCDF format (output of pcp_combine) containing the "
        << "field to be verified (required).\n"

        << "\t\t\"obs_file\" is an observation file in either Grib "
        << "or netCDF format (output of pcp_combine) containing the "
        << "verifying field (required).\n"

        << "\t\t\"config_file\" is a WrfModeConfig file containing "
        << "the desired configuration settings (required).\n"

        << "\t\t\"-config_merge merge_config_file\" overrides the default "
        << "fuzzy engine settings for merging within the fcst/obs fields "
        << "(optional).\n"

        << "\t\t\"-outdir path\" overrides the default output directory ("
        << out_dir << ") (optional).\n"

        << "\t\t\"-plot\" disables plotting (optional).\n"

        << "\t\t\"-obj_plot\" disables the output of the object split "
        << "and composite fields to a NetCDF file (optional).\n"

        << "\t\t\"-obj_stat\" disables the output of the object "
        << "statistics file (optional).\n"

        << "\t\t\"-ct_stat\" disables the output of the contingency "
        << "table standard statistics file (optional).\n"

        << "\t\t\"-v level\" overrides the default level of logging ("
        << verbosity << ") (optional).\n"

        << "\n\tNOTE: The forecast and observation fields must be "
        << "on the same grid.\n\n" << flush;

   return;
}

////////////////////////////////////////////////////////////////////////
